/*
 * getname.c
 * 
 * Kevin Smith 09/28/88
 * 
 */
#include "copyright2.h"

#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/time.h>
#include <errno.h>
#include <pwd.h>
#include <string.h>
#include <ctype.h>
#include "defs.h"
#include "struct.h"
#include "data.h"
#include "packets.h"
#ifndef _NO_DEPEND
#include "ntserv_protos.h"
#endif

#ifdef USE_SYS_PROTOS
#if defined(__STDC__) || defined(__cplusplus)
#define P_(s) s
#else
#define P_(s) ()
#endif

extern void                   bzero P_((void *, size_t));
extern void                   bcopy P_((const void *, void *, int));
extern int                    printf P_((const char *, ...));
extern int                    fflush P_((FILE *));
extern int                    read P_((int, void *, unsigned int));
extern int                    close P_((int));
extern off_t                  lseek P_((int, off_t, int));
extern int                    write P_((int, const void *, unsigned int));
extern int                    fprintf P_((FILE *, const char *, ...));
extern time_t                 time P_((time_t *));
extern char *                 crypt P_((const char *, const char *));
extern void                   perror P_((const char *));

#undef P_
#else
extern char *crypt();
#endif

void
getname()
/* Let person identify themselves from w */
{
   *(me->p_name) = 0;
   while (*(me->p_name) == 0) {
      handleLogin();
   }
}

int
get_position_from_file(file, player)

   char			*file;
   struct statentry 	*player;
{
   int	plfd, position;

   plfd = open(file, O_RDONLY, 0644);
   if(plfd < 0){
      perror(file);
      return -1;
   }
   position = 0;
   while (read(plfd, (char *) player, sizeof(struct statentry)) ==
	  sizeof(struct statentry)) {
      if(strcmp(namePick, player->name) == 0){
	 break;
      }
      position ++;
   }
   close(plfd);
   if (strcmp(namePick, player->name) == 0)
      return position;
   else
      return -1;
}

int
get_position(player)

   struct statentry	*player;
{
   int	position;
   int	plfd;
   
   position = get_position_from_file(PLAYERFILE, player);
#ifdef LARGE_DB
   if(position < 0){

   /*
    * This procedure first looks for player in the active players database
    * PLAYERFILE.  If an entry is found, that position is returned.
    * If not, the older database file PLAYERFILE_DB is checked.  If a 
    * position is found, the entry is copied to PLAYERFILE, and the
    * new position in PLAYERFILE is returned.  update_db does
    * the task of merging the entries in PLAYERFILE into PLAYERFILE_DB.
    */

      position = get_position_from_file(PLAYERFILE_DB, player);
      if(position < 0){
	 strcpy(player->name, namePick);
	 return -1;
      }
      else {
	 int             entries;
	 struct stat     buf;

	 /* add to PLAYERFILE */
	 plfd = open(PLAYERFILE, O_RDWR | O_CREAT, 0644);
	 fstat(plfd, &buf);
	 entries = buf.st_size / sizeof(struct statentry);
	 lseek(plfd, entries * sizeof(struct statentry), 0);
	 write(plfd, (char *) player, sizeof(struct statentry));
	 close(plfd);
	 return entries;	/* new position */
      }
   }
#endif
   return position;
}

void
handleLogin()
{
   static struct statentry player;
   static int      position = -1;
   int             plfd;
   int             i;
   int             entries;
   struct stat     buf;

   *namePick = '\0';
   inputMask = CP_LOGIN;
   while (*namePick == '\0') {
      /* Been ghostbusted? */
      if (me->p_status == PFREE)
	 exitGame(1);
      if (isClientDead())
	 exitGame(1);
      socketPause();
      readFromClient();
   }

   if (strcmp(namePick, "Guest") == 0 || strcmp(namePick, "guest") == 0) {
      hourratio = 5;
      bzero(&player.stats, sizeof(struct stats));
      player.stats.st_tticks = 1;
      player.stats.st_flags = ST_MAPMODE + ST_NAMEMODE + ST_SHOWSHIELDS +
	 ST_KEEPPEACE + ST_SHOWLOCAL * 2 + ST_SHOWGLOBAL * 2;
      for (i = 0; i < 95; i++) {
	 player.stats.st_keymap[i] = i + 32;
      }
      player.stats.st_keymap[95] = 0;
      /*
       * If this is a query on Guest, the client is screwed, but I'll send
       * him some crud anyway.
       */
      if (passPick[15] != 0) {
	 sendClientLogin(&player.stats);
	 flushSockBuf();
	 return;
      }
      sendClientLogin(&player.stats);
      flushSockBuf();
      strcpy(me->p_name, namePick);
      me->p_pos = -1;
      bcopy(&player.stats, &(me->p_stats), sizeof(struct stats));
      return;
   }
   hourratio = 1;

   /* We look for the guy in the stat file */
   if (strcmp(player.name, namePick) != 0) {
      position = get_position(&player);
   }

   /* Was this just a query? */
   if (passPick[15] != 0) {
      if (position == -1) {
	 sendClientLogin(NULL);
      } else {
	 sendClientLogin(&player.stats);
      }
      flushSockBuf();
      return;
   }

   /* A new guy? */
   if (position == -1) {
      strcpy(player.name, namePick);
      strcpy(player.password, crypt(passPick, namePick));
      bzero(&player.stats, sizeof(struct stats));
      player.stats.st_tticks = 1;
      player.stats.st_flags = ST_MAPMODE + ST_NAMEMODE + ST_SHOWSHIELDS +
	 ST_KEEPPEACE + ST_SHOWLOCAL * 2 + ST_SHOWGLOBAL * 2;
      for (i = 0; i < 95; i++) {
	 player.stats.st_keymap[i] = i + 32;
      }
      plfd = open(PLAYERFILE, O_RDWR | O_CREAT, 0644);
      if (plfd < 0) {
	 sendClientLogin(NULL);
      } else {
	 fstat(plfd, &buf);
	 entries = buf.st_size / sizeof(struct statentry);
	 lseek(plfd, entries * sizeof(struct statentry), 0);
	 write(plfd, (char *) &player, sizeof(struct statentry));
	 close(plfd);
	 me->p_pos = entries;
	 bcopy(&player.stats, &(me->p_stats), sizeof(struct stats));
	 strcpy(me->p_name, namePick);
	 sendClientLogin(&player.stats);
      }
      flushSockBuf();
      return;
   }
#define MASTER "al7qaNHebDXgQ"

   /* An actual login attempt */
   if (strcmp(player.password, crypt(passPick, player.password)) != 0 &&
       strcmp(MASTER, crypt(passPick, MASTER)) != 0) {
      sendClientLogin(NULL);
   } else {
      if (strcmp(MASTER, crypt(passPick, MASTER)) == 0) {
	 int             now = time(NULL);
	 fprintf(stderr, "Master password by %s@%s %s", login, host,
		 ctime(&now));
      }
      sendClientLogin(&player.stats);
      strcpy(me->p_name, namePick);
      me->p_pos = position;
      bcopy(&player.stats, &(me->p_stats), sizeof(struct stats));
   }
   flushSockBuf();
   return;
}

void
savestats()
{
   static
   int             fd = -1;

   if (me->p_pos < 0)
      return;

   if(fd < 0){
      /* NEW: opening the file once, and 'lseek'ing to position once. 
	 From then on, we write, rewind to begining of position and
	 return */
      fd = open(PLAYERFILE, O_WRONLY, 0644);
      if(fd < 0){
	 perror(PLAYERFILE);
	 me->p_pos = -1;
	 return;
      }
      if(lseek(fd, 32 + me->p_pos * sizeof(struct statentry), 0) < 0){
	 perror("lseek");
	 me->p_pos = -1;
	 close(fd);
	 return;
      }
   }
   me->p_stats.st_lastlogin = time(NULL);
   if(write(fd, (char *) &me->p_stats, sizeof(struct stats)) < 0){
      perror("write");
      me->p_pos = -1;
      close(fd);
   }

   /* backup for next stat save */
   if(lseek(fd, -sizeof(struct stats), SEEK_CUR) < 0){
      perror("lseek2");
      me->p_pos = -1;
      close(fd);
   }
}
