/*
 * main.c
 */
#include "copyright.h"

#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <setjmp.h>
#include <ctype.h>
#include <pwd.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include "defs.h"
#include "struct.h"
#include "data.h"
#include "packets.h"
#include "patchlevel.h"
#include "version.h"
#ifndef _NO_DEPEND
#include "ntserv_protos.h"
#endif

#ifdef USE_SYS_PROTOS
#if defined(__STDC__) || defined(__cplusplus)
#define P_(s) s
#else
#define P_(s) ()
#endif

extern uid_t getuid P_((void));
extern uid_t geteuid P_((void));
extern int atoi P_((const char *));
extern int fprintf P_((FILE *, const char *,...));
extern int fflush P_((FILE *));
extern void srandom P_((int));
extern pid_t getpid P_((void));
extern time_t time P_((time_t *));
extern int fclose P_((FILE *));
extern int printf P_((const char *,...));
extern void perror P_((const char *));
extern pid_t wait3 P_((union wait *, int, struct rusage *));
extern int strcasecmp P_((char *s1, char *s2));
extern void bzero P_((void *, size_t));
extern int tolower P_((int));

#undef P_
#endif

jmp_buf         env;

int             ignored[MAXPLAYER];

#ifdef COPFIX
/* variables to try to catch the co-pilot problem */
char            start_login[16];
char            start_host[16];
#endif

int             overload = 0;	/* overload indicates a request for a
				 * reserved slot */
int             isbanned = 0;

void
main(argc, argv)
    int             argc;
    char          **argv;
{
   int             team, s_type;
   int             pno, i;
   int             usage = 0;
   int             err = 0;
   char           *name, *ptr;
   struct passwd  *pwent, *getpwuid();
   int             callHost = 0;
   long            starttime;

   name = *argv++;
   argc--;
   if ((ptr = strrchr(name, '/')) != NULL)
      name = ptr + 1;
   while (*argv) {
      if (**argv == '-')
	 ++* argv;
      else
	 break;

      argc--;
      ptr = *argv++;
      while (*ptr) {
	 switch (*ptr) {
	 case 'u':
	    usage++;
	    break;
	 case 'R':
	    if (getuid() == geteuid())
	       overload++;
	    break;
	 case 's':
	    xtrekPort = atoi(*argv);
	    callHost = 1;
	    argv++;
	    argc--;
	    break;
	 case 'd':
	    host = *argv;
	    argc--;
	    argv++;
	    break;
	 default:
	    fprintf(stderr, "%s: unknown option '%c'\n", name, *ptr);
	    fflush(stderr);
	    err++;
	    break;
	 }
	 ptr++;
      }
   }
   if (usage || err) {
      printUsage(name);
      exit(err);
   }
   /* compatability */
   if (argc > 0)
      host = argv[0];
   strcpy(nombreDios, "Destiny");
   srandom(getpid() * time(NULL));
   readsysdefaults();
#ifdef FEATURE
   readFeatures();
#endif
   /* this finds the shared memory information */
   openmem();
   me = NULL;			/* UDP fix */
   if (callHost) {
      if (!connectToClient(host, xtrekPort)) {
	 exit(0);
      }
   } else {
      sock = 0;			/* Because we were forked by inetd! */
      checkSocket();
      initClientData();		/* "normally" called by connectToClient() */
   }
   starttime = time(NULL);
   inputMask = 0;		/* added -- TSH */
   while (userVersion == 0) {
      /*
       * Waiting for user to send his version number. We give him ten seconds
       * to do so...
       */
      if (starttime + 10 < time(NULL)) {
	 exit(1);
      }
      socketPause();
      readFromClient();
   }
   if (!checkVersion())
      exit(1);
#ifdef DS
   signal(SIGALRM, SIG_IGN);
   signal(SIGTERM, forceShutdown);
#endif
   sendMotd();
   pno = findslot(overload);
   if (pno < 0) {
      /* print some appropriate message */
      exit(1);
   }
   me = &players[pno];
   me->p_no = pno;
#ifdef DS
   me->p_process = getpid();
   me->p_timerdelay = timerDelay / 100000;
#endif

   myship = &me->p_ship;
   mystats = &me->p_stats;
   lastm = mctl->mc_current;
   signal(SIGINT, SIG_IGN);
   (void) signal(SIGCHLD, reaper);

   /*
    * We set these so we won't bother updating him on the location of the
    * other players in the galaxy which he is not near.  There is no real
    * harm to doing this, except that he would then get more information than
    * he deserves. It is kind of a hack, but should be harmless.
    */
   me->p_x = -100000;
   me->p_y = -100000;
   me->p_team = 0;
#ifdef RCD
   me->p_gen_distress = 0;
#endif
   updateSelf();		/* so he gets info on who he is */
   updateShips(1);
   updatePlanets();
   flushSockBuf();

   /* Get login name */
   if ((pwent = getpwuid(getuid())) != NULL)
      (void) strncpy(login, pwent->pw_name, sizeof(login));
   else
      (void) strncpy(login, "Bozo", sizeof(login));
   login[sizeof(login) - 1] = '\0';

   strcpy(pseudo, "Guest");

   strcpy(me->p_name, pseudo);
   me->p_team = ALLTEAM;
   getname();
   strcpy(pseudo, me->p_name);
   keeppeace = (me->p_stats.st_flags / ST_KEEPPEACE) & 1;

   /*
    * Set p_hostile to hostile, so if keeppeace is on, the guy starts off
    * hating everyone (like a good fighter should)
    */
   me->p_hostile = (FED | ROM | KLI | ORI);

   s_type = CRUISER;
   me->p_planets = 0;
   me->p_genoplanets = 0;
   me->p_armsbomb = 0;
   me->p_genoarmsbomb = 0;
   /* Set up a reasonable default */
   me->p_whydead = KQUIT;

   (void) strncpy(me->p_login, login, sizeof(me->p_login));
   me->p_login[sizeof(me->p_login) - 1] = '\0';
   (void) strncpy(me->p_monitor, host, sizeof(me->p_monitor));
   me->p_monitor[sizeof(me->p_monitor) - 1] = '\0';

   logEntry();
#ifdef COPFIX
   /* these are checked in redraw.c */
   strncpy(start_login, me->p_login, sizeof(start_login));
   start_login[sizeof(start_login) - 1] = '\0';
   strncpy(start_host, me->p_monitor, sizeof(start_host));
   start_host[sizeof(start_host) - 1] = '\0';
#endif

   for (i = 0; i < MAXPLAYER; i++)
      ignored[i] = 0;

   /* reset ping stats */
   me->p_avrt = -1;
   me->p_stdv = -1;
   me->p_pkls_c_s = -1;
   me->p_pkls_s_c = -1;

#ifdef CYBORGS			/* clear cyborg flag */
   me->p_flags &= ~PFCYBORG;
#endif

   startTkills = me->p_stats.st_tkills;
   startTlosses = me->p_stats.st_tlosses;
   startTarms = me->p_stats.st_tarmsbomb;
   startTplanets = me->p_stats.st_tplanets;
   startTticks = me->p_stats.st_tticks;

   startSBkills = me->p_stats.st_sbkills;
   startSBlosses = me->p_stats.st_sblosses;
   startSBticks = me->p_stats.st_sbticks;
#ifdef BASEPRACTICE
   startTdami = me->p_stats.st_dami;
   startTdamr = me->p_stats.st_damr;
#endif

#ifdef VOTE
   vote_no(&vote->v_resetgalaxy_votes, NULL);
   vote_no(&vote->v_resetplanet_votes, NULL);
#endif

   if (checkspecial(SLOTSCUM_FILE, login, host) == TRUE) {
      /*
       * this person has slot scummed in the past .. therefore we don't allow
       * mutiple slots
       */
      if (checkslots(login, host)) {
	 warning(2, "YOU are not allowed to have multiple slots active.", -1);
	 mustexit = 1;
      }
   }
   if (checkspecial(BYPASS_FILE, login, host) == TRUE)
      bypassed = 1;
   else
      bypassed = 0;

   if (checkspecial(BANNED_FILE, login, host) == TRUE)
      isbanned = 1;

   setjmp(env);			/* Reentry point of game */
   /* give the player the motd and find out which team he wants */
   if (me->p_status != PALIVE) {
      me->p_x = -100000;
      me->p_y = -100000;
      updateSelf();
      updateShips(1);
      teamPick = -1;
      flushSockBuf();
      getEntry(&team, &s_type);
      repCount = 0;		/* Make sure he gets an update immediately */
   }
   if (team == -1) {
      exitGame(0);
   }
   redrawall = 1;
   inputMask = -1;		/* Allow all input now */
   enter(team, host, pno, s_type);
   /*
    * for (i = 0; i < NSIG; i++) { signal(i, SIG_IGN); }
    */

   me->p_status = PALIVE;	/* Put player in game */
   me->p_ghostbuster = 0;
   if (isbanned) {
      char            buf[256];
      FILE           *logfile;
      int             curtime;

      sprintf(buf, "%s (%s@%s) is banned from the game.",
	      me->p_name, me->p_login, host);
      pmessage(buf, 0, MALL, "GOD->ALL");
      warning(2, "You are banned from the game.", -1);
      me->p_explode = 100 / UPDATE_NORM;
      me->p_status = PEXPLODE;
      me->p_whydead = KQUIT;	/* make him self destruct */
      me->p_whodead = 0;
      logfile = fopen(LOGFILENAME, "a");
      if (logfile) {
	 curtime = time(NULL);
	 fprintf(logfile, "Banned and exiting: %s (%s@%s), (%c), %s",
	 me->p_name, me->p_login, host, shipnos[me->p_no], ctime(&curtime));
	 fclose(logfile);
      }
   }
   /* Get input until the player quits or dies */
   input();
}

void
exitGame(gb)
    int             gb;
{
   char            buf[256];
   char            addrbuf[20];

   if (me != NULL && me->p_team != ALLTEAM) {
      savestats();
      printStats();
      if (!gb) {
#ifdef LONGHOST
	 /* new-style leaving message 4/13/92 TC */
	 sprintf(buf, "%s %s (%s) leaving game (%.16s@%.32s)",
		 ranks[me->p_stats.st_rank].name,
		 me->p_name,
		 me->p_mapchars,
		 me->p_login,
		 me->p_monitor);
#else
	 sprintf(buf, "%s %s (%c%c) leaving the game (%.16s@%.16s)",
		 ranks[me->p_stats.st_rank].name,
		 me->p_name,
		 me->p_mapchars[0], me->p_mapchars[1],
		 me->p_login,
		 me->p_monitor);
#endif
	 sprintf(addrbuf, " %c%c->ALL",
		 teamlet[me->p_team], shipnos[me->p_no]);
	 pmessage2(buf, 0, MALL | MLEAVE, addrbuf, me->p_no);
      } else {
#ifdef nodef
	 sprintf(buf, "[slot %d free after client/net failure]",
		 ((int) me - (int) &players[0]) / sizeof(struct player));
	 pmessage2(buf, 0, MALL | MLEAVE, "Server:", 255);
#endif
      }
   }
   if (me) {
      if (me->p_process != getpid()) {
	 fprintf(stderr, "process exiting from co-pilot!\n");
	 fprintf(stderr, "%d != pid: %d\n", me->p_process, getpid());
	 fflush(stderr);
	 exit(0);
      }
      me->p_status = PFREE;
      me->p_ghostbuster = 0;
#ifdef nodef
      int             team = me->p_team;
      bzero(me, sizeof(struct player));
      /* so the message stays the right color */
      me->p_team = team;
#endif

   }
   exit(0);
}

void
printUsage(prog)
    char           *prog;
{
   printf("Usage: %s [-u] -s socket# machine\n",
	  prog);
   fflush(stdout);
}

void
sendMotd()
{
   time_t          curtime;
   struct tm      *tmstruct;
   int             hour, tacc;
   FILE           *motd;
   char            buf[100];	/* big enough... */

   time(&curtime);
   tmstruct = localtime(&curtime);
   if (!(hour = tmstruct->tm_hour % 12))
      hour = 12;
   sprintf(buf, "Connection to server established at local time %d:%02d%s.",
	   hour, tmstruct->tm_min,
	   tmstruct->tm_hour >= 12 ? "pm" : "am");
   sendMotdLine(buf);
   sendMotdLine(" ");

   if (gamedown) {
      FILE           *gd = fopen(GAMEDOWN_F, "r");
      if (!gd) {
	 perror(GAMEDOWN_F);
      } else {
	 while (fgets(buf, sizeof(buf) - 2, gd)) {
	    buf[strlen(buf) - 1] = '\0';
	    sendMotdLine(buf);
	 }
      }
      return;
   }
   /* the following will read a motd */
   if ((motd = fopen(MOTD, "r")) != NULL) {
      while (fgets(buf, sizeof(buf), motd) != NULL) {
	 buf[strlen(buf) - 1] = '\0';
	 sendMotdLine(buf);
      }
      (void) fclose(motd);
   }
   sendConfigMsg();
}

void
sendConfigMsg()
{
   register        i, flag;
   char            buf[100];	/* big enough... */

   /* ATM: send current configuration info as well */
   sendMotdLine(STATUS_TOKEN);	/* indicate start of config info */

   sprintf(buf, "calvin server version %s (patchlevel %d)", version,
	   PATCHLEVEL);
   sendMotdLine(buf);

   if (tournplayers > 8)
      sprintf(buf, "%-30s: disabled", "Tournament Mode");
   else
      sprintf(buf, "%-30s: %d players / side", "Tournament Mode", tournplayers);
   sendMotdLine(buf);

   sprintf(buf, "%-30s: ", "Ships Allowed");
   for (i = 0; i < NUM_TYPES; i++) {
      if (i == ATT)
	 continue;
      if (shipsallowed[i]) {
	 strcat(buf, shiptypes[i]);
	 strcat(buf, " ");
      }
   }
   sendMotdLine(buf);

#ifdef nodef
   sprintf(buf, "%-30s: %s", "Tractor/Pressor Beams",
	   weaponsallowed[WP_TRACTOR] ? "enabled" : "disabled");
   sendMotdLine(buf);

   sprintf(buf, "%-30s: %s", "Plasma Torpedoes",
	   weaponsallowed[WP_PLASMA] ? "enabled" : "disabled");
   sendMotdLine(buf);
#endif

   if (weaponsallowed[WP_PLASMA]) {
      sprintf(buf, "%-30s: %d", "Kills Required for Plasma", plkills);
      sendMotdLine(buf);
   }
#ifdef nodef
   sprintf(buf, "%-30s: %s", "Scanning Beams",
	   weaponsallowed[WP_SCANNER] ? "enabled" : "disabled");
   sendMotdLine(buf);
#endif

   if (shipsallowed[STARBASE]) {
      sprintf(buf, "%-30s: %s (%d)", "Rank Required for SB",
	      ranks[sbrank].name, sbrank);
      sendMotdLine(buf);
   }
   sprintf(buf, "%-30s: ", "Hidden Mode");
   switch (hiddenenemy) {
   case 0:
      strcat(buf, "inactive");
      break;
   case 1:
      strcat(buf, "tournament only");
      break;
   case 2:
      strcat(buf, "active");
      break;
   }
   sendMotdLine(buf);

   flag = 0;			/* see if non-home worlds are allowable entry
				 * points */
   for (i = 0; i < MAXPLANETS; i++) {
      if (startplanets[i] && (i % 10)) {
	 flag = 1;
	 break;
      }
   }
   sprintf(buf, "%-30s: %s", "Multiple Entry Planets", flag ? "yes" : "no");
   sendMotdLine(buf);

   sprintf(buf, "%-30s: %s", "Chaos Mode", chaos ? "enabled" : "disabled");
   sendMotdLine(buf);

   sprintf(buf, "%-30s: %s", "Vector Torps", vectortorps ? "yes" :
	   "no");
   sendMotdLine(buf);

   sprintf(buf, "%-30s: ", "Binary Verification");
   switch (binconfirm) {
   case 0:
      strcat(buf, "disabled");
      break;
   case 1:
      strcat(buf, "RSA only");
      break;
   case 2:
      strcat(buf, "enabled, RSA or non-RSA");
      break;
   case 3:
      strcat(buf, "enabled, cyborgs allowed");
      break;
   }
   sendMotdLine(buf);

   sprintf(buf, "%-30s: ", "Message to GOD Log");
   if (loggod)
      strcat(buf, "Yes");
   else
      strcat(buf, "No");
   sendMotdLine(buf);

#ifdef SB_TRANSWARP
   sprintf(buf, "%-30s: ", "SB Transwarp");
   if (twarpMode)
      strcat(buf, "Yes");
   else
      strcat(buf, "No");
   sendMotdLine(buf);
#endif
}


void
reaper(sig)
    int             sig;
{
   while (wait3((union wait *) 0, WNOHANG, (struct rusage *) 0) > 0);
}

void
printStats()
{
   FILE           *logfile;
   int             curtime;

   logfile = fopen(LOGFILENAME, "a");
   if (!logfile)
      return;
   curtime = time(NULL);
   fprintf(logfile, "Leaving: %-16s (%c%c) %3dP %3dA %3dW/%3dL %3dmin %2d <%s@%s> %s",
	   me->p_name,
	   teamlet[me->p_team],
	   shipnos[me->p_no],
	   me->p_stats.st_tplanets - startTplanets,
	   me->p_stats.st_tarmsbomb - startTarms,
	   me->p_stats.st_tkills - startTkills,
	   me->p_stats.st_tlosses - startTlosses,
	   (me->p_stats.st_tticks - startTticks) / 600,
	   numPlanets(me->p_team),
	   me->p_login,
	   me->p_monitor,
	   ctime(&curtime));
#ifdef BD
   {
      int             count, val, min;

      min = (me->p_stats.st_tticks - startTticks) / 600;
      val = bd_borg(&count);
      if (min > 15 && (double) count / (double) min > 0.33) {
	 fprintf(logfile, "BD %s@%s ** %d (%d) ** in %dmin (\"%s\")\n",
		 me->p_login, me->p_monitor, val, count, min,
		 RSA_client_type);
      }
   }
#endif

#ifdef PACKSTATS
   printPackStats(logfile);	/* socket.c */
#endif
   fclose(logfile);
}

int
checkspecial(file, login, host)
    char           *file;
    char           *login, *host;
{
   register        i;
   FILE           *fi;
   char            log_buf[64], host_buf[64], line_buf[160];
   char           *position;
   int             num1;
   int             Hits = 0;	/* Hits==2 means we're banned */
   char            login_b[64], host_b[64];

   for (i = 0; i < strlen(login); i++)
      login_b[i] = isupper(login[i]) ? tolower(login[i]) : login[i];
   login_b[i] = 0;
   for (i = 0; i < strlen(host); i++)
      host_b[i] = isupper(host[i]) ? tolower(host[i]) : host[i];
   host_b[i] = 0;

   login = login_b;
   host = host_b;

   if ((fi = fopen(file, "r")) == NULL) {
      perror(file);
      return (FALSE);
   }
   while (fgets(line_buf, 160, fi) != NULL) {
      /* Split line up */
      if ((*line_buf == '#') || (*line_buf == '\n'))
	 continue;
      if ((position = strrchr(line_buf, '@')) == 0) {
	 fprintf(stderr, "Bad line in %s\n", file);
	 fflush(stderr);
	 continue;
      }
      for (i = 0; i < strlen(line_buf); i++)
	 if (isupper(line_buf[i]))
	    line_buf[i] = tolower(line_buf[i]);

      num1 = position - line_buf;
      strncpy(log_buf, line_buf, num1);	/* copy login name into log_buf */
      log_buf[num1] = '\0';
      strncpy(host_buf, position + 1, 64);	/* copy host name into
						 * host_buf */
      /* Cut off any extra spaces on the host buffer */
      position = host_buf;
      while (!isspace(*position))
	 position++;
      *position = '\0';

      if (*log_buf == '*')
	 Hits = 1;
      else if (!strcmp(login, log_buf))
	 Hits = 1;
      if (Hits == 1) {
	 if (*host_buf == '*') {/* Lock out any host */
	    Hits++;
	    break;		/* break out now. otherwise Hits will get
				 * reset to one */
	 } else if (strstr(host, host_buf) != NULL) {	/* Lock out subdomains
							 * (eg, "*@usc.edu" */
	    Hits++;
	    break;		/* break out now. otherwise Hits will get
				 * reset to one */
	 } else if (!strcmp(host, host_buf)) {	/* Lock out specific host */
	    Hits++;
	    break;		/* break out now. otherwise Hits will get
				 * reset to one */
	 }
      }
   }
   fclose(fi);
   if (Hits >= 2)
      return (TRUE);
   else
      return (FALSE);
}

int
checkslots(login, host)
    char           *login, *host;
{
   register        i;
   register struct player *j;

   for (i = 0, j = players; i < MAXPLAYER; i++, j++) {
      if (j->p_status == PFREE)
	 continue;
      if (j == me)
	 continue;

      if (strcmp(login, j->p_login) == 0 &&
	  samenet(host, j->p_monitor))
	 return 1;
   }
   return 0;
}

/* simple string check of the last two dot-seperated fields */
int
samenet(host1, host2)
    char           *host1, *host2;
{
   char            hbuf1[64], hbuf2[64];

   r2dot(host1, hbuf1);
   r2dot(host2, hbuf2);

   return strcasecmp(hbuf1, hbuf2) == 0;
}

void
r2dot(h, buf)
    char           *h;
    char           *buf;
{
   int             l = strlen(h);
   register int    dot = 0;

   while (l) {
      *buf = h[--l];
      if (*buf == '.')
	 if (++dot == 2)
	    break;
      buf++;
   }
   *buf = 0;
}

/*
 * .pics file format:
 * 
 * name 
 * index line x y color
 * name 
 * index line x y color
 * ...
 */

void
doMotdPics()
{
   FILE           *ptr, *ftemp;
   char            buf[128], fname[128];
   char  	  *bits;
   char           *result;
   int             index, x, y, line, w, h;
   int             color;
   register	   i = -1;

   ptr = fopen(MOTD_BITMAPS, "r");
   if (ptr == NULL)
      return;
   while (1) {
      result = fgets(fname, 125, ptr);
      if (result == NULL){
	 /* must fclose ptr */
	 break;
      }
      if(result[0] == '#' || result[0] == '\n') continue;

      i++;

      fgets(buf, 125, ptr);

      if (sscanf(buf, "%d %d %d %d %d", &index, &line, &x, &y, &color) !=
	  5){
	 fprintf(stderr, "Format error in .pics file\n");
	 fflush(stderr);
	 continue;
      }

      if (fname[strlen(fname) - 1] == '\n')
	 fname[strlen(fname) - 1] = 0;

      ftemp = fopen(fname, "r");
      bits = NULL;
      if (!ftemp) {
	 fprintf(stderr, "ntserv: couldn't open file %s ...skipping\n", fname);
	 fflush(stderr);
	 continue;
      } else {
	 if(ParseXbmFile(ftemp, &w, &h, &bits) < 0){
	    fprintf(stderr, "%s: bad xbm file format\n", fname);
	    fflush(stderr);
	    fclose(ftemp);
	    continue;
	 }
      }
      fclose(ftemp);

      if(bits){
	 if(index < i){
	    i--;	/* duplicate */
	    sendMotdPic(index, x, y, NULL, line, color, w, h);
	 }
	 else{
	    sendMotdPic(index, x, y, bits, line, color, w, h);
	 }
	 free((void *)bits);
      }
   }
   fclose(ptr);
}
