/*
 * orgmove.c
 */
#include "copyright.h"

#include <X11/Xlib.h>
#include <stdio.h>
#include <signal.h>
#include <values.h>
#include <math.h>
#include <time.h>
#include "defs.h"
#include "struct.h"
#include "data.h"

#include "planets.h"

#ifdef USE_SYS_PROTOS
#if defined(__STDC__) || defined(__cplusplus)
#define P_(s) s
#else
#define P_(s) ()
#endif

extern int                    printf P_((const char *, ...));
extern long                   random P_((void));
extern int                    fprintf P_((FILE *, const char *, ...));
extern int                    fflush P_((FILE *));
extern void                   bzero P_((void *, size_t));

#undef P_
#endif

#define STARTPAUSE	50

#define TPLANETSPEED	4
#define TPLANETDELAY	20
#define TFINALSPEED	3
#define TFINALDELAY	50

#define CENTERFIX	40

#define DECELDIST	18000

extern struct player *perfs[NUMTEAM];
extern int      debug;
extern int      lastm;
extern int      updates;
extern int      org_xloc[9], org_yloc[9];

extern int      locx, locy;

static int      o_clock;
static void	_move();
static int      done[MAXPLAYER];
static int      planetsreset[MAXPLANETS];

unsigned char   getacourse();

void
orgmove()
{
   o_clock++;

   _move();
}

static void
_move()
{
   register        i;
   static int      dntime;

   for (i = 0; i < NUMTEAM; i++) {
      if (!perfs[i])
	 exitOrg(-1);
      if (perfs[i]->p_status == PFREE)
	 exitOrg(-2);
      if (perfs[i]->p_status == PDEAD)
	 exitOrg(-3);
      perfs[i]->p_ghostbuster = 0;

      if (o_clock < STARTPAUSE / updates)
	 continue;

      if (!dntime)
	 moveorg(perfs[i]);
   }

   if (!dntime) {
      dntime = TFINALDELAY / updates;

      for (i = 0; i < NUMTEAM; i++) {
	 if (!done[perfs[i]->p_no])
	    dntime = 0;
      }
   } else {
      if (dntime == TFINALDELAY / updates) {
	 for (i = 0; i < NUMTEAM; i++)
	    perfs[i]->p_flags &= ~PFSHIELD;
	 finaltorps(TMOVE);
      }
      if (dntime == 15 / updates) {
	 leavemesg();
      }
      if (dntime == 10 / updates - 2)
	 finaltorps(TEXPLODE);
      dntime--;
      if (dntime == 1) {
	 signal(SIGALRM, SIG_IGN);
	 exitOrg(-4);
      }
   }
}

void
moveorg(j)
   struct player  *j;
{
   unsigned char   crs;
   int             dist;
   struct planet  *k;

   if (getplanet(j, &crs, &dist, &k)) {
      j->p_desdir = crs;
      j->p_desspeed = 14;
      if (dist < 5000) {
	 j->p_desspeed = 2;
	 resetplanet(j, k, crs);
      } else if (dist < 10000) {
	 posplasma(j, k->pl_x, k->pl_y);
      }
      if (debug) {
	 printf("%s found planet %s\n", j->p_name,
		k->pl_name);
      }
   } else {			/* return home */
      unsigned char   prevdir = j->p_desdir;
      j->p_desdir = getacourse(org_xloc[j->p_team], org_yloc[j->p_team],
			       j->p_x, j->p_y);

      if (angdist(prevdir, j->p_desdir) > 64)
	 j->p_speed = 0;

      if (ihypot(j->p_x - org_xloc[j->p_team], j->p_y - org_yloc[j->p_team])
	  < DECELDIST) {
	 j->p_desspeed = 0;
	 if (j->p_speed == 0) {
	    done[j->p_no] = 1;
	 }
      } else {
	 j->p_desspeed = 18;
      }
   }
}

void
resetplanet(j, k, dir)
   struct player  *j;
   struct planet  *k;
   unsigned char   dir;
{
   if (debug) {
      printf("%s resetting planet %s\n", j->p_name,
	     k->pl_name);
   }
   firephaser(j, k->pl_x, k->pl_y, dir);
   explodepltorp(j, k->pl_x, k->pl_y);
   fireshrapnel(j, k->pl_x, k->pl_y);

   k->pl_armies = 30;
   k->pl_flags = pdata[k->pl_no].pl_flags;
   k->pl_info = (FED | ROM | KLI | ORI);	/* make seen by all */
   k->pl_owner = pdata[k->pl_no].pl_owner;
   k->pl_x = pdata[k->pl_no].pl_x;
   k->pl_y = pdata[k->pl_no].pl_y;

   if (k->pl_flags & PLHOME)
      k->pl_flags |= (PLREPAIR | PLFUEL);

   if (random() % 6 == 0)
      k->pl_flags |= PLAGRI;
   if (random() % 4 == 0)
      k->pl_flags |= PLREPAIR;
   if (random() % 2 == 0)
      k->pl_flags |= PLFUEL;

   planetsreset[k->pl_no] = 1;
}

void
firephaser(j, x, y, dir)
   struct player  *j;
   int             x, y;
   unsigned char   dir;
{
   struct phaser  *mine;

   mine = &phasers[j->p_no];
   mine->ph_dir = dir;
   mine->ph_fuse = 5;
   mine->ph_status = PHHIT2;
   mine->ph_x = x - CENTERFIX;
   mine->ph_y = y - CENTERFIX;
}

void
posplasma(j, x, y)
   struct player  *j;
   int             x, y;
{
   struct plasmatorp *t;

   t = &plasmatorps[j->p_no];

   t->pt_x = x - CENTERFIX;
   t->pt_y = y - CENTERFIX;
   t->pt_status = PTMOVE;
   t->pt_turns = 0;
   t->pt_fuse = 50;
   t->pt_speed = 0;

   j->p_nplasmatorp++;
}

void
explodepltorp(j, x, y)
   struct player  *j;
   int             x, y;
{
   struct plasmatorp *t;

   t = &plasmatorps[j->p_no];

   t->pt_x = x - CENTERFIX;
   t->pt_y = y - CENTERFIX;
   t->pt_status = PTEXPLODE;
   t->pt_fuse = 10;
   t->pt_war = 0;
}

void
fireshrapnel(j, x, y)
   struct player  *j;
   int             x, y;
{
   register        i;
   register struct torp *k;

   for (i = j->p_no * MAXTORP, k = &torps[i]; i < j->p_no * MAXTORP + MAXTORP;
	i++, k++) {

      k->t_no = i;
      k->t_status = TMOVE;
      k->t_owner = j->p_no;
      k->t_team = j->p_team;
      k->t_x = x - CENTERFIX;
      k->t_y = y - CENTERFIX;
      k->t_dir = i * 32;
      k->t_speed = TPLANETSPEED;
      k->t_damage = 0;
      k->t_fuse = TPLANETDELAY;
      k->t_war = 0;
   }
   j->p_ntorp = 8;
}

void
finaltorps(s)
   int             s;
{
   register        i, l;
   register struct player *j;
   register struct torp *k;

   for (i = 0; i < NUMTEAM; i++) {
      j = perfs[i];

      for (l = j->p_no * MAXTORP, k = &torps[l]; l < j->p_no * MAXTORP + MAXTORP;
	   l++, k++) {

	 if (s == TMOVE) {
	    k->t_no = i;
	    k->t_status = TMOVE;
	    k->t_owner = j->p_no;
	    k->t_team = j->p_team;
	    k->t_x = locx - CENTERFIX;
	    k->t_y = locy - CENTERFIX;
	    k->t_dir = (l - j->p_no * MAXTORP) * 32 + i * 8;
	    if (debug)
	       printf("%d\n", k->t_dir);
	    k->t_speed = TFINALSPEED;
	    k->t_damage = 0;
	    k->t_fuse = TFINALDELAY;
	    k->t_war = 0;
	 } else {
	    k->t_status = TEXPLODE;
	    k->t_fuse = 10;
	 }
      }
      if (s == TMOVE)
	 j->p_ntorp = 8;
   }
}

static int      tnos[9] = {0, /* fed */ 0, /* rom */ 1,
   0, /* kli */ 2, 0, 0, 0,
 /* ori */ 3};

int
getplanet(o, crs, dist, pl)
   struct player  *o;
   unsigned char  *crs;
   int            *dist;
   struct planet **pl;
{
   register        i;
   int             tno, starti, stopi;
   register struct planet *j, *f = NULL;
   int             mindist = 200000, dst;

   tno = tnos[o->p_team];
   starti = tno * 10;
   stopi = starti + 10;
   for (i = starti; i < stopi; i++) {
      j = &planets[i];
      if (!planetsreset[j->pl_no] || (j->pl_owner != o->p_team)) {
	 dst = ihypot(o->p_x - j->pl_x, o->p_y - j->pl_y);
	 if (dst < mindist) {
	    mindist = dst;
	    f = j;
	 }
      }
   }
   if (f) {
      *crs = getacourse(f->pl_x, f->pl_y, o->p_x, o->p_y);
      *dist = mindist;
      *pl = f;
      return 1;
   }
   return 0;
}

void
leavemesg()
{
   register        i;
   char            buf[80];
   char            addrbuf[20];

   for (i = 0; i < NUMTEAM; i++) {
      sprintf(buf, "%s (%s) leaving the dimension (%s)",
	      perfs[i]->p_name,
	      perfs[i]->p_mapchars,
	      perfs[i]->p_login);
      sprintf(addrbuf, " %c%c->ALL",
	      teamlet[perfs[i]->p_team], shipnos[perfs[i]->p_no]);
      pmessage2(buf, 0, MALL, addrbuf, perfs[i]->p_no);
   }
}

void
exitOrg(sig)
   int             sig;
{
   register int    i, l;
   register struct player *j;
   register struct torp *k;
   register struct plasmatorp *f;

   signal(SIGALRM, SIG_IGN);

   fprintf(stderr, "organians exiting on signal %d\n", sig);
   fflush(stderr);

   for (i = 0; i < NUMTEAM; i++) {
      j = perfs[i];

      for (l = j->p_no * MAXTORP, k = &torps[l];
	   l < j->p_no * MAXTORP + MAXTORP; l++, k++) {
	 /* torps are free here */
	 if (k->t_status != TFREE)
	    k->t_status = TOFF;
      }

      /* plasma */
      f = &plasmatorps[j->p_no * MAXPLASMA];
      f->pt_status = PTFREE;

      bzero(&phasers[j->p_no], sizeof(struct phaser));
      bzero(j, sizeof(struct player));
      j->p_stats.st_tticks = 1;
   }

   /*
    * note: if we had a segmentation violation or bus error we want a core
    * file to debug
    */
   if (sig == SIGBUS || sig == SIGSEGV)
      abort();
   else
      exit(0);
}

void
pmessage(str, recip, group, address)
   char           *str;
   int             recip;
   int             group;
   char           *address;
{
   pmessage2(str, recip, group, address, 255);
}

void
pmessage2(str, recip, group, address, from)
   char           *str;
   int             recip;
   int             group;
   char           *address;
   unsigned char   from;
{
   struct message *cur;
   int             mesgnum;

   if ((mesgnum = ++(mctl->mc_current)) >= MAXMESSAGE) {
      mctl->mc_current = 0;
      mesgnum = 0;
   }
   cur = &messages[mesgnum];
   cur->m_no = mesgnum;
   cur->m_flags = group;
   cur->m_recpt = recip;
   cur->m_from = from;
   (void) sprintf(cur->m_data, "%-9s %s", address, str);
   cur->m_flags |= MVALID;
#ifdef SHORT_PACKETS
   cur->args[0] = DINVALID; /* to prevent false sending with SPS_S_WARNING*/
#endif
}

/* get course from (mx,my) to (x,y) */
unsigned char 
getacourse(x, y, mx, my)
   int             x, y, mx, my;
{
   unsigned char d;
   if (x == mx && y == my)
      return 0;

   d = (unsigned char) (atan2((double) (x - mx),
				 (double) (my - y)) / 3.14159 * 128.);
   MATH_ERR_CHECK(x-mx + my-y);
   return d;
}

/*
 * gives approximate distance from (x1,y1) to (x2,y2) with only
 * overestimations, and then never by more than (9/8) + one bit uncertainty.
 */

int 
ihypot(x1, y1)
   int             x1, y1;
{
   register        x2 = 0, y2 = 0;

   if ((x2 -= x1) < 0)
      x2 = -x2;
   if ((y2 -= y1) < 0)
      y2 = -y2;
   return (x2 + y2 - (((x2 > y2) ? y2 : x2) >> 1));
}
