/*
 * robotII.c
 */
#include "copyright.h"
#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <signal.h>
#include <setjmp.h>
#include "defs.h"
#include "struct.h"
#include "data.h"
#ifndef _NO_DEPEND
#   include "robotII_protos.h"
#endif

#ifdef USE_SYS_PROTOS
#if defined(__STDC__) || defined(__cplusplus)
#define P_(s) s
#else
#define P_(s) ()
#endif

static void exitRobot P_((int s));

extern int                    atoi P_((const char *));
extern int                    fprintf P_((FILE *, const char *, ...));
extern void                   srandom P_((int));
extern pid_t                  getpid P_((void));
extern time_t                 time P_((time_t *));
extern long                   random P_((void));
extern char *                 strncpy P_((char *, const char *, size_t));
extern int                    setitimer P_((int, struct itimerval *, struct itimerval *));
extern void                   perror P_((const char *));
extern int                    sigsetmask P_((int));
extern int                    pause P_((void));
extern void                   bzero P_((void *, size_t));

#undef P_
#else
static void exitRobot();
#endif

struct itimerval udt;
int             redrawall;
int             lastm;

/* lots of neat flags */
int             hostile;
int             debug;
int             level;
int             fleet;
int             sticky;
int             berserk;
int             practice;
int             nofuel;

char           *rnames[4] = {"M5", "Colossus", "Guardian", "HAL"};

void
main(argc, argv)
   int             argc;
   char          **argv;
{
   register int    i;
   int             team = -1;
   int             bteam=0;
   int             pno;
#ifdef PRACSTARBASE
   int             sb = 0;
#endif

   for (; argc > 1 && argv[1][0] == '-'; argc--, argv++) {
      switch (argv[1][1]) {
      case 'F':
	 nofuel++;
	 break;
      case 'f':
	 fleet++;
	 break;
      case 's':
	 sticky++;
	 break;
      case 'd':
	 debug++;
	 break;
      case 'h':
	 hostile++;
	 break;
      case 'p':
	 practice++;
	 break;
      case 'b':
	 berserk++;
	 break;
#ifdef PRACSTARBASE
      case 'B':
	 sb++;
	 break;
#endif
      case 'l':
	 if (argv[1][2] != NULL)
	    level = atoi(&argv[1][2]);
	 else
	    level = 100;
	 break;
      case 'T':
	 switch (argv[1][2]) {
	 case 'f':
	    team = 0;
	    bteam = FED;
	    break;
	 case 'r':
	    team = 1;
	    bteam = ROM;
	    break;
	 case 'k':
	    team = 2;
	    bteam = KLI;
	    break;
	 case 'o':
	    team = 3;
	    bteam = ORI;
	    break;
	 default:
	    fprintf(stderr, "Unknown team type.  Usage -Tx where x is [frko]\n");
	    exit(1);
	 }
	 break;
      default:
	 fprintf(stderr, "Unknown option '%c'\n", argv[1][1]);
	 exit(1);
      }


   }
   srandom(getpid() * time(NULL));
   readsysdefaults();
   if (team < 0 || team >= 4) {
      if (debug)
	 fprintf(stderr, "Choosing random team.\n");
      team = random() % 4;
   }
   openmem();
   pno = findrslot();		/* Robots are entitled to tester's slots. */
   me = &players[pno];
   me->p_no = pno;
   myship = &me->p_ship;
   mystats = &me->p_stats;
   lastm = mctl->mc_current;
   /*
    * At this point we have memory set up.  If we aren't a fleet, we don't
    * want to replace any other robots on this team, so we'll check the other
    * players and get out if there are any on our team.
    */

   if (!fleet) {
      for (i = 0; i < MAXPLAYER; i++) {
	 if ((players[i].p_status == PALIVE) && (players[i].p_team == bteam)) {
	    if (debug)
	       fprintf(stderr, "Galaxy already defended\n");
	    players[pno].p_status = PFREE;
	    exit(1);
	 }
      }
   }
   if (practice) {
      strcpy(pseudo, "Hoser");
   } else {
      strcpy(pseudo, rnames[team]);
   }
   strcpy(login, "Robot");

   (void) strncpy(me->p_login, login, sizeof(me->p_login));
   me->p_login[sizeof(me->p_login) - 1] = '\0';
   (void) strncpy(me->p_monitor, "Nowhere", sizeof(me->p_monitor));
   me->p_monitor[sizeof(me->p_monitor) - 1] = '\0';
#ifdef PRACSTARBASE
   enter(team, "Nowhere", pno, sb ? STARBASE : BATTLESHIP);
#else
   enter(team, "Nowhere", pno, BATTLESHIP);
#endif
   me->p_pos = -1;		/* So robot stats don't get saved */
   me->p_flags |= PFROBOT;	/* Mark as a robot */
   if (berserk)
      me->p_hostile = (FED | ROM | KLI | ORI);	/* unless they are berserk */
   else if (practice)
      me->p_hostile = bteam;	/* or practice */
   else
      me->p_hostile = 0;	/* robots are peaceful */

   if (practice)
      me->p_flags |= PFPRACTR;	/* Mark as a practice robot */
   signal(SIGTERM, exitRobot);
   signal(SIGALRM, rmove);
   if (!nofuel) {
      myship->s_phasercost = 0;
      myship->s_torpcost = 0;
      myship->s_cloakcost = 0;
      myship->s_plasmacost = 0;
   }
#ifndef DS
   if (practice) {
      udt.it_interval.tv_sec = 1;	/* Robots get to move 1/sec */
      udt.it_interval.tv_usec = 000000;
   } else {
      udt.it_interval.tv_sec = 0;	/* Robots get to move 2/sec */
      udt.it_interval.tv_usec = 500000;
   }
   udt.it_value.tv_sec = 1;
   udt.it_value.tv_usec = 0;
   if (setitimer(ITIMER_REAL, &udt, 0) < 0) {
      perror("setitimer");
      me->p_status = PFREE;	/* Put robot in game */
      exit(1);
   }
#else
   me->p_process = getpid();
   me->p_timerdelay = practice?10:5;
#endif

   /* allows robots to be forked by the daemon -- Evil ultrix bullshit */
   sigsetmask(0);

   me->p_status = PALIVE;	/* Put robot in game */
   while (1) {
      pause();
   }
}

int
findtestslot()
{
   register int    i;

   for (i = MAXPLAYER - TESTERS; i < MAXPLAYER; i++) {
      if (players[i].p_status == PFREE) {	/* We have a free slot */
	 players[i].p_status = POUTFIT;	/* possible race code */
	 break;
      }
   }
   if (i == MAXPLAYER) {
      return -1;		/* no room in tester slots */
   }
   bzero(&players[i].p_stats, sizeof(struct stats));
   players[i].p_stats.st_tticks = 1;
   return (i);
}

int
findrslot()
{
   register int    i;

   /* look for tester slot first */
   i = findtestslot();
   if (i > -1)
      return i;

   for (i = 0; i < MAXPLAYER; i++) {
      if (players[i].p_status == PFREE) {	/* We have a free slot */
	 players[i].p_status = POUTFIT;	/* possible race code */
	 break;
      }
   }
   if (i == MAXPLAYER) {
      fprintf(stderr, "No more room in game\n");
      exit(0);
   }
   bzero(&players[i].p_stats, sizeof(struct stats));
   players[i].p_stats.st_tticks = 1;
   return (i);
}

void
warning(f, mess)
   int	f;
   char	*mess;
{
   if (debug)
      fprintf(stderr, "warning: %s\n", mess);
}

/* do nothing */
void
swarning(whichmessage, argument, argument2)
   unsigned char                   whichmessage, argument,
                                   argument2;
{
}

static void
exitRobot(s)

   int	s;
{
   signal(SIGALRM, SIG_IGN);
   if(me)
      me->p_status = PFREE;
   exit(0);
}

void
handleUpdatesReq(arg)

   int	*arg;
{
   /* XX */
}

#ifdef BD
void
bd_phaser(dir)

   unsigned char	dir;
{
}
#endif
