/*
 * Socket.c
 *
 * Kevin P. Smith  1/26/89 UDP stuff v1.0 by Andy McFadden  Feb-Apr 1992
 *
 * UDP protocol v1.0
 *
 * Provides all of the support for sending packets to the client.
 */
#include "copyright2.h"
#include <stdio.h>
#include <signal.h>
#include <string.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <netdb.h>
#include <math.h>
#include <errno.h>
#include "defs.h"
#include "struct.h"
#include "data.h"
#include "packets.h"
#ifndef _NO_DEPEND
#   include "ntserv_protos.h"
#endif

/* this stuff is used for Fat UDP */
typedef void                   *PTR;	/* adjust this if you lack (void *) */
typedef struct fat_node_t {
   PTR                             packet;
   int                             pkt_size;
   struct fat_node_t              *prev;
   struct fat_node_t              *next;
}                               FAT_NODE;
/* needed for fast lookup of semi-critical fat nodes */

#if defined(__STDC__) || defined(__cplusplus)
#define P_(s) s
#else
#define P_(s) ()
#endif

/* socket.c */
static void updateFat P_((struct player_spacket *packet));
static int fatten P_((void));
static void dequeue P_((FAT_NODE *fatp));
static void enqueue P_((FAT_NODE *fatp, int list));
static void fatMerge P_((void));

#ifdef USE_SYS_PROTOS

extern void exit                P_((int));
extern int printf               P_((const char *,...));
extern long unsigned int inet_addr P_(( /* ??? */ ));
extern int close                P_((int));
extern int shutdown             P_((int s, int how));
extern int socket               P_((int domain, int type, int protocol));
extern int connect              P_((int s, struct sockaddr * name, int namelen));
extern int getpeername          P_((int s, struct sockaddr * name, int *namelen));
extern int fflush               P_((FILE *));
extern int strncmp              P_((const char *, const char *, size_t));
extern char                    *strncpy P_((char *, const char *, size_t));
extern long random              P_((void));
extern int fprintf              P_((FILE *, const char *,...));
extern void bcopy               P_((const void *, void *, int));
extern void bzero               P_((void *, size_t));
extern int select               P_((int width, fd_set * readfds, fd_set * writefds, fd_set * exceptfds, struct timeval * timeout));
extern int read                 P_((int, void *, unsigned int));
extern void perror              P_((const char *));
extern int fclose               P_((FILE *));
extern char                    *strcat P_((char *, const char *));
extern time_t time              P_((time_t *));
extern int bcmp                 P_((const void *, const void *, int));
extern void                    *malloc P_((size_t));
extern int getsockname          P_((int s, struct sockaddr * name, int *namelen));
extern int write                P_((int, const void *, unsigned int));
extern int setitimer            P_((int, struct itimerval *, struct itimerval *));
extern int strncasecmp          P_((char *s1, char *s2, int n));
extern size_t strftime          P_((char *, size_t, const char *, const struct tm *));
extern int flock                P_((int, int));
extern int gethostname          P_((char *, int));
extern char                    *inet_ntoa P_(( /* ??? */ ));
extern int setsockopt           P_((int s, int level, int optname, char *optval, int optlen));
extern char *                 strstr P_((const char *, const char *));
extern pid_t                  fork P_((void));
extern int                    execl P_((const char *, const char *, ...));
extern pid_t                  getpid P_((void));
#else
extern void exit();
#endif

#undef P_

#define CVOID		      struct player_spacket *

/* #define INCLUDE_SCAN		** include Amdahl scanning beams */
/* #define INCLUDE_VISTRACT	** include visible tractor beams */

#define BROKEN			/* for local stress testing; drops 20% of
				]				   packets */
int                             broke = 0;
/* #define HOSED		** combine with BROKEN; drops 90% of packets */
/* #define FATDIAG	** define to get LOTS of fat UDP debugging messages */

static int                      remoteaddr = -1;	/* inet address in net
							 * format */
extern void handleOggV();

struct packet_handler           handlers[] =
{
   {0, NULL},			/* record 0 */
   {sizeof(struct mesg_cpacket), handleMessageReq},	/* CP_MESSAGE */
   {sizeof(struct speed_cpacket), handleSpeedReq},	/* CP_SPEED */
   {sizeof(struct dir_cpacket), handleDirReq},	/* CP_DIRECTION */
   {sizeof(struct phaser_cpacket), handlePhasReq},	/* CP_PHASER */
   {sizeof(struct plasma_cpacket), handlePlasmaReq},	/* CP_PLASMA */
   {sizeof(struct torp_cpacket), handleTorpReq},	/* CP_TORP */
   {sizeof(struct quit_cpacket), handleQuitReq},	/* CP_QUIT */
   {sizeof(struct login_cpacket), handleLoginReq},	/* CP_LOGIN */
   {sizeof(struct outfit_cpacket), handleOutfit},	/* CP_OUTFIT */
   {sizeof(struct war_cpacket), handleWarReq},	/* CP_WAR */
   {sizeof(struct practr_cpacket), handlePractrReq},	/* CP_PRACTR */
   {sizeof(struct shield_cpacket), handleShieldReq},	/* CP_SHIELD */
   {sizeof(struct repair_cpacket), handleRepairReq},	/* CP_REPAIR */
   {sizeof(struct orbit_cpacket), handleOrbitReq},	/* CP_ORBIT */
   {sizeof(struct planlock_cpacket), handlePlanlockReq},	/* CP_PLANLOCK */
   {sizeof(struct playlock_cpacket), handlePlaylockReq},	/* CP_PLAYLOCK */
   {sizeof(struct bomb_cpacket), handleBombReq},	/* CP_BOMB */
   {sizeof(struct beam_cpacket), handleBeamReq},	/* CP_BEAM */
   {sizeof(struct cloak_cpacket), handleCloakReq},	/* CP_CLOAK */
   {sizeof(struct det_torps_cpacket), handleDetTReq},	/* CP_DET_TORPS */
   {sizeof(struct det_mytorp_cpacket), handleDetMReq},	/* CP_DET_MYTORP */
   {sizeof(struct copilot_cpacket), handleCopilotReq},	/* CP_COPLIOT */
   {sizeof(struct refit_cpacket), handleRefitReq},	/* CP_REFIT */
   {sizeof(struct tractor_cpacket), handleTractorReq},	/* CP_TRACTOR */
   {sizeof(struct repress_cpacket), handleRepressReq},	/* CP_REPRESS */
   {sizeof(struct coup_cpacket), handleCoupReq},	/* CP_COUP */
   {sizeof(struct socket_cpacket), handleSocketReq},	/* CP_SOCKET */
   {sizeof(struct options_cpacket), handleOptionsPacket},	/* CP_OPTIONS */
   {sizeof(struct bye_cpacket), handleByeReq},	/* CP_BYE */
   {sizeof(struct dockperm_cpacket), handleDockingReq},	/* CP_DOCKPERM */
   {sizeof(struct updates_cpacket), handleUpdatesReq},	/* CP_UPDATES */
   {sizeof(struct resetstats_cpacket), handleReset},	/* CP_RESETSTATS */
   {sizeof(struct reserved_cpacket), handleReserved},	/* CP_RESERVED */
#ifdef INCLUDE_SCAN
   {sizeof(struct scan_cpacket), handleScan},	/* CP_SCAN (ATM) */
#else
   {0, exit},			/* note exit will never be called */
#endif
   {sizeof(struct udp_req_cpacket), handleUdpReq},	/* CP_UDP_REQ */
   {sizeof(struct sequence_cpacket), handleSequence},	/* CP_SEQUENCE */
#ifdef RSA
   {sizeof(struct rsa_key_cpacket), handleRSAKey},	/* CP_RSA_KEY */
#else
   {0, exit},			/* 37 */
#endif
   {0, exit},			/* 38 */
   {0, exit},			/* 39 */
   {0, exit},			/* 40 */
   {0, exit},			/* 41 */
   {sizeof(struct ping_cpacket), handlePingResponse},	/* CP_PING_RESPONSE */
#ifdef SHORT_PACKETS
   {sizeof(struct shortreq_cpacket), handleShortReq},	/* CP_S_REQ */
   {sizeof(struct threshold_cpacket), handleThresh},	/* CP_S_THRS */
   {-1, handleSMessageReq},	/* CP_S_MESSAGE */
#endif
   {0, exit},			/* 46 */
   {0, exit},			/* 47 */
   {0, exit},			/* 48 */
   {0, exit},			/* 49 */
#ifdef BASEPRACTICE
   {sizeof(struct oggv_cpacket), handleOggV},	/* CP_OGGV */
#else
   {0, exit},			/* 50 */
#endif
   {0, exit},			/* 51 */
   {0, exit},			/* 52 */
   {0, exit},			/* 53 */
   {0, exit},			/* 54 */
   {0, exit},			/* 55 */
   {0, exit},			/* 56 */
   {0, exit},			/* 57 */
   {0, exit},			/* 58 */
   {0, exit},			/* 59 */
#ifdef FEATURE
   {sizeof(struct feature_cpacket), handleFeature },	/* CP_FEATURE */
#endif
};

int                             sizes[] =
{
   0,				/* record 0 */
   sizeof(struct mesg_spacket),	/* SP_MESSAGE */
   sizeof(struct plyr_info_spacket),	/* SP_PLAYER_INFO */
   sizeof(struct kills_spacket),/* SP_KILLS */
   sizeof(struct player_spacket),	/* SP_PLAYER */
   sizeof(struct torp_info_spacket),	/* SP_TORP_INFO */
   sizeof(struct torp_spacket),	/* SP_TORP */
   sizeof(struct phaser_spacket),	/* SP_PHASER */
   sizeof(struct plasma_info_spacket),	/* SP_PLASMA_INFO */
   sizeof(struct plasma_spacket),	/* SP_PLASMA */
   sizeof(struct warning_spacket),	/* SP_WARNING */
   sizeof(struct motd_spacket),	/* SP_MOTD */
   sizeof(struct you_spacket),	/* SP_YOU */
   sizeof(struct queue_spacket),/* SP_QUEUE */
   sizeof(struct status_spacket),	/* SP_STATUS */
   sizeof(struct planet_spacket),	/* SP_PLANET */
   sizeof(struct pickok_spacket),	/* SP_PICKOK */
   sizeof(struct login_spacket),/* SP_LOGIN */
   sizeof(struct flags_spacket),/* SP_FLAGS */
   sizeof(struct mask_spacket),	/* SP_MASK */
   sizeof(struct pstatus_spacket),	/* SP_PSTATUS */
   sizeof(struct badversion_spacket),	/* SP_BADVERSION */
   sizeof(struct hostile_spacket),	/* SP_HOSTILE */
   sizeof(struct stats_spacket),/* SP_STATS */
   sizeof(struct plyr_login_spacket),	/* SP_PL_LOGIN */
   sizeof(struct reserved_spacket),	/* SP_RESERVED */
   sizeof(struct planet_loc_spacket),	/* SP_PLANET_LOC */
#ifdef INCLUDE_SCAN
   sizeof(struct scan_spacket),	/* SP_SCAN (ATM) */
#else
   0,
#endif
   sizeof(struct udp_reply_spacket),	/* SP_UDP_REPLY */
   sizeof(struct sequence_spacket),	/* SP_SEQUENCE */
   sizeof(struct sc_sequence_spacket),	/* SP_SC_SEQUENCE */
#ifdef RSA
   sizeof(struct rsa_key_spacket),	/* SP_RSA_KEY */
#else
   0,				/* 31 */
#endif
   sizeof(struct motd_pic_spacket),	/* SP_MOTD_PIC */
   0,				/* 33 */
   0,				/* 34 */
   0,				/* 35 */
   0,				/* 36 */
   0,				/* 37 */
   0,				/* 38 */
   0,				/* 39 */
#ifdef SHORT_PACKETS
   sizeof(struct shortreply_spacket),	/* SP_S_REPLY */
   -1,				/* SP_S_MESSAGE */
   sizeof(struct warning_s_spacket),	/* SP_S_WARNING */
   sizeof(struct youshort_spacket),	/* SP_S_YOU */
   sizeof(struct youss_spacket),/* SP_S_YOU_SS */
   -1,				/* variable *//* SP_S_PLAYER */
#else
   0,				/* 40 */
   0,				/* 41 */
   0,				/* 42 */
   0,				/* 43 */
   0,				/* 44 */
   0,				/* 45 */
#endif
   sizeof(struct ping_spacket),
#ifdef SHORT_PACKETS
   -1,				/* variable *//* SP_S_TORP  */
   -1,				/* variable *//* SP_S_TORP_INFO */
   20,				/* SP_S_8_TORP */
   -1,				/* variable *//* SP_S_PLANET */
#endif				/* XX - no #else */
   0,				/* 51 */
   0,				/* 52 */
   0,				/* 53 */
   0,				/* 54 */
   0,				/* 55 */
   0,				/* 56 */
   0,				/* 57 */
   0,				/* 58 */
   0,				/* 59 */
#ifdef FEATURE
   sizeof(struct feature_spacket),/* 60 - SP_FEATURE */
#endif
   sizeof(struct bitmap_spacket), /* 61 - SP_BITMAP */
};
#define NUM_PACKETS (sizeof(handlers) / sizeof(handlers[0]) - 1)
#define NUM_SIZES (sizeof(sizes) / sizeof(sizes[0]) - 1)

#ifdef PACKSTATS
static int                      packetsReceived[NUM_PACKETS + 1];
static int                      packetsSent[NUM_SIZES + 1];
static int                      packetsSentBytes[NUM_SIZES + 1];
#endif

/*
 * no need to make this big: 90% packets fall under 1K, those that don't can
 * be fragmented with very small overhead
 */
#define BUFSIZE 1024

#define UDPBUFSIZE 960		/* (tweakable; should be under 1300) */

static char                     buf[BUFSIZE];	/* Socket buffer */
static char                     udpbuf[UDPBUFSIZE];	/* UDP socket buffer */
static char                    *bufptr = buf;
static char                    *udpbufptr = udpbuf;
int                             clientDead = 0;
static long                     sequence;	/* the holy sequence number */

static int                      udpLocalPort = 0;
static int                      udpClientPort = 0;
static int                      udpMode = MODE_SIMPLE;	/* what kind of UDP
							 * trans we want */
#ifdef SAVE_MEM			/* fat UDP is not used all the time.. makes a little sense
										    to dynamically allocate these arrays */

FAT_NODE                       *fat_kills = NULL;
FAT_NODE                       *fat_torp_info = NULL;
FAT_NODE                       *fat_phaser = NULL;
FAT_NODE                       *fat_plasma_info = NULL;
FAT_NODE                        fat_you;
FAT_NODE                        fat_status;
#ifdef SHORT_PACKETS
FAT_NODE	               fat_s_you;
FAT_NODE	               fat_s_planet;
#endif
FAT_NODE                       *fat_planet = NULL;
FAT_NODE                       *fat_flags = NULL;
FAT_NODE                       *fat_hostile = NULL;
#else

FAT_NODE                        fat_kills[MAXPLAYER];
FAT_NODE                        fat_torp_info[MAXPLAYER * MAXTORP];
FAT_NODE                        fat_phaser[MAXPLAYER];
FAT_NODE                        fat_plasma_info[MAXPLAYER * MAXPLASMA];
FAT_NODE                        fat_you;
#ifdef SHORT_PACKETS
FAT_NODE	               fat_s_you;
FAT_NODE	               fat_s_planet;
#endif
FAT_NODE                        fat_status;
FAT_NODE                        fat_planet[MAXPLANETS];
FAT_NODE                        fat_flags[MAXPLAYER];
FAT_NODE                        fat_hostile[MAXPLAYER];
#endif

/* define the lists */
#define MAX_FAT_LIST	5	/* tweakable; should be > 1 */
typedef struct {
   FAT_NODE                       *head;
   FAT_NODE                       *tail;
}                               FAT_LIST;
static
FAT_LIST                        fatlist[MAX_FAT_LIST],
                                tmplist[MAX_FAT_LIST];
/* tweakable parameters; compare with UDPBUFSIZE */
/* NOTE: FAT_THRESH + MAX_FAT_DATA must be < UDPBUFSIZE                */
/* MAX_FAT_DATA must be larger than biggest semi-critical packet */
#ifdef SHORT_PACKETS
#define FAT_THRESH	(send_short?128:400)
#define MAX_FAT_DATA	(send_short?64:100)
#else
#define FAT_THRESH	400
#define MAX_FAT_DATA	100
#endif
#define MAX_NONFAT	10	/* if we have this much left, stop */

struct plyr_info_spacket        clientPlayersInfo[MAXPLAYER];
struct plyr_login_spacket       clientLogin[MAXPLAYER];
struct hostile_spacket          clientHostile[MAXPLAYER];
struct stats_spacket            clientStats[MAXPLAYER];
struct player_spacket           clientPlayers[MAXPLAYER];
struct kills_spacket            clientKills[MAXPLAYER];
struct flags_spacket            clientFlags[MAXPLAYER];
struct pstatus_spacket          clientPStatus[MAXPLAYER];
int                             msgCurrent;
struct torp_info_spacket        clientTorpsInfo[MAXPLAYER * MAXTORP];
struct torp_spacket             clientTorps[MAXPLAYER * MAXTORP];
struct phaser_spacket           clientPhasers[MAXPLAYER];
struct you_spacket              clientSelf;
struct status_spacket           clientStatus;
struct planet_spacket           clientPlanets[MAXPLANETS];
struct planet_loc_spacket       clientPlanetLocs[MAXPLANETS];
struct plasma_info_spacket      clientPlasmasInfo[MAXPLAYER * MAXPLASMA];
struct plasma_spacket           clientPlasmas[MAXPLAYER * MAXPLASMA];
int                             mustUpdate[MAXPLAYER];

extern int                      ignored[];	/* from main.c */
#ifdef SHORT_PACKETS

/* #define SHORT_THRESHOLD */

struct youss_spacket            clientSelfShip;
struct youshort_spacket         clientSelfShort;
/* HW */
static unsigned char            clientVPlanets[MAXPLANETS * sizeof(struct planet_s_spacket)
					       + 2 + 6];
static int                      clientVPlanetCount;
static int                      vtsize[9] =
{4, 8, 8, 12, 12, 16, 20, 20, 24};	/* How big is the SP_S_TORP p acket */
static int                      vtdata[9] =
{0, 3, 5, 7, 9, 12, 14, 16, 18};/* How big is Torpdata */
static int                      mustsend;	/* Flag to remind me that i
						 * must send SP_S_TORP */

static unsigned char            clientVTorps[40];

static unsigned char            clientVTorpsInfo[16];

static unsigned char            clientVPlayers[MAXPLAYER * VPLAYER_SIZE + 16];
#if MAXPLAYER > 32
static unsigned char            clientVXPlayers[33 * 4];
static int                      clientVXPlayerCount;
#endif
static int                      clientVPlayerCount;
static int                      big; /*, small */

 /* int           send_short      = 0; *//* is now in data.c  */
static int                      send_mesg = 1;
static int                      send_kmesg = 1;
static int                      send_warn = 1;

#ifdef SHORT_THRESHOLD
static int                      send_threshold = 0;	/* infinity */
#endif

static int                      spk_update_sall = 0;	/* Small Update: Only
							 * weapons, Kill s and
							 * Planets */
static int                      spk_update_all = 0;	/* Full Update minus
							 * SP_STATS */
#define         SPK_VOFF         0	/* variable packets off */
#define         SPK_VON          1	/* variable packets on */
#define         SPK_MOFF         2	/* message packets off */
#define         SPK_MON          3	/* message packets on */
#define         SPK_M_KILLS      4
#define         SPK_M_NOKILLS    5
#define         SPK_THRESHOLD    6
#define         SPK_M_WARN       7
#define         SPK_M_NOWARN     8
#define SPK_SALL 9		/* only planets,kills and weapons */
#define         SPK_ALL 10	/* Full Update - SP_STATS */

#endif				/* SHORT_PACKETS */

int 
connectToClient(machine, port)
   char                           *machine;
   int                             port;
{
   int                             ns;
   struct sockaddr_in              addr;
   struct hostent                 *hp;
   /* int len,cc; */
   /* int i; */

   if (sock != -1) {
      shutdown(sock, 2);
      sock = -1;
   }
#ifdef DEBUG
   printf("Connecting to %s through %d\n", machine, port);
   fflush(stdout);
#endif

   if ((ns = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
      printf("I cannot create a socket\n");
      fflush(stdout);
      exit(2);
   }
   addr.sin_family = AF_INET;
   addr.sin_port = htons(port);

   if (remoteaddr != -1) {
      addr.sin_addr.s_addr = remoteaddr;
   } else if ((addr.sin_addr.s_addr = inet_addr(machine)) == -1) {
      if ((hp = gethostbyname(machine)) == NULL) {
	 printf("I cannot get host name\n");
	 fflush(stdout);
	 close(ns);
	 exit(1);
      } else {
	 addr.sin_addr.s_addr = *(long *) hp->h_addr;
      }
   }
   remoteaddr = addr.sin_addr.s_addr;

   if (connect(ns, (struct sockaddr *) & addr, sizeof(addr)) < 0) {
#ifdef DEBUG
      printf("I cannot connect through port %d\n", port);
      fflush(stdout);
#endif
      close(ns);
      return (0);
   }
   sock = ns;
#ifdef KEEPALIVE
   set_tcp_opts(sock);
#endif
   initClientData();
   testtime = -1;
   return (1);
}
/*
 * Check the socket to read it's inet addr for possible future use.
 */
void 
checkSocket()
{
   struct sockaddr_in              sin;
   int                             length;
   length = sizeof(sin);
   if (getpeername(sock, (struct sockaddr *) & sin, &length) < 0) {
      /* A bad thing. */
      return;
   }
#ifdef KEEPALIVE
   set_tcp_opts(sock);
#endif

   remoteaddr = sin.sin_addr.s_addr;
}

void 
initClientData()
   /* invalidates all data, so it is all sent to the client */
{
   int                             i;
   clientDead = 0;
   for (i = 0; i < MAXPLAYER; i++) {
      clientHostile[i].hostile = -1;
      clientHostile[i].type = SP_HOSTILE;
      clientHostile[i].pnum = i;

      clientStats[i].losses = -1;
      clientStats[i].type = SP_STATS;
      clientStats[i].pnum = i;

      clientLogin[i].rank = -1;
      clientLogin[i].type = SP_PL_LOGIN;
      clientLogin[i].pnum = i;

      clientPlayersInfo[i].shiptype = -1;
      clientPlayersInfo[i].type = SP_PLAYER_INFO;
      clientPlayersInfo[i].pnum = i;

      clientPStatus[i].status = -1;
      clientPStatus[i].type = SP_PSTATUS;
      clientPStatus[i].pnum = i;

      clientPlayers[i].x = htonl(-1);
      clientPlayers[i].type = SP_PLAYER;
      clientPlayers[i].pnum = i;


      clientPhasers[i].status = -1;
      clientPhasers[i].type = SP_PHASER;
      clientPhasers[i].pnum = i;

      clientKills[i].kills = htonl(-1);
      clientKills[i].type = SP_KILLS;
      clientKills[i].pnum = i;

      clientFlags[i].flags = htonl(-1);
      clientFlags[i].type = SP_FLAGS;
      clientFlags[i].pnum = i;

      mustUpdate[i] = 0;

#ifdef SAVE_MEM
      if (udpMode != MODE_FAT)
	 continue;
#endif

      fat_hostile[i].packet = (PTR) & clientHostile[i];
      fat_hostile[i].pkt_size = sizeof(struct hostile_spacket);
      fat_hostile[i].prev = fat_hostile[i].next = (FAT_NODE *) NULL;
      fat_phaser[i].packet = (PTR) & clientPhasers[i];
      fat_phaser[i].pkt_size = sizeof(struct phaser_spacket);
      fat_phaser[i].prev = fat_phaser[i].next = (FAT_NODE *) NULL;
      fat_kills[i].packet = (PTR) & clientKills[i];
      fat_kills[i].pkt_size = sizeof(struct kills_spacket);
      fat_kills[i].prev = fat_kills[i].next = (FAT_NODE *) NULL;
      fat_flags[i].packet = (PTR) & clientFlags[i];
      fat_flags[i].pkt_size = sizeof(struct flags_spacket);
      fat_flags[i].prev = fat_flags[i].next = (FAT_NODE *) NULL;
   }
   for (i = 0; i < MAXPLAYER * MAXTORP; i++) {
      clientTorpsInfo[i].status = -1;
      clientTorps[i].x = -1;

#ifdef SAVE_MEM
      if (udpMode != MODE_FAT)
	 continue;
#endif

      fat_torp_info[i].packet = (PTR) & clientTorpsInfo[i];
      fat_torp_info[i].pkt_size = sizeof(struct torp_info_spacket);
      fat_torp_info[i].prev = fat_torp_info[i].next = (FAT_NODE *) NULL;
   }
   for (i = 0; i < MAXPLAYER * MAXPLASMA; i++) {
      clientPlasmasInfo[i].status = -1;
      clientPlasmas[i].x = -1;

#ifdef SAVE_MEM
      if (udpMode != MODE_FAT)
	 continue;
#endif

      fat_plasma_info[i].packet = (PTR) & clientPlasmasInfo[i];
      fat_plasma_info[i].pkt_size = sizeof(struct plasma_info_spacket);
      fat_plasma_info[i].prev = fat_plasma_info[i].next = (FAT_NODE *) NULL;
   }
   for (i = 0; i < MAXPLANETS; i++) {
      clientPlanets[i].armies = htonl(-1);
      clientPlanetLocs[i].x = htonl(-1);

#ifdef SAVE_MEM
      if (udpMode != MODE_FAT)
	 continue;
#endif

      fat_planet[i].packet = (PTR) & clientPlanets[i];
      fat_planet[i].pkt_size = sizeof(struct planet_spacket);
      fat_planet[i].prev = fat_planet[i].next = (FAT_NODE *) NULL;
   }
   msgCurrent = (mctl->mc_current + 1) % MAXMESSAGE;
   clientSelf.pnum = -1;

#ifdef SHORT_PACKETS
   clientSelfShip.damage = -1;
   clientSelfShort.pnum = -1;
#endif

#ifdef SAVE_MEM
   if (udpMode != MODE_FAT)
      return;
#endif

#ifdef SHORT_PACKETS
    clientSelfShip.damage = -1;
    clientSelfShort.pnum = -1;
    fat_s_you.packet = (PTR) &clientSelfShort;
    fat_s_you.pkt_size = sizeof(struct youshort_spacket);
    fat_s_you.prev = fat_s_you.next = (FAT_NODE *) NULL;
    fat_s_planet.packet = (PTR) clientVPlanets;
    fat_s_planet.pkt_size = 0; /* It's dynamic */
    fat_s_planet.prev =  fat_s_planet.next = (FAT_NODE *) NULL;
#endif

   fat_you.packet = (PTR) & clientSelf;
   fat_you.pkt_size = sizeof(struct you_spacket);
   fat_you.prev = fat_you.next = (FAT_NODE *) NULL;
   fat_status.packet = (PTR) & clientStatus;
   fat_status.pkt_size = sizeof(struct status_spacket);
   fat_status.prev = fat_status.next = (FAT_NODE *) NULL;

   for (i = 0; i < MAX_FAT_LIST; i++)
      fatlist[i].head = fatlist[i].tail = (FAT_NODE *) NULL;
}

int 
isClientDead()
{
   return (clientDead);
}

static
int			 me_x, me_y;

#ifdef SPEEDUP
/*
 * We can save a large number of calculations by calculating the bounding box
 * around our player once and using the results in the later loops
 */
static int 		me_x_winr;
static int		me_x_winl;
static int		me_y_winb;
static int		me_y_wint;

#define PtOffLocal(x,y)		(x < me_x_winl || \
				 x > me_x_winr || \
				 y > me_y_winb || \
				 y < me_y_wint)

#endif

#define SizeOfUDPUpdate()		(udpbufptr-udpbuf)
#define SizeOfTCPUpdate()		(bufptr-buf)

#ifdef SHORT_THRESHOLD
#define OVER_THRESHOLD(t)       (t && (bufptr-buf > t || \
				       udpbufptr-udpbuf > t))
#endif

void 
updateClient()
{
   int                             i;

   /* since, under SHORT_PACKETS, the coordinate system is dependent
      on my location, we copy it here and use me_x, me_y in place of
      me->p_x, me->p_y so we can be guaranteed it won't change during 
      the update cycle (by the daemon) */
   me_x = me->p_x;
   me_y = me->p_y;

#ifdef SPEEDUP
   /* This must be set before the PtOffLocal() macro can be used */

   me_x_winr = me_x + (SCALE * WINSIDE / 2);
   me_x_winl = me_x - (SCALE * WINSIDE / 2);
   me_y_winb = me_y + (SCALE * WINSIDE / 2);
   me_y_wint = me_y - (SCALE * WINSIDE / 2);

#endif

   for (i = 0; i < MAXPLAYER; i++) {
      mustUpdate[i] = 0;
   }

#ifdef SHORT_PACKETS
   if (send_short) {

      /*
      if(myship->s_type == ATT)
	 updateSelf();
      else
      */
	 updateSelfShort();
      updatePhasers();
      updateShips(0);
      SupdateTorps();		/* Torps MUST be after updateShips(). Because
				 * of new coordinate system HW */
      updatePlasmas();

#ifdef SHORT_THRESHOLD
      if(OVER_THRESHOLD(send_threshold)) goto end_update;
#endif
      updatePlanets();
#ifdef SHORT_THRESHOLD
      if(OVER_THRESHOLD(send_threshold)) goto end_update;
#endif
      updateMessages();
#ifdef SHORT_THRESHOLD
      if(OVER_THRESHOLD(send_threshold)) goto end_update;
#endif
      /* EXPERIMENT:  Don't inflate large packet with non-crucial stuff */
      if(SizeOfUDPUpdate() < 41)
	 updatePlayerStats();
#ifdef SHORT_THRESHOLD
      if(OVER_THRESHOLD(send_threshold)) goto end_update;
#endif
      /* EXPERIMENT:  Don't inflate large packet with non-crucial stuff */
      if(SizeOfUDPUpdate() < 50)
	 updateStatus();
   } else
#endif
   {

      updateSelf();
      /*
       * NOTE: updatePhasers must be before updateShips since a phaser hit on
       * a cloaker sets mustUpdate
       */
      updatePhasers();
      updateShips(0);
      updateTorps();
      updatePlasmas();
      updatePlanets();
      updateMessages();
      updatePlayerStats();
      updateStatus();
   }

#ifdef nodef
   if (bufptr == buf && (commMode != COMM_UDP || udpbufptr == udpbuf)) {
      /* We sent nothing!  We better send something to wake him */
      sendClientPacket((CVOID) & clientSelf);
   }
#endif
   /*
    * NOTE: this must be the last thing we do before flushing the socket
    * buffer.
    */
#ifdef nodef
   /* for SHORT_PACKETS test comment out repCount % */
#endif

   if (ping && ((repCount % efticks(5 * ping_freq)) == 0) && testtime == 0
       &&
   /*
    * these checks are here to make sure we don't ping from the updateClient
    * call in death().  The reason is that savestats() can take > 100ms,
    * invalidating the next ping lag calc
    */
       me->p_status != PDEAD && me->p_status != POUTFIT)
      sendClientPing();		/* ping.c */

#ifdef SHORT_THRESHOLD
   end_update:	;
#endif

   flushSockBuf();
   repCount++;

#ifdef nodef
   /* SHORT_PACKETS TEST */
   printf("done with update %d\n", msetime());
#endif
}

void 
updateStatus()
{
   /* Update status every 10 seconds? */
   if ((status->tourn != (unsigned char)clientStatus.tourn) ||
       (repCount % efticks(50) == 0 && ntohl(clientStatus.timeprod) != 
        timeprod_int())) {
      clientStatus.type = SP_STATUS;
      clientStatus.tourn = status->tourn;
      /* wrap-fix */
      clientStatus.armsbomb = htonl(status->armsbomb/10);
      clientStatus.planets = htonl(status->planets/10);
      clientStatus.kills = htonl(status->kills/10);
      clientStatus.losses = htonl(status->losses/10);
      clientStatus.time = htonl(status->time/10);
      clientStatus.timeprod = htonl(timeprod_int());
      sendClientPacket((CVOID) & clientStatus);
   }
}

void 
updateSelf()
{
   if (ntohl(clientSelf.fuel) != me->p_fuel ||
       ntohl(clientSelf.shield) != me->p_shield ||
       ntohl(clientSelf.damage) != me->p_damage ||
       ntohs(clientSelf.etemp) != me->p_etemp ||
       ntohs(clientSelf.wtemp) != me->p_wtemp ||
       ntohl(clientSelf.flags) != me->p_flags ||
       clientSelf.armies != me->p_armies ||
       clientSelf.swar != me->p_swar ||
       ntohs(clientSelf.whydead) != me->p_whydead ||
       ntohs(clientSelf.whodead) != me->p_whodead ||
       clientSelf.pnum != me->p_no) {

      /* we want to send it, but how? */
      clientSelf.type = SP_YOU;

      if (commMode == COMM_UDP) {
	 if (ntohl(clientSelf.flags) != me->p_flags ||
	     clientSelf.armies != me->p_armies ||
	     clientSelf.swar != me->p_swar) {
	    clientSelf.type = SP_YOU | 0x40;	/* mark as semi-critical */
	 }
	 if (ntohs(clientSelf.whydead) != me->p_whydead ||
	     ntohs(clientSelf.whodead) != me->p_whodead ||
	     clientSelf.pnum != me->p_no) {
	    clientSelf.type = SP_YOU | 0x80;	/* mark as critical */
	 }
      }
      clientSelf.pnum = me->p_no;
      clientSelf.flags = htonl(me->p_flags);
      clientSelf.swar = me->p_swar;
      clientSelf.hostile = me->p_hostile;
      clientSelf.armies = me->p_armies;
      clientSelf.shield = htonl(me->p_shield);
      clientSelf.fuel = htonl(me->p_fuel);
      clientSelf.etemp = htons(me->p_etemp);
      clientSelf.wtemp = htons(me->p_wtemp);
      clientSelf.whydead = htons(me->p_whydead);
      clientSelf.whodead = htons(me->p_whodead);
      clientSelf.damage = htonl(me->p_damage);
#ifdef INCLUDE_VISTRACT
      clientSelf.tractor = (char) me->p_tractor | 0x40;	/* ATM - visible tractor */
#endif
      sendClientPacket((CVOID) & clientSelf);
   }
}

#ifdef SHORT_PACKETS
void 
updateSelfShort()
{
   if (ntohs(clientSelfShip.damage) != me->p_damage ||
       ntohs(clientSelfShip.shield) != me->p_shield ||
       ntohs(clientSelfShip.fuel) != me->p_fuel ||
       ntohs(clientSelfShip.etemp) != me->p_etemp ||
       ntohs(clientSelfShip.wtemp) != me->p_wtemp) {

      clientSelfShip.type = SP_S_YOU_SS;
      clientSelfShip.damage = htons(me->p_damage);
      clientSelfShip.shield = htons(me->p_shield);
      clientSelfShip.fuel = htons(me->p_fuel);
      clientSelfShip.etemp = htons(me->p_etemp);
      clientSelfShip.wtemp = htons(me->p_wtemp);
      /* Feature: SELF_8FLAGS */
      if(F_self_8flags2){
	 /* 8 flags - set the best ones */
	 clientSelfShip.flags8 = (unsigned char) (
	    (me->p_flags & PFSHIELD) 	/* 0x001 */|
	    (me->p_flags & PFREPAIR) 	/* 0x002 */|
	   ((me->p_flags & PFCLOAK)<<2)	/* 0x004 */|
	   ((me->p_flags & PFGREEN)<<7) /* 0x008 */|
	   ((me->p_flags & PFYELLOW)<<7)/* 0x010 */|
	   ((me->p_flags & PFRED)<<7)   /* 0x020 */|
	   ((me->p_flags & PFTRACT)<<15)/* 0x040 */|
	   ((me->p_flags & PFPRESS)<<15)/* 0x080 */
	);
      }
      else /* F_self_8flags */
	 clientSelfShip.flags8 = (unsigned char)(me->p_flags & 0xff);
      sendClientPacket((CVOID)&clientSelfShip);
   }
   if (clientSelfShort.whydead != me->p_whydead ||
       clientSelfShort.whodead != me->p_whodead ||
       clientSelf.pnum != me->p_no) {
      clientSelfShort.type = SP_S_YOU | 0x80;	/* mark as critical */
      clientSelfShort.pnum = me->p_no;
      clientSelfShort.hostile = me->p_hostile;
      clientSelfShort.swar = me->p_swar;
      clientSelfShort.armies = me->p_armies;
      clientSelfShort.whydead = me->p_whydead;
      clientSelfShort.whodead = me->p_whodead;
      clientSelfShort.flags = htonl(me->p_flags);
      sendClientPacket((CVOID)&clientSelfShort);
   }
   else if (
      ntohl(clientSelfShort.flags) != me->p_flags ||
      clientSelfShort.armies != me->p_armies ||
      clientSelfShort.hostile != me->p_hostile ||
      clientSelfShort.swar != me->p_swar){

      clientSelfShort.type = SP_S_YOU;
      clientSelfShort.pnum = me->p_no;
      clientSelfShort.hostile = me->p_hostile;
      clientSelfShort.swar = me->p_swar;
      clientSelfShort.armies = me->p_armies;
      clientSelfShort.whydead = me->p_whydead;
      clientSelfShort.whodead = me->p_whodead;
      clientSelfShort.flags = htonl(me->p_flags);
      sendClientPacket((CVOID)&clientSelfShort);
   }
}
#endif				/* SHORT_PACKETS */
   

/* NEW */
void
updatePlayerStats()
{
   register struct stats_spacket  	*stats;
   register struct player  		*pl;
   register				i = efticks(5);

   if(repCount % i)
      return;

   i = (repCount/i) % MAXPLAYER;

   pl = &players[i];
   if(!pl->p_status) return;

   stats = &clientStats[i];

   /*
    * Send stat packets once per five updates. But, only send one.  We
    * will cycle through them all eventually.
    */
   if (  stats->tkills != htonl(pl->p_stats.st_tkills) ||
         stats->tlosses != htonl(pl->p_stats.st_tlosses) ||
	 stats->kills != htonl(pl->p_stats.st_kills) ||
	 stats->losses != htonl(pl->p_stats.st_losses) ||
	 stats->tticks != htonl(pl->p_stats.st_tticks) ||
         stats->tplanets != htonl(pl->p_stats.st_tplanets) ||
         stats->tarmies != htonl(pl->p_stats.st_tarmsbomb) ||
         stats->sbkills != htonl(pl->p_stats.st_sbkills) ||
         stats->sblosses != htonl(pl->p_stats.st_sblosses) ||
         stats->armies != htonl(pl->p_stats.st_armsbomb) ||
         stats->planets != htonl(pl->p_stats.st_planets)) {

      stats->tkills = htonl(pl->p_stats.st_tkills);
      stats->tlosses = htonl(pl->p_stats.st_tlosses);
      stats->kills = htonl(pl->p_stats.st_kills);
      stats->losses = htonl(pl->p_stats.st_losses);
      stats->tticks = htonl(pl->p_stats.st_tticks);
      stats->tplanets = htonl(pl->p_stats.st_tplanets);
      stats->tarmies = htonl(pl->p_stats.st_tarmsbomb);
      stats->sbkills = htonl(pl->p_stats.st_sbkills);
      stats->sblosses = htonl(pl->p_stats.st_sblosses);
      stats->armies = htonl(pl->p_stats.st_armsbomb);
      stats->planets = htonl(pl->p_stats.st_planets);
      stats->maxkills = htonl((int) (pl->p_stats.st_maxkills * 100));
      stats->sbmaxkills = htonl((int) (pl->p_stats.st_sbmaxkills * 100));
      /*
      stats->type = SP_STATS;
      stats->pnum = i;
      */
      sendClientPacket((CVOID) stats);
   }
}

void 
updateShips(check_myloc)

   int	check_myloc;
   
{
   register int                    i;
   register struct player         *pl;
   register struct plyr_info_spacket *cpli;
   register struct player_spacket *cpl;
   register struct kills_spacket  *kills;
   register struct flags_spacket  *flags;
   register struct pstatus_spacket *pstatus;
   register struct plyr_login_spacket *login;
   register struct hostile_spacket *hostile;
   int                             update;
   int                             x, y;

#ifdef nodef
   /* INL Flag mask */
#define FLAGMASK (PFSHIELD|PFBOMB|PFORBIT|PFCLOAK|PFROBOT|PFPRACTR|PFDOCK|PFTRACT|PFPRESS|PFDOCKOK)
#endif

/* client shouldn't need PFBOMB|PFORBIT|PFTRACT|PFPRESS */
#ifdef BASEPRACTICE	/* robots need a little more */
#define FLAGMASK (PFSHIELD|PFCLOAK|PFROBOT|PFPRACTR|PFDOCKOK|PFORBIT|PFTRACT|PFPRESS|PFDOCKOK)
#else
#define FLAGMASK (PFSHIELD|PFCLOAK|PFROBOT|PFPRACTR|PFDOCKOK)
#endif

#define INVISOMASK (PFCLOAK|PFROBOT|PFPRACTR|PFDOCKOK)

#ifdef SHORT_PACKETS
   clientVPlayerCount = 12;
   clientVPlayers[1] = 0;	/* How many packets are in the Big... */
#if MAXPLAYER > 32
   clientVXPlayers[1] = 0;
   clientVXPlayerCount = 4;
#endif
   big = 0;
#endif

   if(check_myloc){
      me_x = me->p_x;
      me_y = me->p_y;

#ifdef SPEEDUP
   /* This must be set before the PtOffLocal() macro can be used */

      me_x_winr = me_x + (SCALE * WINSIDE / 2);
      me_x_winl = me_x - (SCALE * WINSIDE / 2);
      me_y_winb = me_y + (SCALE * WINSIDE / 2);
      me_y_wint = me_y - (SCALE * WINSIDE / 2);
#endif
   }

   /* Please excuse the ugliness of this loop declaration */
   for (i = 0, pl = players, cpli = clientPlayersInfo, cpl = clientPlayers, kills = clientKills, flags = clientFlags, pstatus = clientPStatus, login = clientLogin, hostile = clientHostile;
	i < MAXPLAYER;
	i++, pl++, cpli++, cpl++, kills++, flags++, pstatus++, login++, hostile++) {
      update = 0;

      /* NEW */
      if(!pstatus->status && !pl->p_status) continue;

      if (strcmp(pl->p_name, login->name) != 0 ||
	  pl->p_stats.st_rank != login->rank ||
#ifdef LONGHOST
	  strncmp(pl->p_monitor, login->monitor, sizeof(login->monitor) - 1) != 0)
#else
	  strcmp(pl->p_monitor, login->monitor) != 0)
#endif
      {
	 strncpy(login->name, pl->p_name, 15);
	 strncpy(login->monitor, pl->p_monitor, 15);
	 strncpy(login->login, pl->p_login, 15);
	 login->name[15] = 0;
	 login->monitor[15] = 0;
	 login->login[15] = 0;
	 /*
	 login->type = SP_PL_LOGIN;
	 login->pnum = i;
	 */
	 login->rank = pl->p_stats.st_rank;
	 sendClientPacket((CVOID) login);
      }
      if ((pl != me && (pl->p_swar & me->p_team) != hostile->war) ||
	  (pl != me && (pl->p_hostile & me->p_team) != hostile->hostile) ||
	  (pl == me && pl->p_swar != hostile->war) ||
	(pl == me && pl->p_hostile != hostile->hostile)) {
	 /*
	 hostile->type = SP_HOSTILE;
	 */
	 if (pl == me) {
	    hostile->war = pl->p_swar;
	    hostile->hostile = pl->p_hostile;
	 } else {
	    hostile->war = (pl->p_swar & me->p_team);
	    hostile->hostile = (pl->p_hostile & me->p_team);
	 }
	 /*
	 hostile->pnum = i;
	 */
	 sendClientPacket((CVOID) hostile);
      }
      if (pl->p_ship.s_type != cpli->shiptype ||
	  pl->p_team != cpli->team) {
	 /*
	 cpli->pnum = i;
	 */
	 cpli->shiptype = pl->p_ship.s_type;
	 cpli->team = pl->p_team;

	 /* NEW: send both UDP & TCP */
	 cpli->type = SP_PLAYER_INFO;
	 sendClientPacket((CVOID) cpli);
	 cpli->type = SP_PLAYER_INFO | 0x80;
	 sendClientPacket((CVOID) cpli);

      }
      if (kills->kills != htonl((int) (pl->p_kills * 100))) {
	 /*
	 kills->type = SP_KILLS;
	 kills->pnum = i;
	 */
	 kills->kills = htonl((int) (pl->p_kills * 100));
	 sendClientPacket((CVOID) kills);
      }
      if (pstatus->status != pl->p_status) {
	 /*
	  * We update the location of people whose status has changed. (like
	  * if they just re-entered...)
	  */
	 update = 1;
	 pstatus->type = SP_PSTATUS;	/* now sent udp by default */
	 /*
	 pstatus->pnum = i;
	 */
	 pstatus->status = pl->p_status;
	 /* turn off ignore when a player leaves */
	 if (pl->p_status == PFREE)
	    ignored[i] = 0;
	 /* UdpTryToFix */
	 if(pl == me && promotion){
	    /* XXX: avoid race condition with rank/SP_PL_LOGIN */
	    pstatus->type = SP_PSTATUS | 0x80;
	    sendClientPacket((CVOID) pstatus);
	    promotion = 0;
	 }
	 else {
	 
	    /* we want to send PEXPLODE/PDEAD by udp.. everything else by
	       UDP and TCP (just to make sure we don't get out of order) */

	    /* UdpTryToFix */
	    switch(pl->p_status){
	       case PEXPLODE:
	       case PDEAD:
		  sendClientPacket((CVOID) pstatus);
		  break;

	       case PALIVE:
		  /* send both */
		  sendClientPacket((CVOID) pstatus);
		  pstatus->type = SP_PSTATUS | 0x80;
		  sendClientPacket((CVOID) pstatus);
		  break;
	       
	       default:
		  /* send both */
		  sendClientPacket((CVOID) pstatus);
		  pstatus->type = SP_PSTATUS | 0x80;
		  sendClientPacket((CVOID) pstatus);
		  break;
	    }
	 }
      }
      if (!hiddenenemy || pl->p_team == me->p_team ||
	  (pl->p_flags & PFSEEN) || (status->tourn == 0 && hiddenenemy < 2)) {
	 x = pl->p_x;
	 y = pl->p_y;
	 if (me != pl && flags->flags != htonl(FLAGMASK & pl->p_flags)) {
	    /*
	    flags->type = SP_FLAGS;
	    flags->pnum = i;
	    */
	    flags->flags = htonl(FLAGMASK & pl->p_flags);
#ifdef INCLUDE_VISTRACT
	    flags->tractor = (char) pl->p_tractor | 0x40;	/* ATM - visible tractor */
#endif
	    sendClientPacket((CVOID) flags);
	 }
      } else {
	 /* A hack to make him inviso */
	 x = -100000;
	 y = -100000;

	 /* reduce flag info if he's inviso 8/7/91 TC */

	 if (me != pl && flags->flags != htonl(INVISOMASK & pl->p_flags)) {
	    /*
	    flags->type = SP_FLAGS;
	    flags->pnum = i;
	    */
	    flags->flags = htonl(INVISOMASK & pl->p_flags);
	    sendClientPacket((CVOID) flags);
	 }
      }

      if (x != ntohl(cpl->x) || y != ntohl(cpl->y) ||
      pl->p_dir != cpl->dir || pl->p_speed != cpl->speed) {
	 /*
	  * We update the player if: 1) haven't updated him for 9 intervals.
	  * 2) he is on the screen 3) he was on the screen recently.
	  */
	 if (!update && repCount % efticks(9) != 0 &&
#ifdef SPEEDUP
	     (PtOffLocal(ntohl(cpl->x), ntohl(cpl->y)) &&
	      PtOffLocal(x, y))
#else
	 (ntohl(cpl->x) < me->p_x - SCALE * WINSIDE / 2 ||
	  ntohl(cpl->x) > me->p_x + SCALE * WINSIDE / 2 ||
	  ntohl(cpl->y) > me->p_y + SCALE * WINSIDE / 2 ||
	  ntohl(cpl->y) < me->p_y - SCALE * WINSIDE / 2) &&
	     (y > me->p_y + SCALE * WINSIDE / 2 ||
	      x > me->p_x + SCALE * WINSIDE / 2 ||
	      x < me->p_x - SCALE * WINSIDE / 2 ||
	      y < me->p_y - SCALE * WINSIDE / 2)
#endif				/* SPEEDUP */
	     )
	    continue;
	 /*
	  * If the guy is cloaked, give information only occasionally, and
	  * make it slightly inaccurate. Also, we don't give a direction. The
	  * client has no reason to know.
	  */
	 if ((pl->p_flags & PFCLOAK) &&
	     (pl->p_cloakphase == (CLOAK_PHASES - 1)) &&
	     (me != pl) && !mustUpdate[i]) {

	    if (repCount % efticks(9) != 0){
	       /* send old position again */
#ifdef SHORT_PACKETS
	       if (send_short)
		  updateVPlayer(cpl);
	       else
		  sendClientPacket((CVOID) cpl);
#else
		  sendClientPacket((CVOID) cpl);
#endif
	       continue;
	    }
#ifdef SHORT_PACKETS
	    /*
	    cpl->type = SP_PLAYER;
	    cpl->pnum = i;
	    */
	    /* Feature: CLOAK_MAXWARP */
	    if(F_cloak_maxwarp)
	       cpl->speed = 0xf;	/* NEW: Joe Rumsey, 15 == cloaked */
	    cpl->x = htonl(x + (random() % 2000) - 1000);
	    cpl->y = htonl(y + (random() % 2000) - 1000);
	    if (send_short) {
	       updateVPlayer(cpl);
	    } else {
	       sendClientPacket((CVOID) cpl);
	    }
#else
	    /*
	    cpl->type = SP_PLAYER;
	    cpl->pnum = i;
	    */
	    cpl->x = htonl(x + (random() % 2000) - 1000);
	    cpl->y = htonl(y + (random() % 2000) - 1000);
	    sendClientPacket((CVOID) cpl);
#endif
	    continue;
	 }
#ifdef SHORT_PACKETS
	 /*
	 cpl->type = SP_PLAYER;
	 cpl->pnum = i;
	 */
	 if(pl == me){
	    cpl->x = htonl(me_x);
	    cpl->y = htonl(me_y);
	 }
	 else{
	    cpl->x = htonl(x);
	    cpl->y = htonl(y);
	 }
	 /* Feature: CLOAK_MAXWARP */
	 if(F_cloak_maxwarp && pl != me && (pl->p_flags & PFCLOAK))
	    cpl->speed = 0xf;  /* NEW: Joe Rumsey, 15 == cloaked */
	 else
	    cpl->speed = pl->p_speed;

	 cpl->dir = pl->p_dir;
	 if (send_short) {
	    updateVPlayer(cpl);
	 } else {
	    sendClientPacket(cpl);
	 }
#else
	 /*
	 cpl->type = SP_PLAYER;
	 cpl->pnum = i;
	 */
	 cpl->x = htonl(x);
	 cpl->y = htonl(y);
	 /* Feature: CLOAK_MAXWARP */
	 if(pl != me && (pl->p_flags & PFCLOAK))
	    cpl->speed = 0xf;  /* NEW: Joe Rumsey, 15 == cloaked */
	 else
	    cpl->speed = pl->p_speed;
	 cpl->dir = pl->p_dir;
	 sendClientPacket((CVOID) cpl);
#endif				/* SHORT_PACKETS */
      }
   }

#ifdef SHORT_PACKETS
   if (send_short && (clientVPlayerCount > 12 
#if MAXPLAYER > 32
	|| clientVXPlayerCount > 4
#endif
		      || big != 0))
      sendVPlayers();
#endif
}

void 
updateTorps()
{
   register struct torp           *torp;
   register int                    i;
   register struct torp_info_spacket *tpi;
   register struct torp_spacket   *tp;
   for (i = 0, torp = torps, tpi = clientTorpsInfo, tp = clientTorps;
	i < MAXPLAYER * MAXTORP;
	i++, torp++, tpi++, tp++) {
      /* NEW */
      if(!torp->t_status && !tpi->status)
	 continue;
      if (torp->t_owner == me->p_no) {
	 if (torp->t_war != tpi->war ||
	     torp->t_status != tpi->status) {

	    tpi->type = SP_TORP_INFO;
	    tpi->war = torp->t_war;
	    tpi->status = torp->t_status;
	    tpi->tnum = htons(i);
	    sendClientPacket((CVOID) tpi);
	 }
#ifdef NEWSPEEDUP		/* packet waste */
	 if (torp->t_status != TFREE)
#endif
	    if (tp->x != htonl(torp->t_x) ||
		tp->y != htonl(torp->t_y)) {

	       tp->type = SP_TORP;
	       tp->x = htonl(torp->t_x);
	       tp->y = htonl(torp->t_y);
	       tp->dir = torp->t_dir;
	       tp->tnum = htons(i);
	       sendClientPacket((CVOID) tp);
	    }
      } else {			/* Someone else's torp... */
	 if (
#ifdef SPEEDUP
	       PtOffLocal(torp->t_x, torp->t_y) ||
#else
	       torp->t_y > me->p_y + SCALE * WINSIDE / 2 ||
	       torp->t_x > me->p_x + SCALE * WINSIDE / 2 ||
	       torp->t_x < me->p_x - SCALE * WINSIDE / 2 ||
	       torp->t_y < me->p_y - SCALE * WINSIDE / 2 ||
#endif				/* SPEEDUP */
	       torp->t_status == TFREE) {
	    if (torp->t_status == TFREE && tpi->status == TEXPLODE) {
	       tpi->status = TFREE;
	       continue;
	    }
	    if (tpi->status != TFREE) {
	       tpi->status = TFREE;
	       tpi->tnum = htons(i);
	       tpi->type = SP_TORP_INFO;
	       sendClientPacket((CVOID) tpi);
	    }
	 } else {		/* in view */
#ifdef NEWSPEEDUP		/* packet waste */
	    if (torp->t_status != TFREE)
#endif
	       if (tp->x != htonl(torp->t_x) ||
		   tp->y != htonl(torp->t_y)) {
		  tp->x = htonl(torp->t_x);
		  tp->y = htonl(torp->t_y);
		  tp->dir = torp->t_dir;
		  tp->tnum = htons(i);
		  tp->type = SP_TORP;
		  sendClientPacket((CVOID) tp);
	       }
	    if (torp->t_status != tpi->status ||
		(torp->t_war & me->p_team) != tpi->war) {
	       /* Let the client fade away the explosion on its own */
	       tpi->war = torp->t_war & me->p_team;
	       tpi->type = SP_TORP_INFO;
	       tpi->tnum = htons(i);
	       tpi->status = torp->t_status;
	       sendClientPacket((CVOID) tpi);
	    }
	 }
      }
   }
}

void 
updatePlasmas()
{
   register struct plasmatorp     *torp;
   register int                    i;
   register struct plasma_info_spacket *tpi;
   register struct plasma_spacket *tp;
   for (i = 0, torp = plasmatorps, tpi = clientPlasmasInfo, tp = clientPlasmas;
	i < MAXPLAYER * MAXPLASMA;
	i++, torp++, tpi++, tp++) {
      /* NEW */
      if(!torp->pt_status && !tpi->status)
	 continue;
      if (torp->pt_owner == me->p_no) {
	 if (torp->pt_war != tpi->war ||
	     torp->pt_status != tpi->status) {
	    tpi->type = SP_PLASMA_INFO;
	    tpi->war = torp->pt_war;
	    tpi->status = torp->pt_status;
	    tpi->pnum = htons(i);
	    sendClientPacket((CVOID) tpi);
	 }
	 if (tp->x != htonl(torp->pt_x) ||
	     tp->y != htonl(torp->pt_y)) {
	    tp->type = SP_PLASMA;
	    tp->x = htonl(torp->pt_x);
	    tp->y = htonl(torp->pt_y);
	    tp->pnum = htons(i);
	    sendClientPacket((CVOID) tp);
	 }
      } else {			/* Someone else's torp... */
	 if(
#ifdef SPEEDUP
	    PtOffLocal(torp->pt_x, torp->pt_y) ||
#else
	    torp->pt_y > me->p_y + SCALE * WINSIDE / 2 ||
	    torp->pt_x > me->p_x + SCALE * WINSIDE / 2 ||
	    torp->pt_x < me->p_x - SCALE * WINSIDE / 2 ||
	    torp->pt_y < me->p_y - SCALE * WINSIDE / 2 ||
#endif				/* SPEEDUP */
	    torp->pt_status == PTFREE) {
	    if (torp->pt_status == PTFREE && tpi->status == PTEXPLODE) {
	       tpi->status = PTFREE;
	       continue;
	    }
	    if (tpi->status != PTFREE) {
	       tpi->status = PTFREE;
	       tpi->pnum = htons(i);
	       tpi->type = SP_PLASMA_INFO;
	       sendClientPacket((CVOID) tpi);
	    }
	 } else {		/* in view */
	    /* Send torp (we assume it moved) */
	    tp->x = htonl(torp->pt_x);
	    tp->y = htonl(torp->pt_y);
	    tp->pnum = htons(i);
	    tp->type = SP_PLASMA;
	    sendClientPacket((CVOID) tp);
	    if (torp->pt_status != tpi->status ||
		(torp->pt_war & me->p_team) != tpi->war) {
	       tpi->war = torp->pt_war & me->p_team;
	       tpi->type = SP_PLASMA_INFO;
	       tpi->pnum = htons(i);
	       tpi->status = torp->pt_status;
	       sendClientPacket((CVOID) tpi);
	    }
	 }
      }
   }
}

void 
updatePhasers()
{
   register int                    i;
   register struct phaser_spacket *ph;
   register struct phaser         *phase;
   register struct player         *pl;
   for (i = 0, ph = clientPhasers, phase = phasers, pl = players;
	i < MAXPLAYER; i++, ph++, phase++, pl++) {
      /* NEW */
      if(!phase->ph_status && !ph->status)
	 continue;
      if (
#ifdef SPEEDUP
	    PtOffLocal(pl->p_x, pl->p_y)
#else
	    pl->p_y > me->p_y + SCALE * WINSIDE / 2 ||
	    pl->p_x > me->p_x + SCALE * WINSIDE / 2 ||
	    pl->p_x < me->p_x - SCALE * WINSIDE / 2 ||
	    pl->p_y < me->p_y - SCALE * WINSIDE / 2
#endif				/* SPEEDUP */
	  ) {
	 if (ph->status != PHFREE) {
	    /*
	    ph->pnum = i;
	    ph->type = SP_PHASER;
	    */
	    ph->status = PHFREE;
	    sendClientPacket((CVOID) ph);
	 }
      } else {
	 if (phase->ph_status == PHHIT) {
	    mustUpdate[phase->ph_target] = 1;
	 }
	 if (ph->status != phase->ph_status ||
	     ph->dir != phase->ph_dir ||
	     ph->target != htonl(phase->ph_target)) {
	    /*
	    ph->pnum = i;
	    ph->type = SP_PHASER;
	    */
	    ph->status = phase->ph_status;
	    ph->dir = phase->ph_dir;
	    ph->x = htonl(phase->ph_x);
	    ph->y = htonl(phase->ph_y);
	    ph->target = htonl(phase->ph_target);
	    sendClientPacket((CVOID) ph);
	 }
      }
   }
}

void 
updatePlanets()
{
   register int                    i;
   register struct planet_spacket *pl;
   register struct planet         *plan;
   register struct planet_loc_spacket *pll;
#ifdef SHORT_PACKETS
   register struct planet_s_spacket *npl =
	       (struct planet_s_spacket *) & clientVPlanets[2];
   register int                    numplanets = 0;
#endif
   static int			   myteam = ALLTEAM;	/* NEW */

#define PLFLAGMASK (PLREPAIR|PLFUEL|PLAGRI|PLHOME|PLCOUP)

   for (i = 0, pl = clientPlanets, plan = planets, pll = clientPlanetLocs;
	i < MAXPLANETS;
	i++, pl++, plan++, pll++) {
      /*
       * Send him info about him not having info if he doesn't but thinks he
       * does. Also send him info on every fifth cycle if the planet needs to
       * be redrawn.
       */
      /* myteam was me->p_team.  Problem:
	 when you switch teams, pl->info should be and'd with previous team,
	 not current team */
      if (((plan->pl_info & me->p_team) == 0) && (pl->info & myteam)) {
	 pl->type = SP_PLANET;
	 pl->pnum = i;
	 pl->info = 0;
#ifdef SHORT_PACKETS
	 if (send_short) {
	    npl->pnum = i;
	    npl->info = 0;
	    npl->armies = 0;
	    npl->owner = 0;
	    npl->flags = 0;
	    npl++;
	    numplanets++;
	 } else
#endif
	    sendClientPacket((CVOID) pl);
      } else if (plan->pl_info & me->p_team &&
		 (pl->info != plan->pl_info ||
		  pl->armies != htonl(plan->pl_armies) ||
		  pl->owner != plan->pl_owner ||
		  pl->flags != htons((short) (plan->pl_flags & PLFLAGMASK)))) {
	 pl->type = SP_PLANET;
	 pl->pnum = i;
	 pl->info = plan->pl_info;
	 pl->flags = htons((short) (plan->pl_flags & PLFLAGMASK));
	 pl->armies = htonl(plan->pl_armies);
	 pl->owner = plan->pl_owner;
#ifdef SHORT_PACKETS
	 if (send_short && plan->pl_armies < 256) {
	    npl->pnum = i;
	    npl->info = plan->pl_info;
	    npl->flags = htons((short) (plan->pl_flags & PLFLAGMASK));
	    npl->armies = (unsigned char) plan->pl_armies;
	    npl->owner = plan->pl_owner;
	    npl++;
	    numplanets++;
	 } else
#endif
	    sendClientPacket((CVOID) pl);
      }
      /* Assume that the planet only needs to be updated once... */
      if (pll->x != htonl(plan->pl_x) || pll->y != htonl(plan->pl_y)) {
	 pll->x = htonl(plan->pl_x);
	 pll->y = htonl(plan->pl_y);
	 pll->pnum = i;
	 strcpy(pll->name, plan->pl_name);
	 pll->type = SP_PLANET_LOC;
	 sendClientPacket((CVOID) pll);
      }
   }
#ifdef SHORT_PACKETS
   /* Now we can send our Planet Packet */
   if (numplanets != 0) {
      clientVPlanetCount = 2 + (numplanets * 6);
      /*
       * make sure on machine word boundary (assuming 4 bytes per w ord)
       */
      if ((clientVPlanetCount % 4) != 0)
	 clientVPlanetCount += (4 - (clientVPlanetCount % 4));

      sizes[SP_S_PLANET] = clientVPlanetCount;
      clientVPlanets[0] = SP_S_PLANET;	/* The type */
      clientVPlanets[1] = numplanets;
      sendClientPacket((CVOID) clientVPlanets);

   }
#endif
   myteam = me->p_team;		/* NEW */
}

void 
updateMessages()
{
   int                             i;
   struct message                 *cur;
   struct mesg_spacket             msg;
   int                             updated = 0;
#ifdef SHORT_PACKETS
   struct mesg_s_spacket          *shortmsg = (struct mesg_s_spacket *) & msg;
   register int                    size = 0;
#endif
#ifdef RCD
   struct distress dist;
#endif

   for (i = msgCurrent; i != (mctl->mc_current + 1) % MAXMESSAGE; i = (i + 1) % MAXMESSAGE) {
      if (i == MAXMESSAGE)
	 i = 0;
      cur = &messages[i];

      if (cur->m_flags & MVALID &&
	  (cur->m_flags & MALL ||
	   (cur->m_flags & MTEAM && cur->m_recpt & me->p_team) ||
       (cur->m_flags & MINDIV && cur->m_recpt == me->p_no)
      || (cur->m_flags & MGOD && cur->m_recpt == me->p_no)
	  )) {
#ifdef SHORT_PACKETS
	 if (send_short && (cur->m_from == 255)) {	/* Test if possible to
							 * send with
							 * SP_S_WARNING */
	    switch (cur->args[0]) {
	    case DMKILL:
	       if (cur->args[3] < 32000 && cur->args[4] < 32 && cur->args[1] < 64 && cur->args[2] < 64) {
		  ushort                          tmp = cur->args[3] | ((cur->args[4] & 16) << 11);

#ifdef FEATURE
		  if(F_why_dead){
		     swarning2(KILLARGS, (uchar) (tmp & 0xff), 
				(uchar) ((tmp >> 8) & 0xff),
		            KILLARGS2, cur->args[5], 0);
		  }
		  else
#endif
		     swarning(KILLARGS, (uchar) (tmp & 0xff), (uchar) ((tmp >> 8) & 0xff));
		  swarning(DMKILL, (uchar) cur->args[1] | ((cur->args[4] & 3) << 6), (uchar) cur->args[2] | ((cur->args[4] & 12) << 4));
		  msgCurrent = (msgCurrent + 1) % MAXMESSAGE;
		  continue;
	       }
	       break;
	    case KILLARGS:	/* Only to help the compiler */
	       break;
	    case KILLARGS2:	/* Only to help the compiler */
	       break;
	    case DMKILLP:
	       swarning(DMKILLP, cur->args[1], cur->args[2]);
	       msgCurrent = (msgCurrent + 1) % MAXMESSAGE;
	       continue;
	    case DMBOMB:
	       swarning2(ARGUMENTS, cur->args[2], 0,	/* Damage */
	                 DMBOMB, cur->args[1], cur->args[3]);	/* To get thevital info
								 * in one packet */
	       msgCurrent = (msgCurrent + 1) % MAXMESSAGE;
	       continue;
	    case DMDEST:
	       swarning(DMDEST, cur->args[1], cur->args[2]);
	       msgCurrent = (msgCurrent + 1) % MAXMESSAGE;
	       continue;
	    case DMTAKE:
	       swarning(DMTAKE, cur->args[1], cur->args[2]);
	       msgCurrent = (msgCurrent + 1) % MAXMESSAGE;
	       continue;
	    case DGHOSTKILL:
	       if (cur->args[2] < 64000) {
		  swarning2(KILLARGS, (uchar) (cur->args[2] & 0xff), 
			(uchar) ((cur->args[2] >> 8) & 0xff),
		        DGHOSTKILL, cur->args[1], 0);
		  msgCurrent = (msgCurrent + 1) % MAXMESSAGE;
		  continue;
	       }
	       break;
	    case INLDMKILLP:
	       swarning2(ARGUMENTS, (uchar) cur->args[3], 0,
	       		 INLDMKILLP, cur->args[1], cur->args[2]);
	       msgCurrent = (msgCurrent + 1) % MAXMESSAGE;
	       continue;
	    case INLDMKILL:
	       if (cur->args[3] < 32000 && cur->args[4] < 32 && cur->args[1] < 64 && cur->args[2] < 64) {
		  ushort                          tmp = cur->args[3] | ((cur->args[4] & 16) << 11);

#ifdef FEATURE
		  if(F_why_dead){
		     swarning2(KILLARGS, (uchar) (tmp & 0xff), 
				(uchar) ((tmp >> 8) & 0xff),
		            KILLARGS2, cur->args[5], 0);
		  }
		  else
#endif
		     swarning(KILLARGS, (uchar) (tmp & 0xff), (uchar) ((tmp >> 8) & 0xff));
		  swarning(INLDMKILL, (uchar) cur->args[1] | ((cur->args[4] & 3) << 6), (uchar) cur->args[2] | ((cur->args[4] & 12) << 4));
		  msgCurrent = (msgCurrent + 1) % MAXMESSAGE;
		  continue;
	       }
	       break;
	    case INLDRESUME:
	       swarning(INLDRESUME, cur->args[1], 0);
	       msgCurrent = (msgCurrent + 1) % MAXMESSAGE;
	       continue;
	    case INLDTEXTE:
	       swarning(INLDTEXTE, cur->args[1], 0);
	       msgCurrent = (msgCurrent + 1) % MAXMESSAGE;
	       continue;
	    case SVALID:	/* We can send it over SP_S_MESSAGE HW */
	       break;
	    case SINVALID:	/* Can't be send over SP_S_MESSAGE */
	       break;
	    default:
	       /*
	        * I must test if the message from GOD has a real header (==
	        * "GOD->")
	        */
	       /* This tests gets all messages from tools */
	       if (strstr(cur->m_data, "GOD->") != NULL)
		  cur->args[0] = SVALID;	/* It has a real header */
	       break;

	    }
	 }
#endif

	 if (!updated) {
	    update_sys_defaults( /* ntserv calling ... */ 0);
	    updated = 1;
	    if (allwar) {
	       me->p_hostile = (FED | ROM | KLI | ORI);
	       me->p_swar = (FED | ROM | KLI | ORI);
	    }
	 }
#ifdef SHORT_PACKETS
	 if (send_short && (cur->m_from < MAXPLAYER || cur->args[0] == SVALID)
	     && (!(cur->m_flags & MLEAVE))) {
	    /*
	     * This is the SAFE version.(All playermessages must have a header
	     * .) It's nicer without MLEAVE
	     */
	    shortmsg->type = SP_S_MESSAGE;
	    shortmsg->m_flags = cur->m_flags;
	    shortmsg->m_recpt = cur->m_recpt;
	    shortmsg->m_from = cur->m_from;
	    cur->m_data[79] = '\0';	/* Is this all right ? */
	    strcpy(&shortmsg->mesg, &cur->m_data[9]);	/* 9 because of GOD mes
							 * sages */
	    /* Now the size */
	    size = strlen(&shortmsg->mesg);
	    size += 6;		/* 1 for '\0', 5 packetheader */
	    if ((size % 4) != 0)
	       size += (4 - (size % 4));
	    shortmsg->length = size;
	    sizes[SP_S_MESSAGE] = size;
	 } else
#endif
	 {

	    msg.type = SP_MESSAGE;
	    strncpy(msg.mesg, cur->m_data, 80);
	    msg.mesg[79] = '\0';
	    msg.m_flags = cur->m_flags;
	    msg.m_recpt = cur->m_recpt;
	    msg.m_from = cur->m_from;
	 }

#ifdef RCD
	 /* kludge for clients that can't handle RCD's */
	 /* if I understood heiko's short_packet code I would put this in
            the above ... but I dont - jmn */

	 if ((cur->m_recpt == me->p_team) && !me->p_gen_distress &&
	     (cur->m_flags == (MTEAM | MDISTR | MVALID))) {
	    char buf[MSG_LEN];
	    buf[0] = 0;
	    msg.m_flags ^= MDISTR; /* get rid of MDISTR flag so client
                                          isn't confused */
	    HandleGenDistr(cur->m_data,cur->m_from,cur->m_recpt,&dist);
	    makedistress(&dist,buf, distmacro[dist.distype].macro);
	    /* note that we do NOT send the F0->FED part */
	    strncpy (msg.mesg, buf, MSG_LEN);
	    msg.mesg[MSG_LEN-1] = '\0';

        }
#endif

	 if(!parseQuery(&msg)){

	    if (cur->m_from < 0 || cur->m_from > MAXPLAYER)
	       sendClientPacket((CVOID) & msg);
	    else if (cur->m_flags & MALL && !(ignored[cur->m_from] & MALL))
	       sendClientPacket((CVOID) & msg);
	    else if (cur->m_flags & MTEAM && !(ignored[cur->m_from] & MTEAM))
	       sendClientPacket((CVOID) & msg);
	    else if (cur->m_flags & MINDIV) {
	       if (cur->m_from < MAXPLAYER && (ignored[cur->m_from] & MINDIV))
		  bounce("That player is currently ignoring you.",
			 cur->m_from);
	       else
		  sendClientPacket((CVOID) & msg);
	    }
	    else if (cur->m_flags & MGOD)
	       sendClientPacket((CVOID) & msg);
	 }
      }
      msgCurrent = (msgCurrent + 1) % MAXMESSAGE;
   }
}

void 
sendQueuePacket(pos)
   short                           pos;
{
   struct queue_spacket            qPacket;
   qPacket.type = SP_QUEUE;
   qPacket.pos = htons(pos);
   sendClientPacket((CVOID) & qPacket);
   flushSockBuf();
}

void 
sendClientPacket(packet)
   /* Pick a random type for the packet */
   struct player_spacket          *packet;
{
   int                             orig_type;
   int                             size, issc;
   int                             cc = 0;
   static int                      oldStatus = POUTFIT;
   orig_type = packet->type;
   packet->type &= ~(0x40 | 0x80);	/* clear special flags */
#ifdef MAYBE
   /*
    * If we're dead, dying, or just born, we definitely want the transmission
    * to get through (otherwise we can get stuck).  I don't think this will
    * be a problem for anybody, though it might hang for a bit if the TCP
    * connection is bad.
    */
   /* Okay, now I'm not so sure.  Whatever. */
   if (oldStatus != PALIVE || (me != NULL && me->p_status != PALIVE))
      orig_type = packet->type | 0x80;	/* pretend it's critical */
#endif
   if (packet->type < 1 || packet->type > NUM_SIZES || sizes[(int) packet->type] == 0) {
      printf("Attempt to send strange packet %d\n", packet->type);
      fflush(stdout);
      return;
   }
#ifdef PACKSTATS
   packetsSent[(int) packet->type]++;
#endif

   /* SHORT_PACKETS TEST
    * printf("sent type %d\n", packet->type);
    */

   if (commMode == COMM_TCP || (commMode == COMM_UDP && udpMode == MODE_TCP)) {
      /*
       * business as usual
       */
      size = sizes[(int) packet->type];
      if (bufptr - buf + size >= BUFSIZE) {
	 if ((cc = gwrite(sock, buf, bufptr - buf)) != bufptr - buf) {
	    fprintf(stderr, "std gwrite failed (%d, error %d)\n", cc, errno);
	    fflush(stderr);
	    clientDead = 1;
	 }
	 bufptr = buf;
      }
      bcopy(packet, bufptr, size);
      bufptr += size;

   } else {
      /*
       * do UDP stuff unless it's a "critical" packet (note that both kinds
       * get a sequence number appended) (FIX)
       */
      issc = 0;

      switch (orig_type) {

      case SP_KILLS:
      case SP_TORP_INFO:
      case SP_PHASER:
      case SP_PLASMA_INFO:
      case SP_YOU | 0x40:
      case SP_STATUS:
      case SP_PLANET:
      case SP_FLAGS:
      case SP_HOSTILE:

#ifdef SHORT_PACKETS
      case SP_S_YOU:
      case SP_S_PLANET:
#endif
      /*
       * these are semi-critical; flag as semi-critical and fall through
       */
	 issc = 1;

#ifdef SHORT_PACKETS
      case SP_S_YOU_SS:
      case SP_S_PLAYER:
      case SP_S_TORP:
      case SP_S_TORP_INFO:
      case SP_S_8_TORP:
#endif

      case SP_PSTATUS:		/* NEW */
      case SP_PLAYER_INFO:	/* NEW */
      case SP_PLAYER:
      case SP_TORP:
      case SP_YOU:
      case SP_PLASMA:
      case SP_STATS:
      case SP_SCAN:
      case SP_PING:
      case SP_UDP_REPLY:	/* only reply when COMM_UDP is SWITCH_VERIFY */
	 /* these are non-critical updates; send them via UDP */
	 V_UDPDIAG(("Sending type %d\n", packet->type));
	 size = sizes[(int) packet->type];
	 packets_sent++;
	 if (udpbufptr - udpbuf + size >= UDPBUFSIZE) {
	    if ((cc = gwrite(udpSock, udpbuf, udpbufptr - udpbuf)) !=
		udpbufptr - udpbuf) {
	       fprintf(stderr, "UDP gwrite failed (%d, error %d)\n",
		       cc, errno);
	       /* clientDead=1; */
	       fflush(stderr);
	       UDPDIAG(("*** UDP disconnected for %s\n", me->p_name));
	       printUdpInfo();
	       closeUdpConn();
	       commMode = COMM_TCP;
	    }
	    udpbufptr = udpbuf + addSequence(udpbuf);
	 }
	 bcopy(packet, udpbufptr, size);
	 udpbufptr += size;
	 if (issc && udpMode == MODE_FAT) {
	    updateFat(packet);
	 }
	 break;

      default:
	 /* these are critical packets; send them via TCP */
	 size = sizes[(int) packet->type];
	 if (bufptr - buf + size >= BUFSIZE) {
	    if (gwrite(sock, buf, bufptr - buf) != bufptr - buf) {
	       fprintf(stderr, "TCP gwrite failed (%d, error %d)\n",
		       cc, errno);
	       fflush(stderr);
	       clientDead = 1;
	    }
	    bufptr = buf /* + addSequence(buf) */ ;
	 }
	 bcopy(packet, bufptr, size);
	 bufptr += size;
	 break;
      }
   }

   if (me != NULL)
      oldStatus = me->p_status;
}

/* Flush the socket buffer */
void 
flushSockBuf()
{
   int                             cc;
   if (clientDead)
      return;
   
   /*
    * This is where we try to add fat.  There's no point in checking at the
    * other places which call gwrite(), because they only call it when the
    * buffer is already full.
    */
   if (udpSock >= 0 && udpMode == MODE_FAT && (udpbufptr - udpbuf) < FAT_THRESH)
      fatten();
   if (udpSock >= 0 && udpbufptr != udpbuf) {
#ifdef BROKEN
      /* debugging only!! */
#ifdef HOSED
      if (sequence % 10)
#else
      if (broke && (random() % 4) == 1) {
#endif
	 /* act as if we did the gwrite(), but don't */
	 udpbufptr = udpbuf + addSequence(udpbuf);
	 goto foo;
      }
#endif
      if ((cc = gwrite(udpSock, udpbuf, udpbufptr - udpbuf))
	  != udpbufptr - udpbuf) {
	 fprintf(stderr, "UDP flush gwrite failed (%d, error %d)\n",
		 cc, errno);
	 /* clientDead=1; */
	 fflush(stderr);
	 UDPDIAG(("*** UDP disconnected for %s\n", me->p_name));
	 printUdpInfo();
	 closeUdpConn();
	 commMode = COMM_TCP;
      }
#ifdef nodef
      /* SHORT_PACKETS TEST */
      printf("UDP: %d bytes\n", udpbufptr-udpbuf);
#endif
      udpbufptr = udpbuf + addSequence(udpbuf);
   }
#ifdef BROKEN
 foo:
#endif

   /* NEW: flush TCP after udp */
   if (bufptr != buf) {
      if ((cc = gwrite(sock, buf, bufptr - buf)) != bufptr - buf) {
	 fprintf(stderr, "std flush gwrite failed (%d, error %d)\n",
		 cc, errno);
	 fflush(stderr);
	 clientDead = 1;
      }
#ifdef nodef
      /* SHORT_PACKETS TEST */
      printf("TCP: %d bytes\n", bufptr-buf);
#endif
      bufptr = buf /* + addSequence(buf) */ ;
   }

   if (udpMode == MODE_FAT)
      fatMerge();
}

void 
socketPause()
{
   struct timeval                  timeout;
   fd_set                          readfds;
   /*
    * this looks like an oversight here .. forgot to select on the udp socket
    * as well
    */

   timeout.tv_sec = 1;
   timeout.tv_usec = 0;
   FD_ZERO(&readfds);
   FD_SET(sock, &readfds);
   if (udpSock >= 0)
      FD_SET(udpSock, &readfds);

   select(32, &readfds, 0, 0, &timeout);
}
/* Find out if client has any requests */
int 
readFromClient()
{
   struct timeval                  timeout;
   fd_set                          readfds;
   int                             retval = 0;
   int                             s;
   if (clientDead)
      return (0);
 again:;			/* for EINTR */
   timeout.tv_sec = 0;
   timeout.tv_usec = 0;
   FD_ZERO(&readfds);
   FD_SET(sock, &readfds);
   if (udpSock >= 0)
      FD_SET(udpSock, &readfds);
   if ((s = select(32, &readfds, 0, 0, &timeout)) > 0) {
      /* Read info from the xtrek client */
      /* read the udp sock first for best numbers */
      if (udpSock >= 0 && FD_ISSET(udpSock, &readfds)) {
	 V_UDPDIAG(("Activity on UDP socket\n"));
	 retval += doRead(udpSock);
      }
      if (FD_ISSET(sock, &readfds)) {
	 retval += doRead(sock);
      }
   } else if (s < 0)
      goto again;
   return (retval != 0);	/* convert to 1/0 */
}

char                            _buf[BUFSIZ * 2];
static int                      rsock;
/* ripped out of above routine */
int 
doRead(asock)
   int                             asock;
{
   int				   sv;
   struct timeval                  timeout;
   /* int readfds; */
   fd_set                          readfds;
   char                           *bufptr;
   int                             size;
   int                             count;
   int                             temp;
   rsock = asock;

   timeout.tv_sec = 0;
   timeout.tv_usec = 0;
   /* readfds = 1<<asock; */
   FD_ZERO(&readfds);
   FD_SET(asock, &readfds);
   /* Read info from the xtrek server */
   count = read(asock, _buf, BUFSIZ * 2);

   /*
    * printf("read %d bytes from %s socket\n", count,
    * asock==sock?"TCP":"UDP");
    */

   if (count <= 0) {
#if DEBUG
      /* (this happens when the client hits 'Q') */
      fprintf(stderr, "%s@%s:1) read() failed in doRead (%d, error %d)\n",
	      login, host, count, errno);
      fprintf(stderr, "asock=%d, sock=%d\n", asock, sock);
      fflush(stderr);
#endif
      if (asock == udpSock) {
	 if (errno == ECONNREFUSED) {
	    struct sockaddr_in              addr;
	    UDPDIAG(("Hiccup(%d)!  Reconnecting\n", errno));
	    addr.sin_addr.s_addr = remoteaddr;
	    addr.sin_port = htons(udpClientPort);
	    addr.sin_family = AF_INET;
	    if (connect(udpSock, (struct sockaddr *) & addr, sizeof(addr)) < 0) {
	       perror("hiccup connect");
	       UDPDIAG(("Unable to reconnect\n"));
	       /* and fall through to disconnect */
	    } else {
	       UDPDIAG(("Reconnect successful\n"));
	       return (0);
	    }
	 }
	 UDPDIAG(("*** UDP disconnected (res=%d, err=%d)\n",
		  count, errno));
	 printUdpInfo();
	 closeUdpConn();
	 commMode = COMM_TCP;
	 return (0);
      }
      clientDead = 1;
      return (0);
   }
   bufptr = _buf;
   while (bufptr < _buf + count) {
      if (*bufptr < 1 || *bufptr > NUM_PACKETS ||
	  handlers[(int) *bufptr].size == 0) {
	 printf("Unknown packet type: %d, aborting...\n", *bufptr);
	 fflush(stdout);
	 return 0;
      }
      size = handlers[(int) *bufptr].size;
#ifdef SHORT_PACKETS
      if (size == -1) {		/* variable packet */
	 switch (*bufptr) {
	 case CP_S_MESSAGE:
	    size = ((char) bufptr[3]);
	    break;
	 default:
            fprintf(stderr, "Unknown variable packet %d\n", *bufptr); 
            return 0;
	 }
	 if ((size % 4) != 0)
	    size += (4 - (size % 4));

      }
#endif
      while (size > count + (_buf - bufptr)) {
	 /*
	  * We wait for up to twenty seconds for rest of packet. If we don't
	  * get it, we assume the client died.
	  */
again:
	 ;
	 timeout.tv_sec = 20;
	 timeout.tv_usec = 0;
	 /* readfds=1<<asock; */
	 FD_ZERO(&readfds);
	 FD_SET(asock, &readfds);
	 sv = select(32, &readfds, 0, 0, &timeout);
	 if (!sv){
	    logmessage("Died while waiting for packet...");
	    fprintf(stderr, "1a) read() failed (%d, error %d)\n",
		    count, errno);
	    fflush(stderr);
	    clientDead = 1;
	    return (0);
	 }
	 else if(sv < 0 && errno == EINTR){
	    goto again;
	 }
	    
	 temp = read(asock, _buf + count, size - (count + (_buf - bufptr)));
	 count += temp;
	 if (temp <= 0) {
	    sprintf(_buf, "Died in second read(), return=%d", temp);
	    logmessage(_buf);
	    fprintf(stderr, "2) read() failed (%d, error %d)\n",
		    count, errno);
	    fflush(stderr);
	    clientDead = 1;
	    return (0);
	 }
      }
      /*
       * Check to see if the handler is there and the request is legal. The
       * code is a little ugly, but it isn't too bad to worry about yet.
       */
#ifdef PACKSTATS
      packetsReceived[(int) *bufptr]++;
#endif
      if (asock == udpSock)
	 packets_received++;
      if (handlers[(int) *bufptr].handler != NULL) {
	 if (((inputMask == -1 || inputMask == *bufptr) &&
	      (me == NULL || !(me->p_flags & (PFWAR | PFREFITTING
#ifdef SB_TRANSWARP
					      | PFTWARP
#endif
			       )))) ||
	     *bufptr == CP_RESETSTATS || 
	     *bufptr == CP_UPDATES ||
	     *bufptr == CP_OPTIONS || 
	     *bufptr == CP_RESERVED ||
	     *bufptr == CP_MESSAGE ||		/* NEW - TSH */
#ifdef SHORT_PACKETS
	     *bufptr == CP_S_MESSAGE ||		/* NEW - TSH */
#endif
#ifdef RSA
	     *bufptr == CP_RSA_KEY ||
#endif
#ifdef FEATURE
	     *bufptr == CP_FEATURE ||
#endif
	     *bufptr == CP_UDP_REQ ||
	     *bufptr == CP_PING_RESPONSE ||
	     *bufptr == CP_SOCKET || 
	     *bufptr == CP_BYE) {
	    if (me && me->p_flags & PFSELFDEST
		&& *bufptr != CP_PING_RESPONSE
		) {
	       me->p_flags &= ~PFSELFDEST;
	       warning(1, "Self Destruct has been canceled", -1);
	    }
#ifdef nodef
	    /*
	     * The condition check is not rigourous enough -- clients can
	     * still send packets at the beginning of the game when me==NULL
	     * and inputMask=-1 condition fixes it(inefficient code but only
	     * executed once in the game..) else if              (me || (!me
	     * && (*bufptr == CP_RESETSTATS || bufptr == CP_UPDATES || bufptr
	     * == CP_OPTIONS || bufptr == CP_RESERVED || bufptr == CP_SOCKET
	     * || bufptr == CP_BYE)))
	     */
#endif
	    (*(handlers[(int) *bufptr].handler)) (bufptr);
	 }
	 /* Otherwise we ignore the request */
      } else {
	 printf("Handler for packet %d not installed...\n", *bufptr);
	 fflush(stdout);
      }
      bufptr += size;
      if (bufptr > _buf + BUFSIZ) {
	 bcopy(buf + BUFSIZ, _buf, BUFSIZ);
	 if (count == BUFSIZ * 2) {
	    /* readfds = 1<<asock; */
	    FD_ZERO(&readfds);
	    FD_SET(asock, &readfds);
	    if (select(32, &readfds, 0, 0, &timeout)) {
	       temp = read(asock, _buf + BUFSIZ, BUFSIZ);
	       count = BUFSIZ + temp;
	       if (temp <= 0) {
		  sprintf(_buf, "Died in third read(), return=%d", temp);
		  fprintf(stderr, "3) read() failed (%d, error %d)\n",
			  count, errno);
		  fflush(stderr);
		  logmessage(_buf);
		  clientDead = 1;
		  return (0);
	       }
	    } else {
	       count = BUFSIZ;
	    }
	 } else {
	    count -= BUFSIZ;
	 }
	 bufptr -= BUFSIZ;
      }
   }
   return (1);
}

void 
handleTorpReq(packet)
   struct torp_cpacket            *packet;
{
#ifdef BD
   bd_dir("torp", packet->dir);
#endif
   ntorp(packet->dir, TMOVE);
}

void 
handlePhasReq(packet)
   struct phaser_cpacket          *packet;
{
   phaser(packet->dir);
}

void 
handleSpeedReq(packet)
   struct speed_cpacket           *packet;
{
   set_speed(packet->speed);
}

void 
handleDirReq(packet)
   struct dir_cpacket             *packet;
{
#ifdef BD
   bd_dir("dir", packet->dir);
#endif
   me->p_flags &= ~(PFPLOCK | PFPLLOCK);
   set_course(packet->dir);
}

void 
handleShieldReq(packet)
   struct shield_cpacket          *packet;
{
   if (packet->state) {
      /* UdpTryToFix */
      if(me->p_flags & PFSHIELD){
	 /* redundant request -- maybe packet dropped.  Solution:
	    give him an update anyway */
	 clientSelf.flags = -1;
#ifdef SHORT_PACKETS
	 clientSelfShort.flags = -1;
#endif
      }
      else
	 shield_up();
   } else {
      if(!(me->p_flags & PFSHIELD)){
	 /* redundant request -- maybe packet dropped.  Solution:
	    give him an update anyway */
	 clientSelf.flags = -1;
#ifdef SHORT_PACKETS
	 clientSelfShort.flags = -1;
#endif
      }
      else
	 shield_down();
   }
}

void 
handleRepairReq(packet)
   struct repair_cpacket          *packet;
{
   if (packet->state) {
      repair();
   } else {
      me->p_flags &= ~(PFREPAIR);
   }
}

void 
handleOrbitReq(packet)
   struct orbit_cpacket           *packet;
{
   if (packet->state) {
      orbit();
   } else {
      me->p_flags &= ~PFORBIT;
      if (me->p_flags & PFDOCK) {
#ifdef SB_DOCK
	 players[me->p_docked].p_numdocked--;
	 players[me->p_docked].p_port[me->p_portdocked] = VACANT;
#else
	 players[me->p_docked].p_docked--;
	 players[me->p_docked].p_port[me->p_port[0]] = VACANT;
#endif
	 me->p_flags &= ~PFDOCK;
      }
   }
}

void 
handlePractrReq(packet)
   struct practr_cpacket          *packet;
{
#ifdef SB_TRANSWARP
   if (twarpMode) {
      register                        i;
      register struct player         *j;
      for (i = 0, j = &players[i]; i < MAXPLAYER; i++, j++) {
	 if (j->p_status != PALIVE)
	    continue;
	 if (j == me)
	    continue;
	 /*
	  * if we reached here there are other players in the game, therefore
	  * practice_robo would fail.
	  */
	 handleTranswarp();
	 return;
      }
   }
#endif
   /*
    * if we reach here there are no other players in the game and we can
    * start a practice robot
    */
   practice_robo();
}

void 
handleBombReq(packet)
   struct bomb_cpacket            *packet;
{
#ifdef MINES
   if (mines && me->p_ship.s_type == STARBASE) {
      mine_det();
      return;
   }
#endif

   if (packet->state) {
      bomb_planet();
   } else {
      me->p_flags &= ~(PFBOMB);
   }
}

void 
handleBeamReq(packet)
   struct beam_cpacket            *packet;
{
   if (packet->state == 1) {
      beam_up();
   } else if (packet->state) {
      beam_down();
   } else {
      me->p_flags &= ~(PFBEAMUP | PFBEAMDOWN);
   }
}

void 
handleCloakReq(packet)
   struct cloak_cpacket           *packet;
{
   if (packet->state) {
      /* UdpTryToFix */
      if(me->p_flags & PFCLOAK){
	 /* redundant request -- maybe packet dropped.  Solution:
	    give him an update anyway */
	 clientSelf.flags = -1;
#ifdef SHORT_PACKETS
	 clientSelfShort.flags = -1;
#endif
      }
      else
	 cloak_on();
   } else {
      if(!(me->p_flags & PFCLOAK)){
	 /* redundant request -- maybe packet dropped.  Solution:
	    give him an update anyway */
	 clientSelf.flags = -1;
#ifdef SHORT_PACKETS
	 clientSelfShort.flags = -1;
#endif
      }
      else
	 cloak_off();
   }
}

void 
handleDetTReq(packet)
   struct det_torps_cpacket       *packet;
{
   detothers();
}

void 
handleCopilotReq(packet)
   struct copilot_cpacket         *packet;
{
   /*
    * Unsupported... if (packet->state) { me->p_flags |= PFCOPILOT; } else {
    * me->p_flags &= ~PFCOPILOT; }
    */
}

void 
handleOutfit(packet)
   struct outfit_cpacket          *packet;
{
   shipPick = packet->ship;
   teamPick = packet->team;
}

void 
sendPickokPacket(state)
   int                             state;
{
   struct pickok_spacket           pickPack;
   pickPack.type = SP_PICKOK;
   pickPack.state = state;
   sendClientPacket((CVOID) & pickPack);
}

void 
handleLoginReq(packet)
   struct login_cpacket           *packet;
{

   strncpy(namePick, packet->name, 16);
   namePick[15] = 0;
   strncpy(passPick, packet->password, 16);
   passPick[15] = 0;
   /* Is this a name query or a login? */
   if (packet->query) {
      passPick[15] = 1;
   }
   strncpy(login, packet->login, 16);
   login[15] = 0;
}

void 
sendClientLogin(stats)
   struct stats                   *stats;
{
   struct login_spacket            logPacket;
   logPacket.type = SP_LOGIN;
   if (stats == NULL) {
      logPacket.accept = 0;
   } else {
      logPacket.accept = 1;
      logPacket.flags = htonl(stats->st_flags);
      bcopy(stats->st_keymap, logPacket.keymap, 96);
   }
   sendClientPacket((CVOID) & logPacket);
}

void 
handlePlasmaReq(packet)
   struct plasma_cpacket          *packet;
{
#ifdef BD
   bd_dir("plasma", packet->dir);
#endif
   nplasmatorp(packet->dir, PTMOVE);
}

void 
handleWarReq(packet)
   struct war_cpacket             *packet;
{
   declare_war(packet->newmask);
}

void 
handlePlanlockReq(packet)
   struct planlock_cpacket        *packet;
{
   lock_planet(packet->pnum);
}

void 
handlePlaylockReq(packet)
   struct playlock_cpacket        *packet;
{
   lock_player(packet->pnum);
}

void 
handleDetMReq(packet)
   struct det_mytorp_cpacket      *packet;
{
   struct torp                    *atorp;
   int                             t;
   t = ntohs(packet->tnum);
   if (t < 0 || t >= MAXPLAYER * MAXTORP)
      return;
   atorp = &torps[t];
#ifdef MINES
   if (atorp->t_mine)
      return;			/* det doesn't remove our mines */
#endif
#ifdef SHORT_PACKETS
   if (send_short) {
      for (t = 0, atorp = &torps[me->p_no * MAXTORP]; t < MAXTORP; t++, atorp++) {
	 if (atorp->t_status == TMOVE || atorp->t_status == TSTRAIGHT)
	    atorp->t_status = TOFF;
      }
      return;
   } else
#endif

   if (atorp->t_owner != me->p_no)
      return;
   if (atorp->t_status == TMOVE || atorp->t_status == TSTRAIGHT) {
      atorp->t_status = TOFF;
   }
}

void 
handleTractorReq(packet)
   struct tractor_cpacket         *packet;
{
   int                             target;
   struct player                  *player;
   if (weaponsallowed[WP_TRACTOR] == 0) {
      warning(1, "Tractor beams haven't been invented yet.", 0);
      return;
   }
   target = packet->pnum;
   if (packet->state == 0) {
      me->p_flags &= ~(PFTRACT | PFPRESS);
      return;
   }
   if (me->p_flags & PFCLOAK) {
      warning(0, "Weapons's Officer:  Cannot tractor while cloaked, sir!", 1);
      return;
   }
   if (target < 0 || target >= MAXPLAYER || target == me->p_no)
      return;
   player = &players[target];
   if (player->p_flags & PFCLOAK)
      return;
   if (hypot((double) me->p_x - player->p_x,
	     (double) me->p_y - player->p_y) <
       ((double) TRACTDIST) * me->p_ship.s_tractrng) {
      if (player->p_flags & PFDOCK) {
#ifdef SB_DOCK
	 players[player->p_docked].p_port[player->p_portdocked] = VACANT;
	 players[player->p_docked].p_numdocked--;
#else
	 players[player->p_docked].p_port[player->p_port[0]] = VACANT;
	 players[player->p_docked].p_docked--;
#endif
      }
      if (me->p_flags & PFDOCK) {
#ifdef SB_DOCK
	 players[me->p_docked].p_numdocked--;
	 players[me->p_docked].p_port[me->p_portdocked] = VACANT;
#else
	 players[me->p_docked].p_docked--;
	 players[me->p_docked].p_port[me->p_port[0]] = VACANT;
#endif
      }
      player->p_flags &= ~(PFORBIT | PFDOCK);
      me->p_flags &= ~(PFORBIT | PFDOCK);
      me->p_tractor = target;
      me->p_flags |= PFTRACT;
   } else {
      warning(0,
	      "Weapon's Officer:  Vessel is out of range of our tractor beam.", 2);
   }
}

void 
handleRepressReq(packet)
   struct repress_cpacket         *packet;
{
   int                             target;
   struct player                  *player;
   if (weaponsallowed[WP_TRACTOR] == 0) {
      warning(1, "Tractor beams haven't been invented yet.", 0);
      return;
   }
   target = packet->pnum;
   if (packet->state == 0) {
      me->p_flags &= ~(PFTRACT | PFPRESS);
      return;
   }
   if (me->p_flags & PFCLOAK) {
      warning(0, "Weapons's Officer:  Cannot tractor while cloaked, sir!", 1);
      return;
   }
   if (target < 0 || target >= MAXPLAYER || target == me->p_no)
      return;
   player = &players[target];
   if (player->p_flags & PFCLOAK)
      return;
   if (hypot((double) me->p_x - player->p_x,
	     (double) me->p_y - player->p_y) <
       ((double) TRACTDIST) * me->p_ship.s_tractrng) {
      if (player->p_flags & PFDOCK) {
#ifdef SB_DOCK
	 players[player->p_docked].p_port[player->p_portdocked] = VACANT;
	 players[player->p_docked].p_numdocked--;
#else
	 players[player->p_docked].p_port[player->p_port[0]] = VACANT;
	 players[player->p_docked].p_docked--;
#endif
      }
      if (me->p_flags & PFDOCK) {
#ifdef SB_DOCK
	 players[me->p_docked].p_numdocked--;
	 players[me->p_docked].p_port[me->p_portdocked] = VACANT;
#else
	 players[me->p_docked].p_docked--;
	 players[me->p_docked].p_port[me->p_port[0]] = VACANT;
#endif
      }
      player->p_flags &= ~(PFORBIT | PFDOCK);
      me->p_flags &= ~(PFORBIT | PFDOCK);
      me->p_tractor = target;
      me->p_flags |= (PFTRACT | PFPRESS);
   } else {
      warning(0,
	      "Weapon's Officer:  Vessel is out of range of our tractor beam.", 2);
   }
}

void 
sendMotdLine(line)
   char                           *line;
{
   struct motd_spacket             motdPacket;
   motdPacket.type = SP_MOTD;
   strncpy(motdPacket.line, line, 80);
   motdPacket.line[79] = '\0';
   sendClientPacket((CVOID) & motdPacket);
}

void 
handleCoupReq(packet)
   struct coup_cpacket            *packet;
{
#ifdef MINES
   /* we can assume that starbases won't ever throw a coup */
   if (mines && me->p_ship.s_type == STARBASE) {
      mine();
      return;
   }
#endif
   coup();
}

void 
handleRefitReq(packet)
   struct refit_cpacket           *packet;
{
   do_refit(packet->ship);
}

#ifdef SHORT_PACKETS

void
handleSMessageReq(packet)
   struct mesg_s_cpacket          *packet;
{
   /* If someone would delete the hardcoded things in handleMessageReq */
   /* like     packet->mesg[69]='\0';   */
   /* we could give handleMessageReq the packet without copying */
   /* But i have no time  HW 04/6/93 */
   /* I have rewritten handleMessageReq, You do not need the copy. BUT */
   /* You cannot send messages with that packet !!!! HW */
   /* In case you want that uncomment the lines below */
   /*
    * struct mesg_cpacket mesPacket; mesPacket.type =  CP_MESSAGE;
    * mesPacket.group = packet->group; mesPacket.indiv  = packet->indiv;
    * strcpy(mesPacket.mesg, packet->mesg); handleMessageReq(& mesPacket);
    */
   /* I hope this was it */
   handleMessageReq((struct mesg_cpacket *) packet);
}
#endif

void 
handleMessageReq(packet)
   struct mesg_cpacket            *packet;
{
   char                            addrbuf[9];
   static long                     lasttime = 0;
   static int                      balance = 0;	/* make sure he doesn't get
						 * carried away */
   time_t                          thistime;
   int                             parseSpecial();
   /*
    * Some random code to make sure the player doesn't get carried away about
    * the number of messages he sends.  After all, he could try to jam
    * peoples communications if we let him.
    */
   thistime = time(NULL);
   if (lasttime != 0) {
      balance = balance - (thistime - lasttime);
      if (balance < 0)
	 balance = 0;
   }
   lasttime = thistime;
   if (balance >= 15) {
      warning(1, "Be quiet", 86);
      /*
      balance += 3;
      */
      if (balance > time(NULL) + 60) {
	 balance = time(NULL) + 60;
      }
      return;
   }
   balance += 3;
#ifdef SHORT_PACKETS
   if (packet->type != CP_S_MESSAGE)
      packet->mesg[69] = '\0';
#else
   packet->mesg[69] = '\0';
#endif

   /* send message while on waitq */
   if(!me){
      return;
   }
      
   sprintf(addrbuf, " %c%c->", teamlet[me->p_team], shipnos[me->p_no]);
   switch (packet->group) {
   case MALL:
      sprintf(addrbuf + 5, "ALL");
      break;
   case MGOD:
      sprintf(addrbuf + 5, "GOD");
      break;
   case (MTEAM|MDISTR):
   case MTEAM:
      if (packet->indiv != FED && packet->indiv != ROM &&
	  packet->indiv != KLI && packet->indiv != ORI)
	 return;
      sprintf(addrbuf + 5, teamshort[(int) packet->indiv]);
      break;
   case MINDIV:
      if (packet->indiv < 0 || packet->indiv >= MAXPLAYER)
	 return;
      if (players[packet->indiv].p_status == PFREE)
	 return;

      if (parseSpecial(packet))
	 return;
      if (ignored[(int) packet->indiv] & MINDIV) {
	 warning(1, "You are ignoring that player.  Message was not sent.", 88);
	 return;
      }
      sprintf(addrbuf + 5, "%c%c ", teamlet[players[packet->indiv].p_team],
	      shipnos[packet->indiv]);
      break;
   case (MINDIV|MCONFIG):
      clientVersion(packet);
      break;

   default:
      return;
   }
   pmessage2(packet->mesg, packet->indiv, packet->group, addrbuf, me->p_no);

   if (checkmessage) {
      if (check_mesgs(packet))
	 return;
   }
}

void 
clientVersion(packet)
   struct mesg_cpacket *packet;
{
#ifdef SYSV
#   define strdup(s)       strcpy((char *)malloc(strlen(s)+1), s)
#else
#ifndef mips
    extern char         *strdup();
#endif
#endif

   if (packet->mesg[0] != '@')
      return;

   cl_version = strdup(strchr(packet->mesg,'@') + 1);
}

void 
handleQuitReq(packet)
   struct quit_cpacket            *packet;
{
   me->p_flags |= PFSELFDEST;
   selfdest = me->p_updates + 100;
   /* 60 seconds to destruct a starbase */
   if (me->p_ship.s_type == STARBASE)
      selfdest = me->p_updates + 600;
   warning(1, "Self destruct initiated", 90);
}

void 
sendMaskPacket(mask)
   int                             mask;
{
   struct mask_spacket             maskPacket;
   maskPacket.type = SP_MASK;
   maskPacket.mask = mask;
   sendClientPacket((CVOID) & maskPacket);
}

void 
handleOptionsPacket(packet)
   struct options_cpacket         *packet;
{
   bcopy(packet->keymap, mystats->st_keymap, 96);
   mystats->st_flags = ntohl(packet->flags);
   keeppeace = (mystats->st_flags / ST_KEEPPEACE) & 1;
}

void 
handleSocketReq(packet)
   struct socket_cpacket          *packet;
{
   nextSocket = ntohl(packet->socket);
   userVersion = packet->version;
   userUdpVersion = packet->udp_version;
}

void 
handleByeReq(packet)
   struct bye_cpacket             *packet;
{
   noressurect = 1;
}

int 
checkVersion()
{
   struct badversion_spacket       packet;
   if (userVersion != SOCKVERSION) {
      packet.type = SP_BADVERSION;
      packet.why = 0;
      sendClientPacket((CVOID) & packet);
      flushSockBuf();
      return (0);
   }
   return (1);
}

void 
logEntry()
{
   FILE                           *logfile;
   long                           curtime;
   logfile = fopen(LOGFILENAME, "a");
   if (!logfile)
      return;
   curtime = time(NULL);
   fprintf(logfile, "Joining: %s (%s@%s) (%c), %s", me->p_name,
	   me->p_login, host,	/* ATM */
	   shipnos[me->p_no], ctime(&curtime));
   fclose(logfile);
}

/* return true if you should eat message */
int 
parseSpecial(packet)
   struct mesg_cpacket            *packet;
{
   char                           *s;
   int                             who;
   int                             what;
   char                            buf[80];
   int                             noneflag;
   extern int                      ignored[];
   int                             previg;
#define FALSE 0
#define TRUE 1

   s = packet->mesg;

   who = packet->indiv;
   /* don't disable special "back door" */
   if (*s != ':' || (*s == ':' && *(s + 1) == ':'))
      return FALSE;

#ifdef CYBORGS
   if (who == me->p_no) {
      if (binconfirm == 1 || binconfirm == 2)
	 warning(1, "No cyborgs allowed in the game at this time.", -1);
      else if (binconfirm == 3) {
	 char                            buf[80];
	 char                            buf2[5];
	 int                             i;
	 int                             cybflag = 0;
	 strcpy(buf, "Possible cyborgs: ");
	 for (i = 0; i < MAXPLAYER; i++)
	    if ((players[i].p_status != PFREE) &&
		(players[i].p_flags & PFCYBORG)) {
	       sprintf(buf2, "%s ", players[i].p_mapchars);
	       strcat(buf, buf2);
	       cybflag = 1;
	       cybflag = 1;
	    }
	 if (!cybflag)
	    strcat(buf, "None");
	 warning(1, buf, -1);
      } else
	 warning(1, "No cyborg checking enabled.", -1);

      return TRUE;
   }
#endif

   previg = ignored[who];

   do {
      what = 0;
      switch (*(++s)) {
      case 'a':
      case 'A':
	 what = MALL;
	 break;
      case 't':
      case 'T':
	 what = MTEAM;
	 break;
      case 'i':
      case 'I':
	 what = MINDIV;
	 break;
      case '\0':
	 what = 0;
	 break;
      default:
	 what = 0;
	 /*
	  * new: we don't want to steal messages that happen to begin with
	  * ':' (like smiley :)
	  */
	 ignored[who] = previg;	/* undo any changes */
	 return FALSE;
	 /*
	  * break;
	  */
      }
      ignored[who] ^= what;
   } while (what != 0);

   strcpy(buf, "Ignore status for this player: ");
   noneflag = TRUE;
   if (ignored[who] & MALL) {
      strcat(buf, "All ");
      noneflag = FALSE;
   }
   if (ignored[who] & MTEAM) {
      strcat(buf, "Team ");
      noneflag = FALSE;
   }
   if (ignored[who] & MINDIV) {
      strcat(buf, "Indiv ");
      noneflag = FALSE;
   }
   if (noneflag)
      strcat(buf, "None");
   warning(1, buf, -1);
   return TRUE;
}

int 
gwrite(fd, wbuf, bytes)
   int                             fd;
   char                           *wbuf;
   register int                    bytes;
{
   long                            orig = bytes;
   register long                   n;
   while (bytes) {
      n = write(fd, wbuf, bytes);

      if (n < 0) {
	 if (errno == ENOBUFS) {
	    struct timeval                  to;
	    /*
	     * The man pages don't mention this as a possibility. Yet, it
	     * happens.  I guess I just wait for 1/10 sec, and continue?
	     */
	    /*
	     * I would use usleep() to do this, but this system ain't got
	     * it...
	     */
	    perror("write");
	    to.tv_sec = 0;
	    to.tv_usec = 100000;
	    select(0, NULL, NULL, NULL, &to);
	    continue;
	 } else if (errno == EINTR)
	    continue;
	 else {
	    char                            tempbuf[80];
	    if (fd == udpSock) {
	       /* do we want Hiccup code here? */
	       UDPDIAG(("Tried to write %d, 0x%x, %d (error %d)\n",
		  fd, (unsigned int) wbuf, bytes, errno));
	       printUdpInfo();
	       logmessage("UDP gwrite failed:");
	    }
	    sprintf(tempbuf, "Died in gwrite, n=%d, errno=%d", n, errno);
	    logmessage(tempbuf);
	    return (-1);
	 }
      }
      bytes -= n;
      wbuf += n;

#ifdef nodef
      if (bytes)
	 printf("incomplete write (%d left)\n", bytes);	/* QQ */
#endif
   }
   return (orig);
}

void 
handleDockingReq(packet)
   struct dockperm_cpacket        *packet;
{
   int                             i;
   if (me->p_ship.s_type == STARBASE) {
#ifdef SB_DOCK
      me->p_numdocked = 0;
#else
      me->p_docked = 0;
#endif
      for (i = 0; i < NUMPORTS; i++) {
	 if (me->p_port[i] != VACANT)
	    players[me->p_port[i]].p_flags &= ~PFDOCK;
	 me->p_port[i] = VACANT;
      }
      if (packet->state)
	 me->p_flags |= PFDOCKOK;
      else
	 me->p_flags &= ~PFDOCKOK;
   }
}

void 
handleReset(packet)
   struct resetstats_cpacket      *packet;
{
   if (packet->verify != 'Y')
      return;

   /*
    * Gee, they seem to want to reset their stats!  Here goes...
    */
   mystats->st_maxkills = 0.0;
   mystats->st_kills = 0;
   mystats->st_losses = 0;
   mystats->st_armsbomb = 0;
   mystats->st_planets = 0;
   mystats->st_ticks = 0;
   mystats->st_tkills = 0;
   mystats->st_tlosses = 0;
   mystats->st_tarmsbomb = 0;
   mystats->st_tplanets = 0;
   mystats->st_tticks = 1;
   mystats->st_rank = 0;
   mystats->st_sbkills = 0;
   mystats->st_sblosses = 0;
   mystats->st_sbticks = 0;
   mystats->st_sbmaxkills = 0.0;

fprintf(stderr, "%s (%s@%s):stats reset \n",
   me->p_name, me->p_login, me->p_monitor);
fflush(stderr);

   startTkills = me->p_stats.st_tkills;
   startTlosses = me->p_stats.st_tlosses;
   startTarms = me->p_stats.st_tarmsbomb;
   startTplanets = me->p_stats.st_tplanets;
   startTticks = me->p_stats.st_tticks;

   startSBkills = me->p_stats.st_sbkills;
   startSBlosses = me->p_stats.st_sblosses;
   startSBticks = me->p_stats.st_sbticks;
#ifdef BASEPRACTICE
   startTdami = me->p_stats.st_dami;
   startTdamr = me->p_stats.st_damr;
#endif

#ifdef BASEPRACTICE
   mystats->st_dami = 1.;
   mystats->st_damr = 1.;
#endif
}

void 
handleUpdatesReq(packet)
   struct updates_cpacket         *packet;
{
#ifndef DS
   struct itimerval                udt;
#endif
   
   if(packet)
      timerDelay = ntohl(packet->usecs);
   /* called from refit/enter */
   else if(me->p_ship.s_type != STARBASE)
      timerDelay = MAXUPDATES;

#ifdef FASTER_SB_MAXUPDATES
  if (me->p_ship.s_type == STARBASE){
    if (timerDelay < 100000)
      timerDelay = 100000;
   }
   else
#endif
      if (timerDelay < MAXUPDATES && me->p_ship.s_type != ATT){
	 char	buf[80];
	 timerDelay = MAXUPDATES;
	 sprintf(buf, "Maximum updates are %d per second.", 1000000/timerDelay);
	 warning(1, buf, -1);
      }

  if (timerDelay >= 1000000)
#ifdef DS
    timerDelay = 1000000;
#else
    timerDelay = 999999;
#endif

   
#ifdef DS
   me->p_timerdelay = timerDelay / 100000;
#else
   udt.it_interval.tv_sec = 0;
   udt.it_interval.tv_usec = timerDelay;
   udt.it_value.tv_sec = 0;
   udt.it_value.tv_usec = timerDelay;
   setitimer(ITIMER_REAL, &udt, 0);
#endif
}

void 
logmessage(string)
   char                           *string;
{
   FILE                           *fp;
   fp = fopen(LOGFILENAME, "a");
   if (fp) {
      fprintf(fp, "%s\n", string);
      fclose(fp);
   }
}

void 
handleReserved(packet)
   struct reserved_cpacket        *packet;
{
   struct reserved_cpacket         mycp;
   struct reserved_spacket         mysp;
#ifdef RSA
   struct rsa_key_spacket          rsp;
#endif
   register                        i, n;
   char                            (*serverNames)[16];
/*
printf("In handleReserved from %s@%s (tt: %d)\n", me->p_login, me->p_monitor,
      testtime);
fflush(stdout);
*/

   if (testtime == 1)
      return;
   if (bcmp(packet->data, testdata, RESERVED_SIZE) != 0) {
      printf("%s@%s: bad data\n", me->p_login, me->p_monitor);fflush(stdout);
      testtime = 1;
      return;
   }
#ifdef RSA
   if (!strncmp(packet->resp, RSA_VERSION, 3)) {
      /* This is an RSA type client */
      RSA_Client = 2;
      warning(1, RSA_VERSION, -1);
      if (!strncmp(packet->resp, RSA_VERSION, strlen("RSA v??"))) {
	 /* This is the right major version */
	 RSA_Client = 1;
	 makeRSAPacket(&rsp);
	 bcopy(rsp.data, testdata, KEY_SIZE);
/*
	 printf("%s@%s: sending RSA packet\n", 
	    me->p_login, me->p_monitor);fflush(stdout);
*/
	 sendClientPacket((CVOID) & rsp);
	 return;
      }
      printf("%s@%s: bad RSA version \"%s\"\n", 
	 me->p_login, me->p_monitor, packet->resp);
      fflush(stdout);
      testtime = 1;
      return;
   }
#endif
   bcopy(testdata, mysp.data, RESERVED_SIZE);

   n = gethostnames(&serverNames);
   for (i = 0; i < n; i++) {
      encryptReservedPacket(&mysp, &mycp, serverNames[i], me->p_no);
      if (bcmp(packet->resp, mycp.resp, RESERVED_SIZE) == 0) {
	 /* Use .sysdef CONFIRM flag to allow old style clients. */
	 if (binconfirm == 2 || binconfirm == 3)
	    testtime = 0;
	 else
	    testtime = 1;
	 return;
      }
   }
   testtime = 1;
   return;
}

#ifdef RSA
void 
handleRSAKey(packet)
   struct rsa_key_cpacket         *packet;
{
   struct rsa_key_spacket          mysp;
   struct rsa_key		   key;
   register                        i, n;
   char                            (*serverNames)[16];
   FILE				   *logfile;
   time_t			   curtime;

/*
printf("In handleRSAKey from %s@%s (tt: %d)\n", me->p_login, me->p_monitor,
   testtime);
fflush(stdout);
*/

   if (testtime == 1)
      return;
   if (RSA_Client != 1)
      return;
   bcopy(testdata, mysp.data, KEY_SIZE);

   n = gethostnames(&serverNames);
   for (i = 0; i < n; i++) {
      key.key_name[0] = 0;
      if (decryptRSAPacket(&mysp, packet, serverNames[i], &key) == 0) {

	 if((logfile = fopen(LOGFILENAME, "a"))){

	    curtime = time(NULL);
	    fprintf(logfile, "[RSA] %s: %s / %s\n", 
	       key.key_name, key.client_type, key.architecture);
	    fprintf(logfile, "      -- (%s) %s@%s at %s", 
	       me->p_name, login, host, ctime(&curtime));
	    fclose(logfile);
	 }
#ifdef nodef
	 /* add rsa type to message log */
	 sprintf(format,"RSA->ALL %%s --> %%.%ds.--- %%.%ds.",
		     KEY_SIZE, KEY_SIZE);
	 sprintf(buf,format,me->p_name,key.client_type,key.architecture);
	 pmessage(buf,0,MALL,"");
#endif

	 sprintf(RSA_client_type,"%s / %s (%s)", 
	    key.client_type, key.architecture, key.key_name);
	 RSA_client_type[70]='\0';    /* was 64? */

#ifdef FEATURE
	 /* NOTE: this assumes client will send features before RSA */
	 if(!F_client_feature_packets)
	    F_tellClient(key.client_type);	/* LAB */
#endif

	 testtime = 0;
	 return;
      }
   }
   if(key.key_name[0]){
      curtime = time(NULL);
      fprintf(stderr, "[RSA] Incorrect verification for key \"%s\"\n",
              key.key_name);
      fprintf(stderr, "      -- (%s) %s@%s at %s", me->p_name, login, host,
                                              ctime(&curtime));
      fflush(stderr);
   }
   else{
      fprintf(stderr, "[RSA] Key not found for %s@%s.\n", login, host);
      fflush(stderr);
   }

   testtime = 1;
   return;
}
#endif

#ifdef INCLUDE_SCAN
void 
handleScan(packet)		/* ATM */
   struct scan_cpacket            *packet;
{
   struct scan_spacket             response;
   struct player                  *pp;

   bzero(&response, sizeof(struct scan_spacket));
   response.type = SP_SCAN;
   response.pnum = packet->pnum;
   if (!weaponsallowed[WP_SCANNER]) {
      warning(1, "Scanners haven't been invented yet", 91);
      response.success = 0;
   } else {
      response.success = scan(packet->pnum);

      if (response.success) {
	 /* fill in all the goodies */
	 pp = &players[packet->pnum];
	 response.p_fuel = htonl(pp->p_fuel);
	 response.p_armies = htonl(pp->p_armies);
	 response.p_shield = htonl(pp->p_shield);
	 response.p_damage = htonl(pp->p_damage);
	 response.p_etemp = htonl(pp->p_etemp);
	 response.p_wtemp = htonl(pp->p_wtemp);
      }
   }
   sendClientPacket((CVOID) & response);
}
#endif

void 
handlePingResponse(packet)
   struct ping_cpacket            *packet;
{
   char                            buf[80];
   /* client requests pings by sending pingme == 1 on TCP socket */

   if (rsock == sock) {
      if (!ping && packet->pingme == 1) {
	 ping = 1;
#if PING_DEBUG >= 1
	 fprintf(stderr, "ping on for %s@%s\n", me->p_name, me->p_monitor);
	 fflush(stderr);
#endif
	 if(ping_freq)
	    sprintf(buf, "Server sending PING packets at %d second intervals",
		    ping_freq);
	 else
	    sprintf(buf, "Server sending PING packets on every update.");
	 warning(1, buf, -1);
	 return;
      }
      /* client says stop */
      else if (ping && !packet->pingme) {
	 ping = 0;
#if PING_DEBUG >= 1
	 fprintf(stderr, "ping off for %s@%s\n", me->p_name, me->p_monitor);
	 fflush(stderr);
#endif
	 warning(1, "Server no longer sending PING packets.", -1);
	 return;
      }
   }
   pingResponse(packet);	/* ping.c */
}

#ifdef BASEPRACTICE
/* these are sent by the robots when a parameter changes */
void 
handleOggV(packet)
   struct oggv_cpacket            *packet;
{
   me->p_df = packet->def;
   me->p_tg = packet->targ;

   if (me->p_df > 100)
      me->p_df = 100;

   me->p_flags |= PFBPROBOT;

#ifdef DEBUG
   printf("%c%c (%s) %d, targ: %d\n",
	  me->p_mapchars[0], me->p_mapchars[1],
	  me->p_name, me->p_df, me->p_tg - 1);
   fflush(stdout);
#endif
}
#endif				/* BASEPRACTICE */

#ifdef FEATURE

void
handleFeature(cpacket)

   struct feature_cpacket	*cpacket;
{
   struct feature_spacket	spacket;

   F_client_feature_packets = 1;	/* XX */

   cpacket->value = ntohl(cpacket->value);
   getFeature(cpacket, &spacket);
   sendFeature(&spacket);

   if(F_motd_bitmaps){
      /* send this here because sending it with the
	 regular motd is too early to know about the feature */
      F_motd_bitmaps = 0;
      doMotdPics();
   }
}

void
sendFeature(packet)

   struct feature_spacket	*packet;
{
   packet->type = SP_FEATURE;
   packet->value = htonl(packet->value);
   sendClientPacket((CVOID) packet);
}

#endif /* FEATURE */
   

/*
 * ---------------------------------------------------------------------------
 * Strictly UDP from here on
 * ---------------------------------------------------------------------------
 * */

void 
handleUdpReq(packet)
   struct udp_req_cpacket         *packet;
{
   struct udp_reply_spacket        response;
   int                             mode;
   response.type = SP_UDP_REPLY;

   if (packet->request == COMM_VERIFY) {
      /* this request should ONLY come through the UDP connection */
      if (commMode == COMM_UDP) {
	 UDPDIAG(("Got second verify from %s; resending server verify\n",
		  me->p_name));
	 response.reply = SWITCH_VERIFY;
	 goto send;
      }
      UDPDIAG(("Receieved UDP verify from %s\n", me->p_name));
      UDPDIAG(("--- UDP connection established to %s\n", me->p_name));
#ifdef BROKEN
      /*
       * warning(1, "WARNING: BROKEN mode is enabled", -1);
       */
#endif

      sequence = 1;		/* reset sequence numbers */
      commMode = COMM_UDP;	/* at last */
      udpMode = MODE_SIMPLE;	/* just send one at a time */

      /* note that we don't NEED to send a SWITCH_VERIFY packet; the client */
      /* will change state when it receives ANY packet on the UDP connection */
      /* (this just makes sure that it gets one) */
      /* (update: recvfrom() currently tosses the first packet it gets...)  */
      response.reply = SWITCH_VERIFY;
      goto send;
      /* return; */
   }
   if (packet->request == COMM_MODE) {
      /* wants to switch modes; mode is in "conmode" */
      mode = packet->connmode;
      if (mode < MODE_TCP || mode > MODE_DOUBLE) {
	 warning(1, "Server can't do that UDP mode", -1);
	 UDPDIAG(("Got bogus request for UDP mode %d from %s\n",
		  mode, me->p_name));
      } else {
	 /* I don't bother with a reply, though it can mess up the opt win */
	 switch (mode) {
	 case MODE_TCP:
	    warning(1, "Server will send with TCP only", 94);
	    break;
	 case MODE_SIMPLE:
	    warning(1, "Server will send with simple UDP", 95);
	    break;
	 case MODE_FAT:
#ifdef nodef
	    mode = MODE_SIMPLE;
	    warning(1, "FAT udp temporarily disabled.", -1);
	    break;
#ifdef SHORT_PACKETS
	    if (send_short) {
	       swarning(TEXTE, 96, 0);
	       mode = MODE_SIMPLE;
	       break;
	    }
#endif

	    if(!send_short){
	       warning(1, "FAT udp only enabled with short-packets.", -1);
	       mode = MODE_SIMPLE;
	       break;
	    }
#endif

	    warning(1, "Server will send with fat UDP; sent full update", -1);
#ifdef SAVE_MEM
#define fmem_alloc(v, size)	v=(FAT_NODE *)malloc(sizeof(FAT_NODE)*(size));\
				bzero(v, sizeof(FAT_NODE)*(size));\
				if(!v) {perror("malloc"); exitGame(1); }
	    /* will only be done once */
	    if (!fat_kills) {

	       fmem_alloc(fat_kills, MAXPLAYER);
	       fmem_alloc(fat_torp_info, MAXPLAYER * MAXTORP);
	       fmem_alloc(fat_phaser, MAXPLAYER);
	       fmem_alloc(fat_plasma_info, MAXPLAYER * MAXPLASMA);
	       fmem_alloc(fat_planet, MAXPLANETS);
	       fmem_alloc(fat_flags, MAXPLAYER);
	       fmem_alloc(fat_hostile, MAXPLAYER);
	       udpMode = MODE_FAT;
	       initClientData();
	    }
#undef fmem_alloc
#endif
	    forceUpdate();
	    V_UDPDIAG(("Sending full update to %s\n", me->p_name));
	    break;
	 case MODE_DOUBLE:
	    warning(1, "Request for double UDP DENIED (set to simple)", -1);
	    mode = MODE_SIMPLE;
	    break;
	 }

	 udpMode = mode;
	 UDPDIAG(("Switching %s to UDP mode %d\n", me->p_name, mode));
      }
      return;
   }
   if (packet->request == COMM_UPDATE) {
      /* client wants a FULL update */
      V_UDPDIAG(("Sending full update to %s\n", me->p_name));
      forceUpdate();

      return;
   }
   UDPDIAG(("Received request for %s mode from %s\n",
	    (packet->request == COMM_TCP) ? "TCP" : "UDP", me->p_name));
   if (packet->request == commMode) {
      /* client asking to switch to current mode */
      if (commMode == COMM_UDP) {
	 /*
	  * client must be confused... whatever the cause, he obviously isn't
	  * connected to us, so we better drop out end and retry.
	  */
	 UDPDIAG(("Rcvd UDP req from %s while in UDP mode; dropping old\n",
		  me->p_name));
	 closeUdpConn();
	 commMode = COMM_TCP;
	 /* ...and fall thru to the UDP request handler */
      } else {
	 /*
	  * Again, client is confused.  This time there's no damage though.
	  * Just tell him that he succeeded.  Could also happen if the client
	  * tried to connect to our UDP socket but failed, and decided to
	  * back off.
	  */
	 UDPDIAG(("Rcvd TCP req from %s while in TCP mode\n", me->p_name));

	 response.reply = SWITCH_TCP_OK;
	 sendClientPacket((CVOID) & response);

	 if (udpSock >= 0) {
	    closeUdpConn();
	    UDPDIAG(("Closed UDP socket\n"));
	 }
	 return;
      }
   }
   /* okay, we have a request to change modes */
   if (packet->request == COMM_UDP) {
      udpClientPort = ntohl(packet->port);	/* where to connect to */
      if (!udpAllowed) {
	 UDPDIAG(("Rejected UDP request from %s\n", me->p_name));
	 response.reply = SWITCH_DENIED;
	 response.port = htons(0);
	 goto send;
      } else {
	 if (userUdpVersion != UDPVERSION) {
	    char                            buf[80];
	    sprintf(buf, "Server UDP is v%.1f, client is v%.1f",
		    (float) UDPVERSION / 10.0,
		    (float) userUdpVersion / 10.0);
	    warning(1, buf, -1);
	    UDPDIAG(("%s (rejected %s)\n", buf, me->p_name));
	    response.reply = SWITCH_DENIED;
	    response.port = htons(1);
	    goto send;
	 }
	 if (udpSock >= 0) {
	    /* we have a socket open, but the client doesn't seem aware */
	    /* (probably because our UDP verify got lost down the line) */
	    UDPDIAG(("Receieved second request from %s, reconnecting\n",
		     me->p_name));
	    closeUdpConn();
	 }
	 /* (note no openUdpConn(); we go straight to connect) */
	 if (connUdpConn() < 0) {
	    response.reply = SWITCH_DENIED;
	    response.port = 0;
	    goto send;
	 }
	 UDPDIAG(("Connected UDP socket (%d:%d) for %s\n", udpSock,
		  udpLocalPort, me->p_name));

	 /* we are now connected to the client, but he's merely bound */
	 /* don't switch to UDP mode yet; wait until client connects */
	 response.reply = SWITCH_UDP_OK;
	 response.port = htonl(udpLocalPort);

	 UDPDIAG(("packet->connmode = %d\n", packet->connmode));
	 if (packet->connmode == CONNMODE_PORT) {
	    /* send him our port # so he can connect to us */
	    goto send;
	 } else {
	    /* send him a packet; he'll get port from recvfrom() */
	    if (gwrite(udpSock, (char *) &response, sizeof(response)) !=
		sizeof(response)) {
	       UDPDIAG(("Attempt to send UDP packet failed; using alt\n"));
	    }
	    goto send;
	 }
      }
   } else if (packet->request == COMM_TCP) {
      closeUdpConn();
      commMode = COMM_TCP;
      response.reply = SWITCH_TCP_OK;
      response.port = 0;
      UDPDIAG(("Closed UDP socket for %s\n", me->p_name));
      goto send;
   } else {
      fprintf(stderr, "ntserv: got weird UDP request (%d)\n",
	      packet->request);
      fflush(stderr);
      return;
   }
 send:
   sendClientPacket((CVOID) & response);
}


int 
connUdpConn()
{
   struct sockaddr_in              addr;
   int                             len;
   if (udpSock > 0) {
      fprintf(stderr, "ntserv: tried to open udpSock twice\n");
      fflush(stderr);
      return (0);		/* pretend we succeeded (this could be bad) */
   }
   if (udpbufptr != udpbuf) {
      udpbufptr = udpbuf;	/* clear out any old data */
      sequence--;		/* we just killed a sequence packet */
   }
   if ((udpSock = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
      perror("ntserv: unable to create DGRAM socket");
      return (-1);
   }
   addr.sin_family = AF_INET;
   addr.sin_addr.s_addr = remoteaddr;	/* addr of our client */
   addr.sin_port = htons(udpClientPort);	/* client's port */

   if (connect(udpSock, (struct sockaddr *) & addr, sizeof(addr)) < 0) {
      perror("ntserv: connect to client UDP port");
      UDPDIAG(("Unable to connect() to %s on port %d\n", me->p_name,
	       udpClientPort));
      close(udpSock);
      udpSock = -1;
      return (-1);
   }
   UDPDIAG(("connect to %s's port %d on 0x%x succeded\n",
	    me->p_name, udpClientPort, remoteaddr));

   /* determine what our port is */
   len = sizeof(addr);
   if (getsockname(udpSock, (struct sockaddr *) & addr, &len) < 0) {
      perror("netrek: unable to getsockname(UDP)");
      UDPDIAG(("Can't get our own socket; connection failed\n"));
      close(udpSock);
      udpSock = -1;
      return (-1);
   }
   udpLocalPort = (int) ntohs(addr.sin_port);

   if (udpAllowed > 2)		/* verbose debug mode? */
      printUdpInfo();

   return (0);
}

int 
closeUdpConn()
{
   V_UDPDIAG(("Closing UDP socket\n"));
   if (udpSock < 0) {
      fprintf(stderr, "ntserv: tried to close a closed UDP socket\n");
      fflush(stderr);
      return (-1);
   }
   shutdown(udpSock, 2);	/* wham */
   close(udpSock);		/* bam */
   udpSock = -1;		/* (nah) */

   return (0);
}

/* used for debugging */
void 
printUdpInfo()
{
   struct sockaddr_in              addr;
   int                             len;
   len = sizeof(addr);
   if (getsockname(udpSock, (struct sockaddr *) & addr, &len) < 0) {
      perror("printUdpInfo: getsockname");
      return;
   }
   UDPDIAG(("LOCAL: addr=0x%x, family=%d, port=%d\n", addr.sin_addr.s_addr,
	    addr.sin_family, ntohs(addr.sin_port)));

   if (getpeername(udpSock, (struct sockaddr *) & addr, &len) < 0) {
      perror("printUdpInfo: getpeername");
      return;
   }
   UDPDIAG(("PEER : addr=0x%x, family=%d, port=%d\n", addr.sin_addr.s_addr,
	    addr.sin_family, ntohs(addr.sin_port)));
}

void 
handleSequence()
{
   /* we don't currently deal with sequence numbers from clients */
}

/*
 * If we're in UDP mode, add a sequence number to the transmission buffer.
 * Returns the #of bytes inserted.
 *
 * This will add a sequence # to transmissions on either channel.  However, the
 * current implementation doesn't put sequences on TCP transmissions because
 * mixed TCP packets and UDP packets rarely arrive in the order in which they
 * were sent.
 */

int 
addSequence(outbuf)
   char                           *outbuf;
{
   struct sequence_spacket        *ssp;
   if (commMode != COMM_UDP || udpMode == MODE_TCP)
      return (0);

   packets_sent++;

   ssp = (struct sequence_spacket *) outbuf;
   ssp->type = SP_SEQUENCE;
   ssp->sequence = htons((unsigned short) sequence);
   sequence++;

   return (sizeof(struct sequence_spacket));
}


/*
 * This is a truncated version of initClientData().  Note that it doesn't
 * explicitly reset all the fat UDP stuff; sendClientData will take care of
 * that by itself eventually.
 *
 * Only semi-critical data is sent, with a few exceptions for non-critical data
 * which would be nice to update (stats, kills, player posn, etc). The
 * critical stuff can't be lost, so there's no point in resending it.
 *
 * (Since the fat data begins in an unqueued state, forceUpdate() should be
 * called immediately after switching to fat mode.  This guarantees that
 * every packet will end up on a queue.  The only real reason for doing this
 * is so that switching to fat mode will clear up your display and keep it
 * cleared; otherwise you could have torps floating around forever because
 * the packet for them isn't on a queue.  Will it reduce the effectiveness of
 * fat UDP? No, because as soon as the player hits the "update all" key it's
 * gonna happen anyway...)
 */

void 
forceUpdate()
{
   static time_t                   lastone = 0;
   time_t                          now;
   int                             i;
#ifdef SHORT_PACKETS

   /* WE must do this or change the torp implementation ( It is now changed ) */
   /* But it should be no problem because normally NO packet is send */
   if (send_short) {
      for (i = 0; i < MAXPLAYER * MAXTORP; i++) {
	 clientTorpsInfo[i].status = TFREE;	/* That's the trick */
      }
      /* And again */
      for (i = 0; i < MAXPLAYER * MAXPLASMA; i++)
	 clientPlasmasInfo[i].status = PTFREE;
      for (i = 0; i < MAXPLAYER; i++)
	 clientPhasers[i].status = PHFREE;
   }
#endif

   now = time(0);
   if (now - lastone < UDP_UPDATE_WAIT) {
#ifdef SHORT_PACKETS
      spk_update_sall = 0;	/* Back to default */
      spk_update_all = 0;
#endif
      warning(1, "Update request DENIED (chill out!)", 98);
      return;
   }
   lastone = now;

   /* clientDead=0; */
#ifdef SHORT_PACKETS
   if (!spk_update_sall) {
#endif

      for (i = 0; i < MAXPLAYER; i++) {
#ifdef SHORT_PACKETS
	 if (spk_update_all && players[i].p_status == PFREE)
	    continue;		/* Do not u pdate free Slots */
#endif
	 clientHostile[i].hostile = -1;
#ifdef SHORT_PACKETS
	 if (!spk_update_all)
	    clientStats[i].losses = -1;
#else
	 clientStats[i].losses = -1;	/* (non-critical, but nice) */
#endif
	 /* clientLogin[i].rank= -1;		(critical) */
	 /* clientPlayersInfo[i].shiptype= -1;	(critical) */
	 /* clientPStatus[i].status= -1;		(critical) */
#ifdef wait_until_later
	 /* NEW -- provide backup in case of screwup with UDP/TCP */
	 clientPStatus[i].status =  -1;
#endif
	 clientPlayers[i].x = htonl(-1);	/* (non-critical, but nice) */
#ifdef SHORT_PACKETS
	 if (!send_short)
	    clientPhasers[i].status = -1;
#else
	 clientPhasers[i].status = -1;
#endif
	 clientKills[i].kills = htonl(-1);	/* (non-critical, but nice) */
	 clientFlags[i].flags = htonl(-1);
	 mustUpdate[i] = 0;
      }
#ifdef SHORT_PACKETS
   }				/* For Player for */
   if (!send_short) {
#endif

      for (i = 0; i < MAXPLAYER * MAXTORP; i++) {
	 clientTorpsInfo[i].status = -1;
	 /* clientTorps[i].x= -1;			(non-critical) */
      }
      for (i = 0; i < MAXPLAYER * MAXPLASMA; i++) {
	 clientPlasmasInfo[i].status = -1;
	 /* clientPlasmas[i].x= -1;			(non-critical) */
      }
#ifdef SHORT_PACKETS
   }
#endif

   for (i = 0; i < MAXPLANETS; i++) {
      clientPlanets[i].info |= me->p_team;
      clientPlanets[i].armies = htonl(-1);
      /* clientPlanetLocs[i].x= htonl(-1);	(critical) */
   }
   /* msgCurrent=(mctl->mc_current+1) % MAXMESSAGE; */
   clientSelf.pnum = -1;
#ifdef SHORT_PACKETS
   if (spk_update_sall) {	/* Small update */
      for (i = 0; i < MAXPLAYER; i++) {
	 if (players[i].p_status == PFREE)
	    continue;		/* Do not update free Slots */
	 clientKills[i].kills = htonl(-1);	/* (non-critical, but ni ce) */
	 mustUpdate[i] = 0;
      }
   }
   clientSelfShort.pnum = -1;
   clientSelfShip.damage = -1;
   spk_update_sall = 0;		/* Back to default */
   spk_update_all = 0;
#endif
}

/*
 * ---------------------------------------------------------------------------
 * Fat City
 * ---------------------------------------------------------------------------
 * */

/*
 * This updates the "fat" tables; it's called from sendClientData().
 */
static void 
updateFat(packet)
   /* Pick a random type for the packet */
   struct player_spacket          *packet;
{
   FAT_NODE                       *fatp;
   struct kills_spacket           *kp;
   struct torp_info_spacket       *tip;
   struct phaser_spacket          *php;
   struct plasma_info_spacket     *pip;
   /* struct you_spacket *yp; */
   /* struct status_spacket *sp; */
   struct planet_spacket          *plp;
   struct flags_spacket           *fp;
   struct hostile_spacket         *hp;
   int                             idx;
   /* step 1 : find the FAT_NODE for this packet */
   switch (packet->type) {
   case SP_KILLS:
      kp = (struct kills_spacket *) packet;
      idx = (int) kp->pnum;
      fatp = &fat_kills[idx];
      break;
   case SP_TORP_INFO:
      tip = (struct torp_info_spacket *) packet;
      idx = (int) ntohs(tip->tnum);
      fatp = &fat_torp_info[idx];
      break;
   case SP_PHASER:
      php = (struct phaser_spacket *) packet;
      idx = (int) php->pnum;
      fatp = &fat_phaser[idx];
      break;
   case SP_PLASMA_INFO:
      pip = (struct plasma_info_spacket *) packet;
      idx = (int) ntohs(pip->pnum);
      fatp = &fat_plasma_info[idx];
      break;
   case SP_YOU:
      /* yp = (struct you_spacket *) packet; */
      fatp = &fat_you;
      break;
   case SP_STATUS:
      /* sp = (struct status_spacket *) packet; */
      fatp = &fat_status;
      break;
   case SP_PLANET:
      plp = (struct planet_spacket *) packet;
      idx = plp->pnum;
      fatp = &fat_planet[idx];
      break;
   case SP_FLAGS:
      fp = (struct flags_spacket *) packet;
      idx = (int) fp->pnum;
      fatp = &fat_flags[idx];
      break;
   case SP_HOSTILE:
      hp = (struct hostile_spacket *) packet;
      idx = (int) hp->pnum;
      fatp = &fat_hostile[idx];
      break;
#ifdef SHORT_PACKETS
    case SP_S_YOU:
      fatp = &fat_s_you;
      break;
    case SP_S_PLANET:
      fat_s_planet.pkt_size = sizes[SP_S_PLANET];
      fatp = &fat_s_planet;
      break;
#endif

   default:
      fprintf(stderr, "Fat error: bad semi-critical type in updateFat\n");
      fflush(stderr);
      return;
   }

   if (fatp->packet != (PTR) packet) {
      fprintf(stderr, "Fat error: fatp->packet=0x%.8lx, packet=0x%.8lx\n",
	      (long unsigned int) fatp->packet, (long unsigned int) packet);
      fflush(stderr);
      return;
   }
   /* step 2 : move this dude to temporary list 0 */
   dequeue(fatp);
   enqueue(fatp, 0);
}

/*
 * This fattens up the transmission, adding up to MAX_FAT_DATA bytes.  The
 * packets which get added will be moved to a higher queue, giving them less
 * priority for next time.  Note that they are added to a parallel temporary
 * list (otherwise they'd could be sent several times in the same
 * transmission as the algorithm steps through the queues), and merged later
 * on.
 *
 * Packets are assigned from head to tail, on a first-fit basis.  If a
 * semi-critical packet is larger than MAX_FAT_DATA, this routine will never
 * send it, but it will skip around it.
 *
 * This routine is called from flushSockBuf, before the transmission is sent.
 *
 * A possible improvement is to have certain packets "expire", never to be seen
 * again.  This way we don't keep resending torp packets for players who are
 * long dead.  This raises the possibility that the dead player's torps will
 * never go away though.
 */

static int 
fatten()
{
   int                             bytesleft;
   FAT_NODE                       *fatp, *nextfatp;
   int                             list;
#ifdef FATDIAG
   printf("--- fattening\n");
#endif
   bytesleft = MAX_FAT_DATA;
   for (list = 0; list < MAX_FAT_LIST; list++) {
      fatp = fatlist[list].head;
      while (fatp != NULL) {
	 nextfatp = fatp->prev;	/* move toward tail */
#ifdef FATDIAG
	 if (nextfatp == fatp) {
	    printf("Hey!  nextfatp == fatp!\n");
	    abort();
	 }
#endif
	 if (fatp->pkt_size < bytesleft) {
	    /* got one! */
	    bcopy(fatp->packet, udpbufptr, fatp->pkt_size);
	    udpbufptr += fatp->pkt_size;
	    bytesleft -= fatp->pkt_size;

	    packets_sent++;

	    /* move the packet to a higher queue (if there is one) */
	    dequeue(fatp);
	    if (list + 1 == MAX_FAT_LIST) {
	       /* enqueue(fatp, list); *//* keep packets on high queue? */
	    } else {
	       enqueue(fatp, list + 1);
	    }

	    /* done yet? */
	    if (bytesleft < MAX_NONFAT)
	       goto done;	/* don't waste time searching anymore */
	 }
	 fatp = nextfatp;
      }
   }

 done:
   /* at this point, we either filled with fat or ran out of queued packets */
#ifdef FATDIAG
   printf("--- done, added %d grams of fat\n", MAX_FAT_DATA - bytesleft);
#endif
   V_UDPDIAG(("- Added %d grams of fat\n", MAX_FAT_DATA - bytesleft));
   return (0);			/* some compilers need something after a goto */
}


/*
 * Remove a FAT_NODE from a queue.  If it's at the head or the tail of a
 * list, then we need to figure out which list it's in and update the head or
 * tail pointer.  It's easier to go searching than to maintain a queue number
 * in every FAT_NODE.
 *
 * This routine looks too complex... there must be a simpler way to do this.
 */

static void 
dequeue(fatp)
   FAT_NODE                       *fatp;
{
   int                             i;
#ifdef V_FATDIAG
   for (i = 0; i < MAX_FAT_LIST; i++) {
      printf("fatlist[i].head = 0x%.8lx tail = 0x%.8lx\n",
	     fatlist[i].head, fatlist[i].tail);
      if ((fatlist[i].head == NULL && fatlist[i].tail != NULL) ||
	  (fatlist[i].head != NULL && fatlist[i].tail == NULL)) {
	 printf("before!\n");
	 abort();
      }
   }
#endif


   if (fatp->next == NULL) {
      /* it's at the head or not in a queue */
      for (i = 0; i < MAX_FAT_LIST; i++) {
	 if (fatlist[i].head == fatp) {
	    fatlist[i].head = fatp->prev;	/* move head back */
	    if (fatlist[i].head != NULL)
	       (fatlist[i].head)->next = NULL;	/* amputate */
	    break;
	 }
      }
   } else {
      /* it's not at the head */
      if (fatp->prev != NULL)
	 fatp->prev->next = fatp->next;
   }

   if (fatp->prev == NULL) {
      /* it's at the tail or not in a queue */
      for (i = 0; i < MAX_FAT_LIST; i++) {
	 if (fatlist[i].tail == fatp) {
	    fatlist[i].tail = fatp->next;	/* move head fwd */
	    if (fatlist[i].tail != NULL)
	       (fatlist[i].tail)->prev = NULL;	/* amputate */
	    break;
	 }
      }
   } else {
      /* it's not at the tail */
      if (fatp->next != NULL)
	 fatp->next->prev = fatp->prev;
   }

#ifdef FATDIAG
   printf("Removed 0x%.8lx...", fatp);	/* FATDIAG */
   for (i = 0; i < MAX_FAT_LIST; i++) {
      if ((fatlist[i].head == NULL && fatlist[i].tail != NULL) ||
	  (fatlist[i].head != NULL && fatlist[i].tail == NULL)) {
	 printf("after: %d %.8lx %.8lx\n", i, fatlist[i].head, fatlist[i].tail);
	 abort();
      }
   }
#endif
   fatp->prev = NULL;
   fatp->next = NULL;
}

/*
 * Add a FAT_NODE to the tail of a temporary queue.  The merge() routine
 * merges the temporary queues with the fatlists once the transmission is
 * sent.
 */

static void 
enqueue(fatp, list)
   FAT_NODE                       *fatp;
   int                             list;
{
#ifdef FATDIAG
   printf("added to tmplist %d\n", list);	/* FATDIAG */
#endif

   if (tmplist[list].tail == NULL) {
      /* list was empty */
      tmplist[list].tail = tmplist[list].head = fatp;
   } else {
      /* list wasn't empty */
      fatp->next = tmplist[list].tail;
      fatp->next->prev = fatp;
      tmplist[list].tail = fatp;
   }
}

/*
 * This gets called from flushSockBuf after the transmission is sent.  It
 * appends all the packets sitting in temporary queues to the corresponding
 * fat queues, where they will be eligible for fattening next transmission.
 */

static void 
fatMerge()
{
   int                             i;
   for (i = 0; i < MAX_FAT_LIST; i++) {
      if (tmplist[i].head == NULL)
	 continue;		/* temp list is empty, nothing to do */

      if (fatlist[i].head == NULL) {
	 /* fatlist is empty; just copy pointers */
	 fatlist[i].head = tmplist[i].head;
	 fatlist[i].tail = tmplist[i].tail;

      } else {
	 /* stuff in both */ 
	 (tmplist[i].head)->next = fatlist[i].tail;	/* fwd pointer */
	 (fatlist[i].tail)->prev = tmplist[i].head;	/* back pointer */
	 fatlist[i].tail = tmplist[i].tail;	/* move tail back */
      }
      tmplist[i].head = tmplist[i].tail = NULL;
#ifdef FATDIAG
      printf("merged list %d: %.8lx %.8lx\n", i, fatlist[i].head,
	     fatlist[i].tail);
#endif
   }
}

int 
check_mesgs(packet)
   struct mesg_cpacket            *packet;
{
   char                            buf[256];
   char                            addrb[10];
   char                           *addrs();
   FILE                           *glog;
   /*
    * Log messages to god or messages preceeded by "god:" ("GOD:")
    */

   if (loggod) {
      long                            t;
      char                            tbuf[40];
      struct tm                      *tmv;
      if (strncasecmp(packet->mesg, "god:", 4) == 0) {

	 glog = fopen(GODLOG, "a");
	 if (!glog) {
	    perror(GODLOG);
	    return 0;
	 }
	 t = time(NULL);
	 tmv = localtime(&t);

	 strftime(tbuf, 39, "%R %D", tmv);
	 sprintf(buf, "%s %s@%s \"%s\"", tbuf, login, host,
		 packet->mesg);

#ifndef SYSV
	 flock(fileno(glog), LOCK_EX);
#endif
	 fprintf(glog, "%s\n", buf);
#ifndef SYSV
	 flock(fileno(glog), LOCK_UN);
#endif
	 fclose(glog);

	 /* response */
	 sprintf(addrb, "GOD->%c%c", teamlet[me->p_team], shipnos[me->p_no]);
	 sprintf(buf, "Thank you, %s, for your comments.",
		 me->p_login);
	 pmessage(buf, me->p_no, MINDIV, addrb);

	 sprintf(buf, "Your message has been logged.");
	 pmessage(buf, me->p_no, MINDIV, addrb);
      }
   }
   return 0;			/* continue processing all messages */
}

char                           *
addrs(packet)
   struct mesg_cpacket            *packet;
{
   static char                     addrb[10];
   switch (packet->group) {
   case MALL:
      return "ALL";
   case MGOD:
      return "GOD";
   case MTEAM:
      switch (packet->indiv) {
      case FED:
      case ROM:
      case KLI:
      case ORI:
	 return teamshort[(int) packet->indiv];
      default:
	 return "T?";
      }
      break;
   case MINDIV:
      if (packet->indiv < 0 || packet->indiv > MAXPLAYER)
	 return "I?";
      if (players[packet->indiv].p_status == PFREE)
	 return "D?";
      sprintf(addrb, "%c%c", teamlet[players[packet->indiv].p_team],
	      shipnos[packet->indiv]);
      return addrb;
   default:
      return "??";
   }
}

#ifdef KEEPALIVE
void 
set_tcp_opts(sock)
   int                             sock;
{
   int                             optval = 1;
   /*
    * int             optlen = sizeof(int);
    */
   struct protoent                *ent;
   ent = getprotobyname("TCP");
   if (!ent) {
      fprintf(stderr, "TCP protocol not found.\n");
      return;
   }
   /*
    * if(getsockopt(sock, SOL_SOCKET, SO_KEEPALIVE, &optval, &optlen) < 0)
    * perror("getsockopt"); printf("TCP SO_KEEPALIVE: %d\n", optval);
    */
#ifdef nodef
   if (getsockopt(sock, SOL_SOCKET, SO_SNDBUF, &optval, &optlen) < 0)
      perror("getsockopt");
   fprintf(stderr, "TCP SO_SNDBUF: %d\n", optval);
   if (getsockopt(sock, SOL_SOCKET, SO_RCVBUF, &optval, &optlen) < 0)
      perror("getsockopt");
   fprintf(stderr, "TCP SO_RCVBUF: %d\n", optval);
   fflush(stderr);
#endif

   optval = 1;
   if (setsockopt(sock, SOL_SOCKET, SO_KEEPALIVE, (char *) &optval,
		  sizeof(int)) < 0)
                                      perror("setsockopt");
   /*
    * if(getsockopt(sock, SOL_SOCKET, SO_KEEPALIVE, &optval, &optlen) < 0)
    * perror("getsockopt"); printf("TCP SO_KEEPALIVE: %d\n", optval);
    */
}
#endif

#ifdef PACKSTATS
void 
printPackStats(fo)
   FILE                           *fo;
{
   register                        i;
   /* only packets sent for now */
   for (i = 0; i < NUM_SIZES; i++) {
      fprintf(fo, "%s: %d (%d bytes), ", namePack(i), packetsSent[i],
	      packetsSentBytes[i]);
      if (!(i % 3))
	 fprintf(fo, "\n");
   }
}

char                           *
namePack(i)
   int                             i;
{
   switch (i) {
   case SP_MESSAGE:
      return "SP_MESSAGE";
   case SP_PLAYER_INFO:
      return "SP_PLAYER_INFO";
   case SP_KILLS:
      return "SP_KILLS";
   case SP_PLAYER:
      return "SP_PLAYER";
   case SP_TORP_INFO:
      return "SP_TORP_INFO";
   case SP_TORP:
      return "SP_TORP";
   case SP_PHASER:
      return "SP_PHASER";
   case SP_PLASMA_INFO:
      return "SP_PLASMA_INFO";
   case SP_PLASMA:
      return "SP_PLASMA";
   case SP_WARNING:
      return "SP_WARNING";
   case SP_MOTD:
      return "SP_MOTD";
   case SP_YOU:
      return "SP_YOU";
   case SP_QUEUE:
      return "SP_QUEUE";
   case SP_STATUS:
      return "SP_STATUS";
   case SP_PLANET:
      return "SP_PLANET";
   case SP_PICKOK:
      return "SP_PICKOK";
   case SP_LOGIN:
      return "SP_LOGIN";
   case SP_FLAGS:
      return "SP_FLAGS";
   case SP_MASK:
      return "SP_MASK";
   case SP_PSTATUS:
      return "SP_PSTATUS";
   case SP_BADVERSION:
      return "SP_BADVERSION";
   case SP_HOSTILE:
      return "SP_HOSTILE";
   case SP_STATS:
      return "SP_STATS";
   case SP_PL_LOGIN:
      return "SP_PL_LOGIN";
   case SP_RESERVED:
      return "SP_RESERVED";
   case SP_PLANET_LOC:
      return "SP_PLANET_LOC";
   case SP_UDP_REPLY:
      return "SP_UDP_REPLY";
   case SP_SEQUENCE:
      return "SP_SEQUENCE";
   case SP_SC_SEQUENCE:
      return "SP_SC_SEQUENCE";

   default:
      return "UNKNOWN";
   }
}
#endif

/*
 * give session stats if you send yourself (or any player) a '?' 2/27/92 TC
 * give lag stats if you send yourself (or any player) a '!' 4/4/93 (Lars
 * Bernhardsson) Return true if you should eat message.
 */

int 
parseQuery(packet)
   struct mesg_spacket            *packet;
{
   /* 0-8 for address, 9 is space */
   char                           *cchar;
#ifdef SHORT_PACKETS
   if (packet->type == SP_S_MESSAGE)
      cchar = &packet->mesg[2];
   else
#endif
      cchar = &packet->mesg[10];

   if (*cchar == '?' && *(cchar+1) == '\0')
#ifdef BASEPRACTICE
      return bounceBaseStats(packet);
#else
      return bounceSessionStats(packet);
#endif
   if (*cchar == '!' && *(cchar+1) == '\0')
      return bouncePingStats(packet);
#ifdef RSA
   if (*cchar == '#' && *(cchar+1) == '\0')
      return bounceRSAClientType(packet);
#endif
#ifdef BD
   if (*cchar == '&' && *(cchar+1) == '\0')
      return bounceBD(packet);
#endif
   if (*cchar == '^' && *(cchar+1) == '\0')
      return bounceSBStats(packet);
   
   /* DEBUG */
   if (strcmp(cchar, "debug")==0){
      char	buf[80];
      sprintf(buf, "dumpshmem > %s.dump.%d", login, getpid());
      if(fork() == 0){
	 signal(SIGALRM, SIG_DFL);
	 execl("/bin/sh", "sh", "-c", buf, NULL);
	 perror("execl");
	 _exit(1);
      }
   }
      
   return FALSE;
}

int 
bounceSessionStats(packet)
   struct mesg_spacket            *packet;
{
   char                            buf[80];
   double                          sessionBombing, sessionPlanets,
                                   sessionOffense, sessionDefense;
   int                             deltaArmies, deltaPlanets,
                                   deltaKills, deltaLosses,
                                   deltaTicks;
   deltaPlanets = me->p_stats.st_tplanets - startTplanets;
   deltaArmies = me->p_stats.st_tarmsbomb - startTarms;
   deltaKills = me->p_stats.st_tkills - startTkills;
   deltaLosses = me->p_stats.st_tlosses - startTlosses;
   deltaTicks = me->p_stats.st_tticks - startTticks;

   if (deltaTicks == 0)
      deltaTicks = 1;

   sessionPlanets = (double) deltaPlanets *status->timeprod /
                                   ((double) deltaTicks * status->planets);

   sessionBombing = (double) deltaArmies *status->timeprod /
                                   ((double) deltaTicks * status->armsbomb);

   sessionOffense = (double) deltaKills *status->timeprod /
                                   ((double) deltaTicks * status->kills);

   sessionDefense = (double) deltaTicks *status->losses /
                                   (deltaLosses != 0 ?
				                                    ((double) deltaLosses * status->timeprod) :
                                      (status->timeprod));
   sprintf(buf, "%2s stats: %d planets and %d armies. %d wins/%d losses. %5.2f hours.",
	   me->p_mapchars,
	   deltaPlanets,
	   deltaArmies,
	   deltaKills,
	   deltaLosses,
	   (double) deltaTicks / 36000.0);
   bounce(buf, packet->m_from);
   sprintf(buf, "Ratings: Pla: %5.2f  Bom: %5.2f  Off: %5.2f  Def: %5.2f  Ratio: %4.2f",
	   sessionPlanets,
	   sessionBombing,
	   sessionOffense,
	   sessionDefense,
	   (double) deltaKills /
	   (double) ((deltaLosses == 0) ? 1 : deltaLosses));
   bounce(buf, packet->m_from);
   return TRUE;
}

int
bounceSBStats(packet)

    struct mesg_spacket         *packet;
{
   char 			buf[80];
   float			sessionRatio, sessionKPH, sessionDPH,
				overallKPH, overallDPH, overallRatio;
   int				deltaKills, deltaLosses,
				deltaTicks;

   deltaKills = me->p_stats.st_sbkills - startSBkills;
   deltaLosses = me->p_stats.st_sblosses - startSBlosses;
   deltaTicks = me->p_stats.st_sbticks - startSBticks;
   overallRatio = (float) (me->p_stats.st_sblosses==0) ?
      (float) me->p_stats.st_sbkills :
      (float) me->p_stats.st_sbkills / (float) me->p_stats.st_sblosses;

   if (deltaTicks != 0) {
	 sessionKPH = (float) deltaKills * 36000.0 / (float) deltaTicks;
	 sessionDPH = (float) deltaLosses * 36000.0 / (float) deltaTicks;
   } else {
	 sessionKPH = sessionDPH = 0.0;
   }

   sessionRatio = (float) (deltaLosses==0) ? (float) deltaKills :
	 (float) deltaKills / (float) deltaLosses;

   if (me->p_stats.st_sbticks != 0) {
	 overallKPH = (float) me->p_stats.st_sbkills * 36000.0 /
	    (float) me->p_stats.st_sbticks;
	 overallDPH = (float) me->p_stats.st_sblosses * 36000.0 /
	    (float) me->p_stats.st_sbticks;
   } else {
	 overallKPH = overallDPH = 0.0;
   }
   sprintf(buf,
      "%2s overall SB stats: %d wins/%d losses. %5.2f hours. Ratio: %5.2f",
      me->p_mapchars,
      me->p_stats.st_sbkills,
      me->p_stats.st_sblosses,
      (float) me->p_stats.st_sbticks/36000.0,
      overallRatio);
   bounce(buf, packet->m_from);

   if (deltaTicks){
      sprintf(buf,
          "%2s session SB stats: %d wins/%d losses. %5.2f hours. Ratio: %5.2f",
          me->p_mapchars,
          deltaKills,
          deltaLosses,
          (float) deltaTicks/36000.0,
          sessionRatio);
      bounce(buf, packet->m_from);
      sprintf(buf,
          "Kills/Hour: %5.2f (%5.2f total), Deaths/Hour: %4.2f (%4.2f total)",
          sessionKPH,
          overallKPH,
          sessionDPH,
          overallDPH);
      bounce(buf, packet->m_from);
   }
   return TRUE;
}

#ifdef BASEPRACTICE
int
bounceBaseStats(packet)

   struct mesg_spacket	*packet;
{
   char		buf[80];

   int		deltaKills, deltaLosses, deltaTicks;
   double	deltaDami, deltaDamr, sessionDefense,
		sessionDami, sessionDamr;

   deltaKills = me->p_stats.st_tkills - startTkills;
   deltaLosses = me->p_stats.st_tlosses - startTlosses;
   deltaDami = me->p_stats.st_dami - startTdami;
   deltaDamr = me->p_stats.st_damr - startTdamr;
   deltaTicks = me->p_stats.st_tticks - startTticks;

   if(deltaTicks == 0)
      deltaTicks = 1;
   
   sessionDami = deltaDami * status->timeprod /
		   ((double) deltaTicks * status->dami);

   sessionDamr = deltaDamr * status->timeprod /
		   ((double) deltaTicks * status->damr);

   sessionDefense = ((double)deltaKills / 
		(deltaLosses + 2.0 + (double)deltaTicks / 14400))/
		((double) status->kills / (status->losses + 2.0 + 
			(double)(status->timeprod)/14400));

   sprintf(buf, "%2s: %d w/%d l, %5.2f dam infl/ %5.2f dam rprd, %5.2f df, %5.2f hours.",
	 me->p_mapchars,
	 deltaKills,
	 deltaLosses,
	 sessionDami,
	 sessionDamr,
	 sessionDefense,
	(double) deltaTicks / 36000.0);
   bounce(buf, packet->m_from);

   /* total actually */
   sessionDami = damInflictedRating(me);
   sessionDamr = damRepairedRating(me);
   sessionDefense = sbRateTradNorm(me);	/* not quite defense */
   sprintf(buf, "TOTAL RATING: Dam Infl: %5.2f / Dam Rprd: %5.2f / Def: %5.2f = %5.2f",
	 sessionDami,
	 sessionDamr,
	 sessionDefense,
	 sessionDami + sessionDamr + sessionDefense);
   bounce(buf, packet->m_from);
   return TRUE;
}
#endif

int 
bouncePingStats(packet)
   struct mesg_spacket            *packet;
{
   char                            buf[80];
   if (me->p_avrt == -1) {
      /* client doesn't support it or server not pinging */
      sprintf(buf, "No PING stats available.");
   } else {
      sprintf(buf, 
	 "PING stats: Average: %d ms, Stdv: %d ms, Loss: %d%%/%d%% s->c/c->s",
	      me->p_avrt,
	      me->p_stdv,
	      me->p_pkls_s_c,
	      me->p_pkls_c_s);
   }
   bounce(buf, packet->m_from);

   return TRUE;
}

#ifdef RSA
int 
bounceRSAClientType(packet)
   struct mesg_spacket            *packet;
{
   char                            buf[160];
   if (*RSA_client_type)
      sprintf(buf, "Client: %s", RSA_client_type);
   else
      strcpy(buf, "Not using an RSA client.");
   buf[72] = 0;
   bounce(buf, packet->m_from);

   return TRUE;
}
#endif

#ifdef BD
int 
bounceBD(packet)
   struct mesg_spacket            *packet;
{
   char                            buf[80];
   int				   val, count;
   val = bd_borg(&count);

   sprintf(buf, "%5.2f (%d (%d))",
      (float)count/(me->p_stats.st_tticks-startTticks),
      val, count);
   bounce(buf, packet->m_from);
   return TRUE;
}
#endif

/* new code, sends bouncemsg to bounceto from GOD 4/17/92 TC */
void 
bounce(bouncemsg, bounceto)
   char                           *bouncemsg;
   int                             bounceto;
{
   pmessage2(bouncemsg, bounceto, MINDIV, getaddr2(MINDIV, me->p_no, bounceto),
      me->p_no);
}

#ifdef SHORT_PACKETS

void
handleShortReq(packet)
   struct shortreq_cpacket        *packet;
{
   struct shortreply_spacket       resp;
   switch (packet->req) {
   case SPK_VOFF:
      send_short = 0;
      warning(2, "Not sending variable and short packets.  Back to default.",
	      -1);
      if (udpSock >= 0 && udpMode == MODE_FAT) forceUpdate();
#ifdef nodef
      udpMode = MODE_SIMPLE;	/* disable FAT udp for non-short packets */
#endif
      break;

   case SPK_VON:
      if (packet->version != (char) SHORTVERSION) {
	 warning(2, "Your SHORT Protocol Version is not right!", -1);
	 packet->req = SPK_VOFF;
	 break;
      }
      if (!send_short)
	 warning(2, "Sending variable and short packets.", -1);	/* send only firsttime */
      send_short = 1;
      resp.winside = htons(WINSIDE);
      resp.gwidth = htonl(GWIDTH);
      break;

   case SPK_MOFF:
      send_mesg = 1;
      warning(2, "Obsolete!", -1);
      packet->req = SPK_MON;
      break;

   case SPK_MON:
      send_mesg = 1;
      warning(2, "All messages sent.", -1);
      break;

   case SPK_M_KILLS:
      send_kmesg = 1;
      warning(2, "Kill messages sent", -1);
      break;

   case SPK_M_NOKILLS:
      send_kmesg = 1;
      warning(2, "Obsolete!", -1);
      packet->req = SPK_M_KILLS;
      break;

   case SPK_M_WARN:
      send_warn = 1;
      warning(2, "Warn messages sent", -1);
      break;

   case SPK_M_NOWARN:
      send_warn = 1;
      warning(2, "Obsolete!", -1);
      packet->req = SPK_M_WARN;
      break;

   case SPK_SALL:
      if (send_short) {
	 spk_update_sall = 1;
	 spk_update_all = 0;
	 forceUpdate();
      } else
	 warning(2, "Activate SHORT Packets first!", -1);
      return;

   case SPK_ALL:
      if (send_short) {
	 spk_update_sall = 0;
	 spk_update_all = 1;
	 forceUpdate();
      } else
	 warning(2, "Activate SHORT Packets first!", -1);
      return;

   default:
      warning(2, "Unknown short packet code", -1);
      return;
   }

   resp.type = SP_S_REPLY;
   resp.repl = (char) packet->req;

   sendClientPacket((struct player_spacket *) & resp);
}

void
handleThresh(packet)
   struct threshold_cpacket       *packet;
{
#ifdef SHORT_THRESHOLD
   char                            buf[80];
   send_threshold = packet->thresh;
   if (send_threshold == 0) {
      warning(2, "Threshold test deactivated.", -1);
   } else if (send_threshold < 60) {	/* my low value */
      send_threshold = 60;	/* means: 1 SP_S_PLAYER+SP_S_YOU + 16 bytes */
   }
   sprintf(buf, "Threshold set to %d.", send_threshold);
   warning(2, buf, -1);
#else
   warning(2, "Server is compiled without Thresholdtesting!", -1);
#endif
}

void
updateVPlayer(p)
   struct player_spacket          *p;
{
   register int                    i = clientVPlayerCount;

#if MAXPLAYER > 32
   register int 		   j = clientVXPlayerCount;
#endif
   register long                   x = ntohl(p->x), y = ntohl(p->y);
   register int                    dir, dx, dy;
   int                             view = SCALE * WINSIDE / 2;

   if (p->pnum == me->p_no) {
      struct player_s_spacket        *pl = (struct player_s_spacket *) & clientVPlayers[0];
      pl->x = htonl(x) /* p->x */ ;
      pl->y = htonl(y) /* p->y */ ;
      pl->speed = p->speed;
      pl->dir = p->dir;
      big = 1;			/* Flag to remind me that we must send big
				 * header */
   } else 
#if MAXPLAYER >= 32
   if (p->pnum < 32) 
#endif
   {
      clientVPlayers[1] += 1;	/* Another packet... */
      /* Where is the ship? On local or galactic map. */

      dx = x - me_x;
      dy = y - me_y;

      if (dx > view || dx < -view || dy > view || dy < -view) {	/* Galactic */

	 dx = (x * WINSIDE) / GWIDTH;	/* Orig. */
	 dy = (y * WINSIDE) / GWIDTH;
	 if (dx < 0 || dy < 0 || dx >= WINSIDE || dy >= WINSIDE) {
	    dx = 501;		/* clipped */
	    dy = 501;
	 }
	 dir = rosette(p->dir);
	 clientVPlayers[i++] = (unsigned char) ((p->pnum) | ((dx & 256) >> 2) | ((dy & 256) >> 1) | 32);
	 clientVPlayers[i++] = (unsigned char) ((p->speed & 0x0f) | (dir << 4));
	 clientVPlayers[i++] = (unsigned char) (dx & 0xff);
	 clientVPlayers[i++] = (unsigned char) (dy & 0xff);
	 clientVPlayerCount = i;
      }
      else{
	 dx = x/SCALE - me_x/SCALE + WINSIDE/2;
	 dy = y/SCALE - me_y/SCALE + WINSIDE/2;

	 /*
	 dx = dx / SCALE + WINSIDE / 2;
	 dy = dy / SCALE + WINSIDE / 2;
	 */

	 dir = rosette(p->dir);
	 clientVPlayers[i++] = (unsigned char) ((p->pnum) | ((dx & 256) >> 2) | ((dy & 256) >> 1));
	 clientVPlayers[i++] = (unsigned char) ((p->speed & 0x0f) | (dir << 4));
	 clientVPlayers[i++] = (unsigned char) (dx & 0xff);
	 clientVPlayers[i++] = (unsigned char) (dy & 0xff);
	 clientVPlayerCount = i;
      }
   }
#if MAXPLAYER > 32
   else 
#if MAXPLAYER >= 64
   if (p->pnum < 64) 
#endif
   {
      clientVXPlayers[1] += 1;
      /* Where is the ship? On local or galactic map. */
      dx = x - me_x;
      dy = y - me_y;
      if (dx > view || dx < -view || dy > view || dy < -view) {	/* Galactic */
	 dx = x * WINSIDE / GWIDTH;	/* Orig. */
	 dy = y * WINSIDE / GWIDTH;
	 if (dx < 0 || dy < 0 || dx >= WINSIDE || dy >= WINSIDE) {
	    dx = 501;		/* clipped */
	    dy = 501;
	 }
	 dir = rosette(p->dir);
	 clientVXPlayers[j++] = (unsigned char) ((p->pnum - 32) | ((dx & 256) >> 2) | ((dy & 256) >> 1) | 32);
	 clientVXPlayers[j++] = (unsigned char) ((p->speed & 0x0f) | (dir << 4));
	 clientVXPlayers[j++] = (unsigned char) (dx & 0xff);
	 clientVXPlayers[j++] = (unsigned char) (dy & 0xff);
	 clientVXPlayerCount = j;
      } else {			/* Local */
	 dx = dx / SCALE + WINSIDE / 2;
	 dy = dy / SCALE + WINSIDE / 2;
	 dir = rosette(p->dir);
	 clientVXPlayers[j++] = (unsigned char) ((p->pnum - 32) | ((dx & 256) >> 2) | ((dy & 256) >> 1));
	 clientVXPlayers[j++] = (unsigned char) ((p->speed & 0x0f) | (dir << 4));
	 clientVXPlayers[j++] = (unsigned char) (dx & 0xff);
	 clientVXPlayers[j++] = (unsigned char) (dy & 0xff);
	 clientVXPlayerCount = j;
      }
   }
#endif				/* MAXPLAYER > 32 */
#if MAXPLAYER > 64
   else
      sendClientPacket((struct player_spacket *) p);
#endif

}

void
sendVPlayers()
{
   int                             size;
   /* initialize type & length fields */
   clientVPlayers[0] = (unsigned char) SP_S_PLAYER;
#if MAXPLAYER > 32
   clientVXPlayers[0] = (unsigned char) SP_S_PLAYER;
#endif

   if (big) {					/* Big Packet */
      size = 12 + (clientVPlayers[1] * 4);
      sizes[SP_S_PLAYER] = size;
      sendClientPacket((CVOID) clientVPlayers);
   } else if (clientVPlayerCount > 12) {	/* Small packet */
      size = 4 + (clientVPlayers[1] * 4);
      sizes[SP_S_PLAYER] = size;
      clientVPlayers[8] = (unsigned char) SP_S_PLAYER;
      clientVPlayers[9] = clientVPlayers[1] | 64;
      clientVPlayers[10] = 0;
      clientVPlayers[11] = 0;
      sendClientPacket((CVOID) & clientVPlayers[8]);
   }
#if MAXPLAYER > 32
   if (clientVXPlayerCount > 4) {	/* Small + extended Packet */
      sizes[SP_S_PLAYER] = (clientVXPlayers[1] * 4) + 4;
      clientVXPlayers[1] |= 128;/* To mark it as a short header + Ext ended */
      sendClientPacket((CVOID) clientVXPlayers);
   }
#endif
}

/* New Torp Routine */
void
SupdateTorps()
{
   register struct torp           *torp;
   register int                    i, j;
   register struct torp_info_spacket *tpi;
   register struct torp_spacket   *tp;
   register unsigned char         *tinfo;	/* TorpInfo data   */
   register unsigned char         *torp_xy;	/* The torp coord. */
   register unsigned char          infobitset, torpbitset;
   int                             numtorps;	/* How many torps ? */

   int                             shift;	/* Shift for 9 bit xy
						 * coordinate */
   register int                    dx, dy;
   int                             view = SCALE * WINSIDE / 2;
   mustsend = 0;

   for (j = 0, torp = torps, tpi = clientTorpsInfo, tp = clientTorps;
	j < (MAXPLAYER * MAXTORP) / 8; j++) {

      /* NEW */
      if(!players[j].p_status) {
	 torp += MAXTORP;
	 tpi += MAXTORP;
	 tp += MAXTORP;
	 continue;
      }

      shift = 0;
      numtorps = 0;
      infobitset = 0;
      torpbitset = 0;
      torp_xy = &clientVTorps[4];
      *torp_xy = 0;		/* Clear first data byte */
      tinfo = &clientVTorpsInfo[0];

      for (i = 0; i < 8; i++, torp++, tpi++, tp++) {

	 if (torp->t_owner == me->p_no) {
	    if (torp->t_war != tpi->war ||
		torp->t_status != tpi->status) {
	       if (torp->t_war == tpi->war) {
		  /*
		   * Don't send TMOVE/TFREE . It's encoded in the SP_S_TORP
		   * bitset.
		   */
		  switch (torp->t_status) {
		  case TFREE:
		     if(tpi->status == TEXPLODE){ /* send real TFREE */
			infobitset = infobitset | (01 << i);
			*tinfo ++ = ((uchar)tpi->war & 0x0f)|
				    ((uchar)torp->t_status << 4);
		     }
		     else
			mustsend = 1;		/* XX? */
		     tpi->status = torp->t_status;
		     /* Don't forget... */
		     tp->x = htonl(torp->t_x);
		     tp->y = htonl(torp->t_y);
		     tp->dir = torp->t_dir;

		     continue;	/* We continue , because server sends
				 * obsolete data otherwise */
		  case TMOVE:
		     tpi->status = torp->t_status;
		     break;
		  case TSTRAIGHT:
		     tpi->status = torp->t_status;
		     break;
		  default:
		     tpi->status = torp->t_status;
		     infobitset = infobitset | (01 << i);	/* set bit for torp */
		     *tinfo++ = ((uchar) tpi->war & 0x0f) | ((uchar) tpi->status << 4);
		     break;
		  }
	       } else {
		  tpi->war = torp->t_war;
		  tpi->status = torp->t_status;
		  infobitset = infobitset | (01 << i);	/* set bit for torp */
		  *tinfo++ = ((uchar) tpi->war & 0x0f) | ((uchar) tpi->status << 4);
	       }
	    }
	    if (tp->x != htonl(torp->t_x) ||
		tp->y != htonl(torp->t_y) || 
		(torp->t_speed == 0 && torp->t_status != TFREE)){
	       tp->type = SP_TORP;
	       tp->x = htonl(torp->t_x);
	       tp->y = htonl(torp->t_y);
	       tp->dir = torp->t_dir;
	       /* Now the fun begins .. */
	       torpbitset = torpbitset | (01 << i);	/* set bit for torp */
	       numtorps++;
	       /* Now the Coordi. */
	       dx = torp->t_x - me_x;
	       dy = torp->t_y - me_y;
	       if (dx > view || dx < -view || dy > view || dy < -view) {
		  /* any torps off screen appear at 501 = clipped */
		  dx = dy = 501;
		  dx <<= shift;
		  *torp_xy++ |= (dx & 255);
		  *torp_xy = (((unsigned int) dx >> 8) & 255);
		  shift++;
		  dy <<= shift;
		  *torp_xy++ |= (dy & 255);
		  *torp_xy = (((unsigned int) dy >> 8) & 255);
		  shift++;
		  if (shift == 8) {
		     shift = 0;
		     torp_xy++;
		     *torp_xy = 0;
		  }
		  continue;
	       }
	       dx = torp->t_x/SCALE - me_x/SCALE + WINSIDE/2;
	       dy = torp->t_y/SCALE - me_y/SCALE + WINSIDE/2;
	       /*
	       dx = dx / SCALE + WINSIDE / 2;
	       dy = dy / SCALE + WINSIDE / 2;
	       */
	       dx <<= shift;
	       *torp_xy++ |= (dx & 0xff);
	       *torp_xy = (((unsigned int) dx >> 8) & 255);
	       shift++;
	       dy <<= shift;
	       *torp_xy++ |= (dy & 255);
	       *torp_xy = (((unsigned int) dy >> 8) & 255);
	       shift++;
	       if (shift == 8) {
		  shift = 0;
		  torp_xy++;
		  *torp_xy = 0;
	       }
	    }
	 } else {		/* Someone else's torp... */
	    if (PtOffLocal(torp->t_x, torp->t_y) ||
		torp->t_status == TFREE) {
	       if (torp->t_status == TFREE && tpi->status == TEXPLODE) {
		  tpi->status = TFREE;
		  continue;
	       }
	       if (tpi->status != TFREE) {
		  tpi->status = TFREE;
		  mustsend = 1;
	       }
	    } else {		/* in view */
	       if (torp->t_status != tpi->status ||
		 (torp->t_war & me->p_team) != tpi->war) {
		  if ((torp->t_war & me->p_team) == tpi->war) {
		     /*
		      * Don't send TMOVE/TFREE . It's encoded in the
		      * SP_S_TORP bitset.
		      */
		     switch (torp->t_status) {
		     case TFREE:	/* We continue , because server sends
					 * obsolete data otherwise */
			tpi->status = torp->t_status;
			mustsend = 1;		/* XX? */
			/* Don't forget... */
			tp->x = htonl(torp->t_x);
			tp->y = htonl(torp->t_y);
			tp->dir = torp->t_dir;
			continue;
		     case TMOVE:
			tpi->status = torp->t_status;
			break;
		     case TSTRAIGHT:
			tpi->status = torp->t_status;
			break;
		     default:
			tpi->status = torp->t_status;
			infobitset = infobitset | (01 << i);	/* set bit for torp */
			*tinfo++ = ((uchar) tpi->war & 0x0f) | ((uchar) tpi->status << 4);
			break;
		     }
		  } else {
		     /* Let the client fade away the explosion on its own */
		     tpi->war = torp->t_war & me->p_team;
		     tpi->status = torp->t_status;
		     infobitset = infobitset | (01 << i);	/* set bit for torp */
		     *tinfo++ = ((unsigned char) tpi->war & 0x0f) | ((unsigned char) tpi->status << 4);
		  }
	       }
	       if (tp->x != htonl(torp->t_x) || tp->y != htonl(torp->t_y)
		   || (torp->t_speed == 0 && torp->t_status != TFREE)) {
		  tp->x = htonl(torp->t_x);
		  tp->y = htonl(torp->t_y);
		  tp->dir = torp->t_dir;
		  /* Now the fun begins .. */
		  torpbitset = torpbitset | (01 << i);	/* set bit for torp */
		  numtorps++;
		  /* Now the Coordi. */
/*
		  dx = torp->t_x - me_x;
		  dx = dx / SCALE + WINSIDE / 2;
		  dy = torp->t_y - me_y;
		  dy = dy / SCALE + WINSIDE / 2;
*/
		  dx = torp->t_x/SCALE - me_x/SCALE + WINSIDE/2;
		  dy = torp->t_y/SCALE - me_y/SCALE + WINSIDE/2;

		  dx <<= shift;
		  *torp_xy++ |= (dx & 255);
		  *torp_xy = (((unsigned int) dx >> 8) & 255);
		  shift++;
		  dy <<= shift;
		  *torp_xy++ |= (dy & 255);
		  *torp_xy = (((unsigned int) dy >> 8) & 255);
		  shift++;
		  if (shift == 8) {
		     shift = 0;
		     torp_xy++;
		     *torp_xy = 0;
		  }
	       }		/* second if */
	    }
	 }
      }				/* 2. for */

      /* Is there something to send ? */
      if (infobitset != 0) {
	 /* send torp info */
	 int                             size = tinfo - clientVTorpsInfo;
	 unsigned char                  *dest = &clientVTorps[4 + vtdata[numtorps]];
	 unsigned char                  *source = &clientVTorpsInfo[0];
	 /* I must copy the Torpinfo behind the Torpdata */
#ifdef nodef
	 while (--size >= 0) {	/* size doubly used */
	    *dest++ = *source++;
	 }
#endif
	 bcopy(source, dest, size);

	 size = 4 + vtdata[numtorps] + (tinfo - clientVTorpsInfo);
	 if ((size % 4) != 0)
	    size += (4 - (size % 4));

	 sizes[SP_S_TORP_INFO] = size;
	 clientVTorps[0] = SP_S_TORP_INFO;	/* The type */
	 clientVTorps[1] = (unsigned char) torpbitset;
	 clientVTorps[2] = (unsigned char) j;	/* which torps */
	 clientVTorps[3] = (unsigned char) infobitset;
	 sendClientPacket((CVOID) & clientVTorps[0]);

      } else if (torpbitset != 0) {	/* NO TorpInfo */
	 /* send torp location */
	 mustsend = 0;
	 if (numtorps == 8) {	/* we do not need bitset */
	    clientVTorps[2] = SP_S_8_TORP;	/* The type */
	    clientVTorps[3] = (unsigned char) j;	/* which torps */
	    sendClientPacket((CVOID) & clientVTorps[2]);
	 } else {
	    sizes[SP_S_TORP] = vtsize[numtorps];
	    clientVTorps[1] = SP_S_TORP;	/* The type */
	    clientVTorps[2] = (unsigned char) torpbitset;
	    clientVTorps[3] = (unsigned char) j;	/* which torps */
	    sendClientPacket((CVOID) & clientVTorps[1]);
	 }
      } else if (mustsend != 0) {	/* TFREE's */
	 /* ?? */
	 mustsend = 0;
	 sizes[SP_S_TORP] = 4;
	 clientVTorps[1] = SP_S_TORP;	/* The type */
	 clientVTorps[2] = (unsigned char) torpbitset;
	 clientVTorps[3] = (unsigned char) j;	/* which torps */
	 sendClientPacket((CVOID) & clientVTorps[1]);
      }
   }				/* auesseres for */

   /* That's the main loop */
   /* Now we must process the rest torps */
   /* I know this is very ugly */
#if (MAXPLAYER*MAXTORP) % 8
   if ((MAXPLAYER * MAXTORP) % 8) {	/* Rest torps */

      shift = 0;
      numtorps = 0;
      infobitset = 0;
      torpbitset = 0;
      torp_xy = &clientVTorps[4];
      *torp_xy = 0;		/* Clear first data byte */
      tinfo = &clientVTorpsInfo[0];
      mustsend = 0;

      for (i = 8 * j; i < MAXPLAYER * MAXTORP; i++, torp++, tpi++, tp++) {

	 if (torp->t_owner == me->p_no) {
	    if (torp->t_war != tpi->war ||
		torp->t_status != tpi->status) {
	       if (torp->t_war == tpi->war) {
		  /*
		   * Don't send TMOVE/TFREE . It's encoded in the SP_S_TORP
		   * bitset.
		   */
		  switch (torp->t_status) {
		  case TFREE:	/* We continue , because server sends
				 * obsolete data */
		     if(tpi->status == TEXPLODE){ /* send real TFREE */
			infobitset = infobitset | (01 << i);
			*tinfo++ = ((uchar) tpi->war & 0x0f)|
				   ((uchar) torp->t_status << 4);
		     }
		     else
			mustsend = 1;
		     tpi->status = torp->t_status;
		     /* Don't forget... */
		     tp->x = htonl(torp->t_x);
		     tp->y = htonl(torp->t_y);
		     tp->dir = torp->t_dir;

		     continue;
		  case TMOVE:
		     tpi->status = torp->t_status;
		     break;
		  case TSTRAIGHT:
		     tpi->status = torp->t_status;
		     break;
		  default:
		     tpi->status = torp->t_status;
		     infobitset = infobitset | (01 << i);	/* set bit for torp */
		     *tinfo++ = ((uchar) tpi->war & 0x0f) | ((uchar) tpi->status << 4);
		     break;
		  }
	       } else {
		  tpi->war = torp->t_war;
		  tpi->status = torp->t_status;
		  infobitset = infobitset | (01 << i);	/* set bit for torp */
		  *tinfo++ = ((uchar) tpi->war & 0x0f) | ((uchar) tpi->status << 4);
	       }
	    }
	    if (tp->x != htonl(torp->t_x) ||
		tp->y != htonl(torp->t_y) || 
		(torp->t_speed == 0 && torp->t_status != TFREE)) {
	       tp->type = SP_TORP;
	       tp->x = htonl(torp->t_x);
	       tp->y = htonl(torp->t_y);
	       tp->dir = torp->t_dir;
	       /* Now the fun begins .. */
	       torpbitset = torpbitset | (01 << i);	/* set bit for torp */
	       numtorps++;
	       /* Now the Coordi. */
	       dx = torp->t_x - me_x;
	       dy = torp->t_y - me_y;
	       if (dx > view || dx < -view || dy > view || dy < -view) {
		  /* any torps off screen appear at 501 = clipped */
		  dx = dy = 501;
		  dx <<= shift;
		  *torp_xy++ |= (dx & 255);
		  *torp_xy = (((unsigned int) dx >> 8) & 255);
		  shift++;
		  dy <<= shift;
		  *torp_xy++ |= (dy & 255);
		  *torp_xy = (((unsigned int) dy >> 8) & 255);
		  shift++;
		  if (shift == 8) {
		     shift = 0;
		     torp_xy++;
		     *torp_xy = 0;
		  }
		  continue;
	       }
	       dx = torp->t_x/SCALE - me_x/SCALE + WINSIDE/2;
	       dy = torp->t_y/SCALE - me_y/SCALE + WINSIDE/2;
	       /*
	       dx = dx / SCALE + WINSIDE / 2;
	       dy = dy / SCALE + WINSIDE / 2;
	       */
	       dx <<= shift;
	       *torp_xy++ |= (dx & 0xff);
	       *torp_xy = (((unsigned int) dx >> 8) & 255);
	       shift++;
	       dy <<= shift;
	       *torp_xy++ |= (dy & 255);
	       *torp_xy = (((unsigned int) dy >> 8) & 255);
	       shift++;
	       if (shift == 8) {
		  shift = 0;
		  torp_xy++;
		  *torp_xy = 0;
	       }
	    }
	 } else {		/* Someone else's torp... */
	    if (torp->t_y > me_y + SCALE * WINSIDE / 2 ||
	      torp->t_x > me_x + SCALE * WINSIDE / 2 ||
	      torp->t_x < me_x - SCALE * WINSIDE / 2 ||
	      torp->t_y < me_y - SCALE * WINSIDE / 2 ||
		torp->t_status == TFREE) {
	       if (torp->t_status == TFREE && tpi->status == TEXPLODE) {
		  tpi->status = TFREE;
		  continue;
	       }
	       if (tpi->status != TFREE) {
		  tpi->status = TFREE;
		  mustsend = 1;
	       }
	    } else {		/* in view */
	       if (torp->t_status != tpi->status ||
		 (torp->t_war & me->p_team) != tpi->war) {
		  if ((torp->t_war & me->p_team) == tpi->war) {
		     /*
		      * Don't send TMOVE/TFREE . It's encoded in the
		      * SP_S_TORP bitset.
		      */
		     switch (torp->t_status) {
		     case TFREE:	/* We continue , because server sends
					 * obsolete data */
			tpi->status = torp->t_status;
			mustsend = 1;
			/* Don't forget... */
			tp->x = htonl(torp->t_x);
			tp->y = htonl(torp->t_y);
			tp->dir = torp->t_dir;
			continue;
		     case TMOVE:
			tpi->status = torp->t_status;
			break;
		     case TSTRAIGHT:
			tpi->status = torp->t_status;
			break;
		     default:
			tpi->status = torp->t_status;
			infobitset = infobitset | (01 << i);	/* set bit for torp */
			*tinfo++ = ((uchar) tpi->war & 0x0f) | ((uchar) tpi->status << 4);
			break;
		     }
		  } else {
		     /* Let the client fade away the explosion on its own */
		     tpi->war = torp->t_war & me->p_team;
		     tpi->status = torp->t_status;
		     infobitset = infobitset | (01 << i);	/* set bit for torp */
		     *tinfo++ = ((unsigned char) tpi->war & 0x0f) | ((unsigned char) tpi->status << 4);
		  }
	       }
	       if (tp->x != htonl(torp->t_x) ||
		   tp->y != htonl(torp->t_y) ||
		   (torp->t_speed == 0 && torp->t_status != TFREE) {
		  tp->x = htonl(torp->t_x);
		  tp->y = htonl(torp->t_y);
		  tp->dir = torp->t_dir;
		  /* Now the fun begins .. */
		  torpbitset = torpbitset | (01 << i);	/* set bit for torp */
		  numtorps++;
		  /* Now the Coordi. */
/*
		  dx = torp->t_x - me_x;
		  dx = dx / SCALE + WINSIDE / 2;
		  dy = torp->t_y - me_y;
		  dy = dy / SCALE + WINSIDE / 2;
*/

		  dx = torp->t_x/SCALE - me_x/SCALE + WINSIDE/2;
		  dy = torp->t_y/SCALE - me_y/SCALE + WINSIDE/2;

		  dx <<= shift;
		  *torp_xy++ |= (dx & 255);
		  *torp_xy = (((unsigned int) dx >> 8) & 255);
		  shift++;
		  dy <<= shift;
		  *torp_xy++ |= (dy & 255);
		  *torp_xy = (((unsigned int) dy >> 8) & 255);
		  shift++;
		  if (shift == 8) {
		     shift = 0;
		     torp_xy++;
		     *torp_xy = 0;
		  }
	       }
	    }
	 }
      }				/* 2. for */

      /* Is there something to send ? */
      if (infobitset != 0) {
	 int                             size = tinfo - clientVTorpsInfo;
	 unsigned char                  *dest = &clientVTorps[4 + vtdata[numtorps]];
	 unsigned char                  *source = &clientVTorpsInfo[0];
	 /* I must copy the Torpinfo behind the Torpdata */
	 while (--size >= 0) {	/* size doubly used */
	    *dest++ = *source++;
	 }
	 size = 4 + vtdata[numtorps] + (tinfo - clientVTorpsInfo);
	 if ((size % 4) != 0)
	    size += (4 - (size % 4));

	 sizes[SP_S_TORP_INFO] = size;
	 clientVTorps[0] = SP_S_TORP_INFO;	/* The type */
	 clientVTorps[1] = (unsigned char) torpbitset;
	 clientVTorps[2] = (unsigned char) j;	/* which torps */
	 clientVTorps[3] = (unsigned char) infobitset;
	 sendClientPacket((struct torp_spacket *) & clientVTorps[0]);

      } else if (torpbitset != 0) {	/* NO TorpInfo */
	 mustsend = 0;
	 if (numtorps == 8) {	/* we do not need bitset */
	    clientVTorps[2] = SP_S_8_TORP;	/* The type */
	    clientVTorps[3] = (unsigned char) j;	/* which torps */
	    sendClientPacket((struct torp_spacket *) & clientVTorps[2]);
	 } else {
	    sizes[SP_S_TORP] = vtsize[numtorps];
	    clientVTorps[1] = SP_S_TORP;	/* The type */
	    clientVTorps[2] = (unsigned char) torpbitset;
	    clientVTorps[3] = (unsigned char) j;	/* which torps */
	    sendClientPacket((struct torp_spacket *) & clientVTorps[1]);
	 }
      } else if (mustsend != 0) {	/* TFREE's */
	 mustsend = 0;
	 sizes[SP_S_TORP] = 4;
	 clientVTorps[1] = SP_S_TORP;	/* The type */
	 clientVTorps[2] = (unsigned char) torpbitset;
	 clientVTorps[3] = (unsigned char) j;	/* which torps */
	 sendClientPacket((struct torp_spacket *) & clientVTorps[1]);
      }
   }				/* last if */
#endif /* (MAXPLAYER*MAXTORP) % 8 */
}				/* Function */
#endif				/* SHORT_PACKETS */

/*
 * Try to find out all interface names of this machine
 */

int 
gethostnames(sbuf)
   char                            (**sbuf)[16];
{
#ifdef CALVIN
   static char                     addresses[2][16] =
   {"128.125.142.143", "128.125.253.129"};
   *sbuf = addresses;
   return 2;
#else
   struct hostent                 *gethostbyname(), *hent;
   struct in_addr                **adrp;
   char                            hostname[80];
   static char                     addresses[10][16];
   int                             i;
   gethostname(hostname, 80);
   hent = gethostbyname(hostname);
   if (!hent)
      return 0;

   for (i = 0, adrp = (struct in_addr **) hent->h_addr_list; *adrp; ++adrp, i++) {
      if (i == 10) {
	 fprintf(stderr, "Oops, static array 'addresses' not large enough.\n");
	 fprintf(stderr, "(gethostnames`socket.c: increase constant 10.)\n");
	 exitGame(1);
      }
      strncpy(addresses[i], inet_ntoa(**adrp), 16);
   }

   *sbuf = addresses;
   return i;
#endif	/* CALVIN */
}

#ifdef DS
void
forceShutdown(s)
   
   int	s;
{
   signal(SIGALRM, SIG_IGN);
   fprintf(stderr, "forceShutdown for %s@%s: %d\n", login, host,
      s);
   fflush(stderr);
   shutdown(sock, 2);
   close(sock);
   if(udpSock >= 0){
      shutdown(udpSock, 2);
      close(udpSock);
   }
   exit(1);
}
#endif


void
sendMotdPic(index, x, y, bits, line, color, width, height)

   int	   index;
   int     x;
   int     y;
   char   *bits;
   int     line;
   int	   color;
   int     width;
   int     height;
{
   struct motd_pic_spacket	temppack;
   int				size;

   size = ((width+7)/8) * height;

   temppack.type = SP_MOTD_PIC;
   temppack.color = (short)color;
   temppack.index = htons((short)index);
   temppack.line = htons((short)line);
   temppack.x = htons((short)x);
   temppack.y = htons((short)y);
   sendClientPacket((CVOID) &temppack);

   /* duplicate bitmaps are not sent twice */
   if(bits){
      /* followed by a bitmap packet */
      sendBitmap(BITMAP_MOTD, width, height, size, bits);
   }
}

void
sendBitmap(bitmap_type, width, height, size, bits)

   int	bitmap_type;
   int	width, height;
   int	size;
   char	*bits;
{
   struct	bitmap_spacket	temppack;
   register	i = 0, j = size;

#define BITMAP_FRAG_SIZE	sizeof(temppack.bits)

   /* we fragment the large bitmaps into pieces of BITMAP_FRAG_SIZE.
      The client code automatically reassembles them into a single
      bitmap */
   while(j > BITMAP_FRAG_SIZE){
      temppack.type = SP_BITMAP;
      temppack.bitmap_type = (char)bitmap_type;
      temppack.width = htons((short) width);
      temppack.height = htons((short) height);
      temppack.size = htonl(BITMAP_FRAG_SIZE);
      bcopy(&bits[i], temppack.bits, BITMAP_FRAG_SIZE);
      sendClientPacket((CVOID) &temppack);
      j -= BITMAP_FRAG_SIZE;
      i += BITMAP_FRAG_SIZE;
   }
   temppack.type = SP_BITMAP;
   temppack.bitmap_type = (char)bitmap_type;
   temppack.width = htons((short) width);
   temppack.height = htons((short) height);
   temppack.size = htonl(j);
   bcopy(&bits[i], temppack.bits, j);
   sendClientPacket((CVOID) &temppack);
}
