/*
 * warning.c
 */
#include "copyright.h"

#include <stdio.h>
#include <math.h>
#include <signal.h>
#include "defs.h"
#include "struct.h"
#include "data.h"
#include "packets.h"
#ifndef _NO_DEPEND
#   include "ntserv_protos.h"
#endif

#define CVOID                 struct player_spacket *

#if __STDC__ || defined(__cplusplus)
#define P_(s) s
#else
#define P_(s) ()
#endif

static void spwarning	P_((char *text, int index));

#ifdef USE_SYS_PROTOS

extern char *                 strncpy P_((char *, const char *, size_t));

#endif

#undef P_

/*
 * * The warning in text will be printed in the warning window. * The message
 * will last WARNTIME/10 seconds unless another message * comes through and
 * overwrites it.
 */

#ifdef SHORT_PACKETS
extern int                      sizes[];
#endif

void
warning(force, text, sid)
   int                             force;
   char                           *text;
   int                             sid;
{
   struct warning_spacket          warn;
   static int                      last_warn, vi, last_sid;
   int                             n = mstime();

#ifdef SHORT_PACKETS
   if (sid > -1){
      if(last_sid == sid && n - last_warn < 750)
	 return;
      last_sid = sid;
   }
#endif
      

   if (!force) {
      if (n - last_warn < 750) {
	 return;
      }
   }
   if (vi) {
      if (n - last_warn < 2000) {
	 return;
      }
      vi = 0;
   }
   if (force == 2)
      vi = 1;

   last_warn = n;

#ifdef SHORT_PACKETS
   if (sid != -1) {
      spwarning(text, sid);
      return;
   }
#endif

#ifdef SHORT_PACKETS
   if (send_short) {		/* send it like a CP_S_MESSAGE */
      int                             size;
      strncpy(warn.mesg, text, 80);
      warn.mesg[79] = '\0';
      size = strlen(warn.mesg);
      size += 5;		/* 1 for '\0', 4 packetheader */
      if ((size % 4) != 0)
	 size += (4 - (size % 4));
      warn.pad3 = (char) size;
      sizes[SP_S_WARNING] = size;
      warn.type = SP_S_WARNING;
      warn.pad1 = SHORT_WARNING;
      warn.pad2 = 0;
      sendClientPacket((CVOID)&warn);
      sizes[SP_S_WARNING] = sizeof(struct warning_s_spacket);	/* Back to default */
      return;
   }
#endif
   warn.type = SP_WARNING;
   strncpy(warn.mesg, text, 80);
   warn.mesg[79] = '\0';
   sendClientPacket((CVOID)&warn);
}

#ifdef SHORT_PACKETS
/*
 * * The warning in text will be printed in the warning window. * The message
 * will last WARNTIME/10 seconds unless another message * comes through and
 * overwrites it.
 */

static void
spwarning(text, index)
   char                           *text;
   int                             index;	/* The index into the
						 * warnings array in client */
{
   struct warning_spacket          warn;
   struct warning_s_spacket        swarn;

   if (!send_short) {
      warn.type = SP_WARNING;
      strncpy(warn.mesg, text, 80);
      warn.mesg[79] = '\0';
      sendClientPacket((CVOID)&warn);
   } else {			/* These are only 'TEXTE' warnings */
      swarn.type = SP_S_WARNING;
      swarn.whichmessage = TEXTE;
      swarn.argument = (uchar) index & 0xff;
      swarn.argument2 = ((uchar) index >> 8) & 0xff;
      sendClientPacket((CVOID)&swarn);
   }
}

/* This function is for new warnings that are not compiled in the client */
void
s_warning(text, index)
   char                           *text;
   int                             index;	/* The index into the
						 * warnings array in client */
{
   struct warning_spacket          warn;
   struct warning_s_spacket        swarn;
   int                             size;
   static char                     sent_to_client[256];	/* This is a (sortof)
							 * boolean arry to
							 * remind th e server
							 * that he has sent */
   /* a warning to the client */

   if (!send_short) {
      warn.type = SP_WARNING;
      strncpy(warn.mesg, text, 80);
      warn.mesg[79] = '\0';
      sendClientPacket((CVOID)&warn);
   } else if (sent_to_client[index]) {	/* The client has that warning */
      swarn.type = SP_S_WARNING;
      swarn.whichmessage = STEXTE;
      swarn.argument = (uchar) index;
      swarn.argument2 = 0;
      sendClientPacket((CVOID)&swarn);
   } else {			/* first time */
      strncpy(warn.mesg, text, 80);
      warn.mesg[79] = '\0';
      size = strlen(warn.mesg);
      size += 5;		/* 1 for '\0', 4 packetheader */
      if ((size % 4) != 0)
	 size += (4 - (size % 4));
      warn.pad3 = (char) size;
      sizes[SP_S_WARNING] = size;
      warn.type = SP_S_WARNING;
      warn.pad1 = STEXTE_STRING;
      warn.pad2 = (uchar) index;/* This is for future reference */
      sent_to_client[index] = 1;
      sendClientPacket((CVOID)&warn);
      sizes[SP_S_WARNING] = sizeof(struct warning_s_spacket);	/* Back to default */
   }
}

void
swarning(whichmessage, argument, argument2)
   unsigned char                   whichmessage, argument,
                                   argument2;
{
   struct warning_s_spacket        swarn;
   static int                      last_warn, 
				   last_whichmessage,
				   last_argument,
				   last_argument2;
   int                             n = mstime();

   if(n - last_warn < 750){
      if(last_whichmessage == whichmessage &&
	 last_argument == argument &&
	 last_argument2 == argument2)
      return;
   }
   last_warn = n;
   last_whichmessage = whichmessage;
   last_argument = argument;
   last_argument2 = argument2;

   swarn.type = SP_S_WARNING;
   swarn.whichmessage = whichmessage;
   swarn.argument = argument;
   swarn.argument2 = argument2;
   sendClientPacket((CVOID)&swarn);
}

void
swarning2(w_1, arg1_1, arg2_1, w_2, arg1_2, arg2_2)
   
   unsigned char	w_1, arg1_1, arg2_1, 
			w_2, arg1_2, arg2_2;
{
   struct warning_s_spacket	swarn;
   static int			last_warn,
				lw_1, larg1_1, larg2_1,
				lw_2, larg1_2, larg2_2;
   int				n = mstime();

   if(n - last_warn < 750){
      if(lw_1 == w_1 && larg1_1 == arg1_1 && larg2_1 == arg2_1 &&
         lw_2 == w_2 && larg1_2 == arg1_2 && larg2_2 == arg2_2)
	 return;
   }
   last_warn = n;
   lw_1 = w_1;
   larg1_1 = arg1_1;
   larg2_1 = arg2_1;
   lw_2 = w_2;
   larg1_2 = arg1_2;
   larg2_2 = arg2_2;
   
   swarn.type = SP_S_WARNING;
   swarn.whichmessage = w_1;
   swarn.argument = arg1_1;
   swarn.argument2 = arg2_1;
   sendClientPacket((CVOID)&swarn);

   swarn.whichmessage = w_2;
   swarn.argument = arg1_2;
   swarn.argument2 = arg2_2;
   sendClientPacket((CVOID)&swarn);
}

#endif
