#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/file.h>
#include <sys/errno.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <netdb.h>
#include <signal.h>
#include <syslog.h>

#if defined(__STDC__) || defined(__cplusplus)
#define P_(s) s
#else
#define P_(s) ()
#endif

/* ntstart.c */
void main P_((int argc, char **argv, char **envp));
void read_portfile P_((void));

#ifdef USE_SYS_PROTOS
extern uid_t getuid P_((void));
extern uid_t geteuid P_((void));
extern void perror P_((const char *));
extern int atoi P_((const char *));
extern int printf P_((const char *,...));
extern int fprintf P_((FILE *, const char *,...));
extern unsigned int alarm P_((unsigned int));
extern int close P_((int));
extern int fflush P_((FILE *));
extern int shutdown P_((int s, int how));
extern int socket P_((int domain, int type, int protocol));
extern int connect P_((int s, struct sockaddr * name, int namelen));
extern int getpeername P_((int s, struct sockaddr * name, int *namelen));
extern pid_t wait3 P_((union wait *, int, struct rusage *));
extern int bind P_((int s, struct sockaddr * name, int namelen));
extern int listen P_((int s, int backlog));
extern char *inet_ntoa P_((/* ??? */));
extern time_t time P_((time_t *));
extern int dup2 P_((int, int));
extern int ioctl P_((int, int, ...));
extern int accept P_((int s, struct sockaddr *addr, int *addrlen));
extern int setsockopt P_((int s, int level, int optname, char *optval, int optlen));
extern void                   bzero P_((void *, size_t));
extern int                    sscanf P_((const char *, const char *, ...));
extern int                    fclose P_((FILE *));
extern int                    select P_((int width, fd_set *readfds, fd_set *writefds, fd_set *exceptfds, struct timeval *timeout));
extern int                    fputc P_((int, FILE *));
#endif

#undef P_

/* port/program configuration file */
#define PORTFILE	"./.ports"
#define NTSTART		"./ntstart"
#define MAXPROG			16

static
struct progrecord {
   unsigned short	port;
   int			sock;
   char			prog[512], 
			progname[512],
			/* any special arguments */
			arg[4][64];
} prog[MAXPROG];

int	num_progs=0;

void
main(argc, argv, envp)

   int             argc;
   char          **argv, **envp;
{
   time_t	   curtime;
   struct sockaddr_in addr;
   extern int      errno;
   int             optval;
   static struct linger ling = {0, 0};
   register struct progrecord	*pr;

#if defined(SERVERDIR)

   /* we'll put this in so can be started from boot */

   if (chdir(SERVERDIR) < 0) {
      perror("ntstart'chdir");
      exit(1);
   }

   /* if we're started as root set the id to the owner of 'ntstart' */

   if (getuid() == 0 || geteuid() == 0) {
      struct stat     sbuf;
      if (stat(argv[0], &sbuf) < 0) {
	 perror("ntstart'stat");
      } else {
	 setuid(sbuf.st_uid);
	 setgid(sbuf.st_gid);
      }
   }
#endif

   read_portfile();
   if(!num_progs) exit(0);

   pr=prog;	/* use only one program to test */

   if ((pr->sock = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
      fprintf(stderr, "Socket create failed on socket\n");
      perror("ntstart'socket");
      pr->sock = -1;
      exit(1);
   }
   /* allow local address resuse */
   optval = 1;
   if (setsockopt(pr->sock, SOL_SOCKET, SO_REUSEADDR, (char *)&optval, 
      sizeof(int)) < 0) {
      perror("ntstart'setsockopt");
   }
   /* set linger on close */ 
   if (setsockopt(pr->sock, SOL_SOCKET, SO_LINGER, (char *)&ling, 
      sizeof(ling)) < 0) {
      perror("ntstart'startup:setsockopt(linger)");
   }

   /* create the address that we will bind to */
   addr.sin_family = AF_INET;
   addr.sin_addr.s_addr = INADDR_ANY;	/* wild card */
   addr.sin_port = htons(pr->port);

   /* try to bind to the address */
   if (bind(pr->sock, (struct sockaddr *)&addr, sizeof(addr)) < 0){
      if(errno == EADDRINUSE){
	 shutdown(pr->sock, 2);
	 close(pr->sock);
	 exit(0);
      }
      else {
	 fprintf(stderr, "Unknown error from bind");
	 perror("bind");
	 shutdown(pr->sock, 2);
	 close(pr->sock);
      }
   }
   else {
      /* otherwise ntstart must not be running */
      shutdown(pr->sock, 2);
      close(pr->sock);

      time(&curtime);
      fprintf(stderr, "Ntstart not running, restarting on %s", ctime(&curtime));
      fflush(stderr);
      execl(NTSTART, "ntstart", 0);
      perror("execl");
      _exit(1);
   }
}

void
read_portfile()

{
   FILE	*fi = fopen(PORTFILE, "r");
   char	buf[BUFSIZ];
   int	i=0, p;
   struct stat     sbuf;

   if(!fi){
      perror(PORTFILE);
      exit(1);
   }

   while(fgets(buf, BUFSIZ-1, fi)){
      if(sscanf(buf, "%d %s \"%[^\"]\" %s %s %s %s", &p,
				     prog[i].prog,
				     prog[i].progname,
				     prog[i].arg[0],
				     prog[i].arg[1],
				     prog[i].arg[2],
				     prog[i].arg[3]) >= 3){
	 prog[i].port = (unsigned short)p;
	 prog[i].sock = -1;
	 if (stat(prog[i].prog, &sbuf) < 0 || !(sbuf.st_mode & S_IRWXU)) {
	    fprintf(stderr, "\"%s\" not accessible or not executable.\n",
	       prog[i].prog);
	    fflush(stderr);
	 }
	 else{
	    i++;
	 }
      }
   }
   fclose(fi);

   num_progs = i;
}
